#if !defined (UNITTEST)
#include "globals.h"
#else
#include "StdAfx.h"
#endif
#include "ImeDoc.h"
#include "ImeBuffer.h"
#include "TMarker.h"
#include "jstring.h"
#include "ImeConfig.h"
#include "TCommuSession.h"

#define	NLOOP_IMEDOC	256

/*========================================================================
 *	public functions (ImeUI inferface)
 */
CImeDoc::CImeDoc () :
	m_pPC				(NULL),
	m_nRetPC			(0),
	m_nUnreadMessages	(0),
	m_nBuffer			(0),
	m_nUnusedBuffer		(0),
	m_pCurBuffer		(NULL),
	m_pbStackArea		(0),
	m_nStackHead		(0),
	m_nStackAreaSize	(0),
	m_nRegStackHead		(0),
	m_pRegStack			(NULL),
	m_nSignal			(LMSIGNAL_NONE),
	m_iThisCommand		(NFUNC_INVALID_CHAR),
	m_iLastCommand		(NFUNC_INVALID_CHAR),
	m_nbufMessage		(0),
	m_pPreCommandHook	(NULL),
	m_pPostCommandHook	(NULL),
	m_dwUpdateFlag		(0)
{
	return ;
}

CImeDoc::~CImeDoc ()
{
	if (m_pbStackArea != NULL) {
		delete[]	m_pbStackArea ;
		m_pbStackArea	= NULL ;
	}
	if (m_pRegStack != NULL) {
		delete[]	m_pRegStack ;
		m_pRegStack	= NULL ;
	}
	return ;
}

BOOL
CImeDoc::bInit ()
{
	CImeBuffer*		pBuffer ;
	int				i ;

	m_pPC				= CImeDoc::LM_iMessageLoop ;
	m_nRetPC			= 0 ;

	m_nUnreadMessages	= 0 ;
	m_nBuffer			= 0 ;
	m_nUnusedBuffer		= 0 ;
	m_pCurBuffer		= NULL ;
	m_pbStackArea		= 0 ;
	m_nStackHead		= 0 ;
	m_nStackAreaSize	= 0 ;
	m_nRegStackHead		= 0 ;
	m_pRegStack			= NULL ;

	m_nSignal			= LMSIGNAL_NONE ;
	m_iThisCommand		= NFUNC_INVALID_CHAR ;
	m_iLastCommand		= NFUNC_INVALID_CHAR ;
	memset (&m_LastEvent, 0, sizeof (m_LastEvent)) ;
	m_nbufMessage		= 0 ;
	m_pPreCommandHook	= NULL ;
	m_pPostCommandHook	= NULL ;
	m_dwUpdateFlag		= 0 ;

	if (! _bInitConfig ())
		return	FALSE ;

	pBuffer	= pCreateBuffer (NULL, 0, TRUE) ;
	if (pBuffer == NULL)
		return	FALSE ;

	m_pbStackArea			= new BYTE [NSIZE_IMEDOC_STACK] ;
	if (m_pbStackArea == NULL) 
		return	FALSE ;
	m_nStackAreaSize		= NSIZE_IMEDOC_STACK ;

	m_pRegStack				= new struct LMREG [NSIZE_REGSTACK] ;	//(struct LMREG*) MALLOC (sizeof (struct LMREG) * NSIZE_REGSTACK) ;
	if (m_pRegStack == NULL)
		return	FALSE ;

	for (i = 0 ; i < MYARRAYSIZE (m_rREGS) ; i ++)
		m_rREGS [i].m_nType	= LMREGTYPE_NIL ;

	m_pCurBuffer		= pBuffer ;

	/*	skk-mode-invoke
	 */
	m_pPreCommandHook	= CImeDoc::LM_bSkkPreCommand ;
	m_pPostCommandHook	= CImeDoc::LM_bSkkPostCommand ;

	(void) CTCommunicateSession::bInitPipeName () ;

	m_CandInfo.vClear () ;
	m_MessageInfo.vClear () ;
	return	TRUE ;
}

void
CImeDoc::vUninit ()
{
	int		i ;

	for (i = 0 ; i < m_nBuffer ; i ++) {
		if (m_rpBuffer [i] != NULL) 
			delete	m_rpBuffer [i] ;
		m_rpBuffer [i]	= NULL ;
	}
	for (i = 0 ; i < m_nUnusedBuffer ; i ++) {
		if (m_rpUnusedBuffer [i] != NULL) 
			delete	m_rpUnusedBuffer [i] ;
		m_rpUnusedBuffer [i]	= NULL ;
	}
	m_nBuffer = m_nUnusedBuffer = 0 ;

	_vUninitConfig () ;
	m_CandInfo.vClear () ;
	m_MessageInfo.vClear () ;
	return ;
}

void
CImeDoc::vClear ()
{
	int		i ;

	for (i = 0 ; i < MYARRAYSIZE (m_rREGS) ; i ++)
		m_rREGS [i].m_nType	= LMREGTYPE_NIL ;

	/*	stack ʒu擪ɖ߂B
	 */
	m_nRegStackHead		= 0 ;
	m_nStackHead		= 0 ;
	m_pPC				= CImeDoc::LM_iMessageLoop ;
	m_nRetPC			= 0 ;
	m_nUnreadMessages	= 0 ;
	m_nSignal			= LMSIGNAL_NONE ;

	for (i = 1 ; i < m_nBuffer ; i ++) {
		if (bDeleteBuffer (m_rpBuffer [i]))
			m_rpBuffer [i]	= NULL ;
	}
	if (m_rpBuffer [0] != NULL) {
		m_rpBuffer [0]->bClear () ;
		m_pCurBuffer	= m_rpBuffer [0] ;
		(void) m_pCurBuffer->bSetConversionMode (IMECMODE_HIRAGANA) ;
	} else {
		m_pCurBuffer	= NULL ;
	}
	m_iThisCommand		= NFUNC_INVALID_CHAR ;
	m_iLastCommand		= NFUNC_INVALID_CHAR ;
	memset (&m_LastEvent, 0, sizeof (m_LastEvent)) ;
	m_nbufMessage			= 0 ;
	m_CandInfo.vClear () ;
	m_MessageInfo.vClear () ;
	return ;
}

LPCWSTR
CImeDoc::pGetPreeditText (int* pnLength)
{
	LPCWSTR	strPreeditText ;
	int		nLength ;

	ASSERT (m_nBuffer > 0 && m_rpBuffer [0] != NULL) ;

	strPreeditText	= m_rpBuffer [0]->pBufferRawString (&nLength) ;

	if ((bRecursiveEditp () || m_rpBuffer [0]->bInputByCodeOrMenuModep () || m_rpBuffer [0]->bShowHenkanCandidatesModep ()) && nLength == 0) {
		/*	TF_SELECTION.RANGE  Empty  ITfContextView->GetTextExt ̕Ԃ RECT 
		 *	RAvP[V݂邽߁AҏW̃eLXgsB
		 *	(Empty ɂ͂ȂȂ悤)
		 *	ɂ Message Window (Status Window) ҏW̃eLXgɏdȂ肪
		 *	ɘa锤B܂iD悭Ȃ@ł邪cB
			strPreeditText	= L" " ;
			nLength			= 1 ;
		 *
		 *	ƁAӂƎv̂ǁAStatus Window \󔒂ɖ߂ƂǂȂ̂
		 *	낤H
		 *	݂̎sȂA L" " 𗘗p邵Ȃ낤ǁcB
		 */
		strPreeditText	= L"" ;
		nLength			= 0 ;
	}
	if (pnLength != NULL)
		*pnLength	= nLength ;
	return	strPreeditText ;
}

BOOL
CImeDoc::bGetPreeditCursor (int* pnCursor) const
{
	CTMarker*	pmkPoint ;

	ASSERT (m_nBuffer > 0 && m_rpBuffer [0] != NULL) ;
	ASSERT (pnCursor != NULL) ;

	if (! m_rpBuffer [0]->bGetMarker (CImeBuffer::MARKER_POINT, &pmkPoint) || pmkPoint == NULL)
		return	FALSE ;
	*pnCursor	= pmkPoint->iGetPosition () ;
	return	TRUE ;
}

#if !defined (UNITTEST)
const IMECANDIDATES*
CImeDoc::pGetStatusText () const
{
	if (m_pCurBuffer->bShowHenkanCandidatesModep () ||
		m_pCurBuffer->bInputByCodeOrMenuModep () ||
		m_pCurBuffer->bInputByCodeOrMenu1Modep ()) {
		return	m_CandInfo.bIsEnabledp ()? &m_CandInfo : NULL ;
	} else {
		if (m_rpBuffer [0] != m_pCurBuffer) 
			return	m_MessageInfo.bIsEnabledp ()? &m_MessageInfo : NULL ;
	}
	return	NULL ;
}
#else
LPCWSTR
CImeDoc::pGetStatusText (int* pnLength)
{
	ASSERT (m_nBuffer > 0 && m_rpBuffer [0] != NULL) ;
	ASSERT (pnLength != NULL) ;

	if (m_nbufMessage > 0) {
		*pnLength	= m_nbufMessage ;
		return	m_bufMessage ;
	} else {
		if (m_rpBuffer [0] != m_pCurBuffer)
			return	m_pCurBuffer->pBufferRawString (pnLength) ;
	}
	*pnLength	= 0 ;
	return	NULL ;
}
#endif

BOOL
CImeDoc::bGetStatusCursor (int* pnCursor) const
{
	CTMarker*	pmkPoint ;

	ASSERT (pnCursor != NULL) ;

#if 0
	if (m_nbufMessage > 0)
		return	FALSE ;
	if (m_rpBuffer [0] == m_pCurBuffer)
		return	FALSE ;

	if (! m_pCurBuffer->bGetMarker (CImeBuffer::MARKER_POINT, &pmkPoint) || pmkPoint == NULL)
		return	FALSE ;
	*pnCursor	= pmkPoint->iGetPosition () ;
#else
	if (m_rpBuffer [0] != m_pCurBuffer && ! bShowHenkanCandidatesModep () && ! bInputByCodeOrMenuModep (NULL)) {
		if (! m_pCurBuffer->bGetMarker (CImeBuffer::MARKER_POINT, &pmkPoint) || pmkPoint == NULL)
			return	FALSE ;
		*pnCursor	= pmkPoint->iGetPosition () ;
	} else {
		return	FALSE ;
	}
#endif
	return	TRUE ;
}

BOOL
CImeDoc::bIsStatusActivep () const
{
	return	(m_rpBuffer [0] != m_pCurBuffer) ;
}

BOOL
CImeDoc::bRecursiveEditp () const
{
	return	(m_nBuffer > 1 && m_pCurBuffer != m_rpBuffer [0])? TRUE : FALSE ;
}

BOOL
CImeDoc::bQueryFilterEvent	(
	const struct TMSG*		pMsg,
	BOOL*					pfEaten)
{
	BOOL	bEaten ;

	if (pMsg == NULL)
		return	FALSE ;

	/*	͑S process ƂĂBɏCB
	 */
	bEaten	= FALSE ;
	if (bRecursiveEditp ()) {
		bEaten	= TRUE ;
	} else {
		if (m_pCurBuffer->bShowHenkanCandidatesModep () ||
			m_pCurBuffer->bInputByCodeOrMenuModep ()) {
			bEaten	= TRUE ;
		} else if (m_pCurBuffer->bHavePrefixp ()) {
			/*	prefix ݂΁Askk-pre-command sȂ΂ȂȂ̂ŁA
			 *	eaten
			 */
			bEaten	= TRUE ;
		} else {
			int		nText, nFuncNo ;

			/* ŎۂɐHׂL[ǂ`FbNB*/
			(void) m_pCurBuffer->pBufferString (&nText) ;
			if (nText > 0) {
				/* 炩̃eLXg͂Ă΁AL[ꉞHׂB*/
				bEaten	= TRUE ;
			} else {
				if (bLookupKeymap (pMsg, &nFuncNo) && _bProcessFunctionp (nFuncNo))
					bEaten	= TRUE ;
			}
		}
	}
	if (pfEaten != NULL)
		*pfEaten	= bEaten ;
	return	TRUE ;
}

BOOL
CImeDoc::bFilterEvent	(
	const struct TMSG*		pMsg)
{
	BOOL	bRetval ;
	int		iConversionMode ;

	if (! bPushEvent (pMsg))
		return	FALSE ;

	/* Cxǵuvɐݒ肳B*/
	m_UnprocessEvent.m_nMessage	= WM_NULL ;
	/* update flag ͏B*/
	vClearUpdateFlag () ;
	vClearMessage () ;
	iConversionMode	= iGetConversionMode () ;
	bRetval	= _bTick () ;
	if (bRetval && iConversionMode != iGetConversionMode ()) {
		vSetUpdateFlag (IMEDOC_UPDATE_CONVERSIONMODE) ;
	}
	/* status candidate  update */
	_bUpdateCandidateInfoForStatusText () ;
	return	bRetval ;
}

BOOL
CImeDoc::bQueryToggleIMEEvent	(
	const struct TMSG*		pMsg,
	BOOL*					pfEaten)
{
	BOOL	bEaten ;
	int		nFuncNo ;

	if (pMsg == NULL)
		return	FALSE ;

	if (bLookupKeymap (pMsg, &nFuncNo) && nFuncNo == NFUNC_TOGGLE_IME) {
		bEaten	= TRUE ;
	} else {
		bEaten	= FALSE ;
	}
	if (pfEaten != NULL)
		*pfEaten	= bEaten ;
	return	TRUE ;
}

const struct TMSG*
CImeDoc::pGetUnprocessEvent ()
{
	if (m_UnprocessEvent.m_nMessage == WM_NULL)
		return	NULL ;

	return	&m_UnprocessEvent ;
}

int
CImeDoc::iGetReadingText (
	LPWSTR					pwReading,
	int						nReadingSize,
	int						nCompPosition,
	int*					pnReadingPosition)
{
	/*	buffer class  (henkan-start-point, henkan-end-point, henkan-key) ̃yA
	 *	L^BāAbuffer-string 鎞 (henkan-start-point, henkan-end-point) 
	 *	 henkan-key ɒuB
	 *	ő reading-text ͂ȂɎ擾ł锤B(top-buffer ł̂ݎ΂
	 *	v)
	 */
	if (pwReading == NULL || m_rpBuffer [0] == NULL)
		return	0 ;

	return	m_rpBuffer [0]->iGetReadingText (pwReading, nReadingSize, nCompPosition, pnReadingPosition) ;
}

/*	ϊ[h؂ւBtopbuffer ؂ւ̂͗ǂ̂Arecursive-edit-mode
 *	ꍇɂ͗ǂ̂낤H ͎̂dl̍ă`FbNKv낤B
 */
BOOL
CImeDoc::bSetConversionMode (
	int						nConversionMode)
{
	CImeBuffer*	pBuffer ;
	struct TMSG	msg ;
	int			nCurMode ;

	if (m_pCurBuffer == NULL)
		return	FALSE ;

	pBuffer	= m_pCurBuffer ;

	/*	ǂ̂悤ɂ conversion-mode ύXΗǂ̂낤H
	 *	܂A[hA[hAskk-show-henkan-candidates-mode 
	 *	skk-input-by-code-or-menu-mode 肵 conversion-mode ύXƂ
	 *	͉wׂȂ̂낤H
	 */
	nCurMode		= iGetConversionMode () ;
	if (nCurMode == nConversionMode)
		return	TRUE ;

	/* Cxǵuvɐݒ肳B*/
	m_UnprocessEvent.m_nMessage	= WM_NULL ;
	/* update flag ͏B*/
	vClearUpdateFlag () ;			

	msg.m_nMessage	= WM_LM_COMMAND ;
	msg.m_pvData	= NULL ;
	msg.m_lParam	= 0 ;
	msg.m_nTime		= GetTickCount () ;
	msg.m_pt.x		= msg.m_pt.y	= 0 ;

	switch (nConversionMode) {
	case	IMECMODE_HIRAGANA:
	case	IMECMODE_KATAKANA:
		/* conversion-mode -> hiragana ւ route */
		if (! pBuffer->bJModep ()) {
			/* skk-mode-on ĂяoB*/
			msg.m_wParam	= NFUNC_SKK_MODE ;
			if (! bPushEvent (&msg))
				return	FALSE ;
			/* katakana-mode  FALSE ɂȂBŊB*/
			if (nConversionMode == IMECMODE_KATAKANA) {
				msg.m_wParam	= NFUNC_SKK_TOGGLE_CHARACTERS ;
				if (! bPushEvent (&msg))
					return	FALSE ;
			}
		} else {
			if (pBuffer->bLatinModep () || pBuffer->bJisx0208LatinModep ()) {
				/* skk-kakutei ł܂ j-mode ɂB*/
				msg.m_wParam	= NFUNC_SKK_MODE ;
				if (! bPushEvent (&msg))
					return	FALSE ;
				if ((  pBuffer->bKatakanaModep () && nConversionMode == IMECMODE_HIRAGANA) ||
					(! pBuffer->bKatakanaModep () && nConversionMode == IMECMODE_KATAKANA)) {
					msg.m_wParam	= NFUNC_SKK_TOGGLE_CHARACTERS ;
					if (! bPushEvent (&msg))
						return	FALSE ;
				}
			} else if (pBuffer->bJisx0201Modep ()) {
				/* ܂ĂȂB*/
				/* toggle-jisx0201 ĂяoB*/
				/* ̌Akatakana-mode ǂ肾cB*/
				/* ̌Atoggle-characters */
			} else if (pBuffer->bKatakanaModep ()) {
				if (nConversionMode == IMECMODE_HIRAGANA) {
					msg.m_wParam	= NFUNC_SKK_TOGGLE_CHARACTERS ;
					if (! bPushEvent (&msg))
						return	FALSE ;
				}
			} else {
				if (nConversionMode == IMECMODE_KATAKANA) {
					msg.m_wParam	= NFUNC_SKK_TOGGLE_CHARACTERS ;
					if (! bPushEvent (&msg))
						return	FALSE ;
				}
			}
		}
		break ;
	case	IMECMODE_ZENKAKU:
	case	IMECMODE_ASCII:
		if (! pBuffer->bJModep  ()) {
			/* skk-mode-on ĂяoB*/
			msg.m_wParam	= NFUNC_SKK_MODE ;
			if (! bPushEvent (&msg))
				return	FALSE ;
		}
		msg.m_wParam	= (nConversionMode == IMECMODE_ZENKAKU)? NFUNC_SKK_JISX0208_LATIN_MODE : NFUNC_SKK_LATIN_MODE ;
		if (! bPushEvent (&msg))
			return	FALSE ;
		break ;
	case	IMECMODE_HANKANA:
		/* B*/
		return	FALSE ;
	case	IMECMODE_OFF:
		/*	 IME ƂƂƎvǁcɗ邱Ƃ͂ȂƎvB */
		return	FALSE ;
	}
	if (! _bTick ()) {
		vClear () ;
	}
	return	FALSE ;
}

int
CImeDoc::iGetConversionMode ()
{
	ASSERT (m_nBuffer > 0 && m_rpBuffer [0] != NULL) ;
	return	m_rpBuffer [0]->iGetConversionMode () ;
}

BOOL
CImeDoc::bSetConversionString (
	LPCWSTR				wstrText,
	int					nText)
{
	/*	̓[hɂē삪قȂB
	 */
	vClear () ;
	return	m_pCurBuffer->bSetConversionString (wstrText, nText, ImeConfig_bSkkCompositionAutoShiftp ()) ;
}

BOOL
CImeDoc::bGetSelectedRegion (
	int*				pnStartPos,
	int*				pnEndPos)
{
	return	m_pCurBuffer->bGetSelectedRegion (pnStartPos, pnEndPos) ;
}

BOOL
CImeDoc::bGetConvertedRegion (
	int*				pnStartPos,
	int*				pnEndPos)
{
	return	m_pCurBuffer->bGetConvertedRegion (pnStartPos, pnEndPos) ;
}

BOOL
CImeDoc::bQueryUpdateContext (
	int*				pnShift,
	int*				pnCursor,
	BOOL*				pfContinue)
{
	ASSERT (m_nBuffer > 0 && m_rpBuffer [0] != NULL) ;
	ASSERT (pnShift    != NULL) ;
	ASSERT (pnCursor   != NULL) ;
	ASSERT (pfContinue != NULL) ;

	if (m_rpBuffer [0] != m_pCurBuffer) {
		CTMarker*	pmkPoint ;

		if (! m_rpBuffer [0]->bGetMarker (CImeBuffer::MARKER_POINT, &pmkPoint) || pmkPoint == NULL)
			return	FALSE ;

		*pnShift	= 0 ;
		*pnCursor	= pmkPoint->iGetPosition () ;
		*pfContinue	= TRUE ;
		return	TRUE ;
	}

	/*	Vtgł͂ȂA newline ɂ鋭 update ĂȂꍇɂ
	 *	ȂBcontinue ǂ͏󋵂ɂB
	 */
	if (! ImeConfig_bSkkCompositionAutoShiftp () && (m_dwUpdateFlag & IMEDOC_NEWLINE) == 0) {
		CTMarker*	pmkPoint ;

		if (! m_rpBuffer [0]->bGetMarker (CImeBuffer::MARKER_POINT, &pmkPoint) || pmkPoint == NULL)
			return	FALSE ;

		*pnShift	= 0 ;
		if (bProcessEventp ()) {
			*pnCursor	= pmkPoint->iGetPosition () ;
			*pfContinue	= TRUE ;
		} else {
			*pnCursor	= 0 ;
			*pfContinue	= FALSE ;
		}
		return	TRUE ;
	}
	if (ImeConfig_bSkkIsNewlineKakuteiAllp () && (m_dwUpdateFlag & IMEDOC_NEWLINE) != 0) {
		if (! m_rpBuffer [0]->bSkkHenkanActivep () && m_rpBuffer [0]->pSkkGetSkkCurrentRuleTree () == NULL) {
			CTMarker*	pmkPoint ;
			CTMarker*	pmkBufferEnd ;
			int			iMove ;

			if (! m_rpBuffer [0]->bGetMarker (CImeBuffer::MARKER_POINT, &pmkPoint) || pmkPoint == NULL)
				return	FALSE ;
			if (! m_rpBuffer [0]->bGetMarker (CImeBuffer::MARKER_BUFFEREND, &pmkBufferEnd) || pmkBufferEnd == NULL)
				return	FALSE ;
			iMove	= pmkBufferEnd->iGetPosition () - pmkPoint->iGetPosition () ;
			if (! pmkPoint->bForward (iMove))
				return	FALSE ;
		}
	}
	return	m_rpBuffer [0]->bQueryUpdateContext (pnShift, pnCursor, pfContinue) ;
}

BOOL
CImeDoc::bUpdateContext ()
{
	ASSERT (m_nBuffer > 0 && m_rpBuffer [0] != NULL) ;

	if (m_rpBuffer [0] != m_pCurBuffer) 
		return	TRUE ;

	/*	Vtgł͂ȂA newline ɂ鋭 update ĂȂꍇɂ
	 *	ȂBcontinue ǂ͏󋵂ɂB
	 */
	if (! ImeConfig_bSkkCompositionAutoShiftp () && (m_dwUpdateFlag & IMEDOC_NEWLINE) == 0) 
		return	TRUE ;
	return	m_rpBuffer [0]->bUpdateContext () ;
}

void
CImeDoc::vUpdateConfig ()
{
	/*	Configuration  Update ̂ŁAԂɖ߂BRuleTree ǂĂr 
	 *	RuleTree XVꂽ肷Ɨ邵ȂB
	 */
//	_vUninitConfig () ;
	vClear () ;
	return ;
}

BOOL
CImeDoc::bRevertText ()
{
	int		nConversionMode ;

	/*	eLXgSLZB̏ꍇ͑S buffer  clear 
	 *	stack āAmessage-loop  pointer ߂΂̂B
	 */
	nConversionMode	= m_rpBuffer [0]->iGetConversionMode () ;
	vClear () ;
	/*	Clear Ԃ SetConversionMode ͗eՁB
	 */
	m_rpBuffer [0]->bSetConversionMode (nConversionMode) ;
	return	TRUE ;
}

BOOL
CImeDoc::bCompleteText ()
{
	CImeBuffer*		pBuffer ;
	struct TMSG		msg ;
	int				nLoop, nConversionMode ;

	/*	eLXgSmB( top ȊO cancel )
	 *	safety ̂߂ɍő僋[v񐔂ݒ肵ĂB
	 */
	for (nLoop = NLOOP_IMEDOC ; nLoop > 0 ; nLoop --) {
		pBuffer	= pGetCurrentBuffer () ;
		if (pBuffer->bInputByCodeOrMenuModep () ||
			pBuffer->bShowHenkanCandidatesModep ()) {
			/*	LZ̃L[pӁB
			 *	{ keyboard-quit  abort-recursive-edit 蓖ĂĂ
			 *	key Ȃ cancel Ƃē삷ƂR[hɂȂ΂ȂȂ
			 *	 > input-by-code-or-menu  henkan-show-candidates
			 */
			msg.m_nMessage	= WM_CHAR ;
			msg.m_wParam	= 0x07 ;			/*,.. */
			msg.m_lParam	= 0 ;
			msg.m_nTime		= GetTickCount () ;
			msg.m_pt.x		= msg.m_pt.y	= 0 ;
		} else {
			if (! bRecursiveEditp ())
				break ;

			if (! bWhereIsInternal (NFUNC_ABORT_RECURSIVE_EDIT, &msg)) {
				/* dȂ̂ŁAOR}h𔭍sB*/
				msg.m_nMessage	= WM_LM_COMMAND ;
				msg.m_nTime		= GetTickCount () ;
				msg.m_wParam	= NFUNC_ABORT_RECURSIVE_EDIT ;
				msg.m_lParam	= 0 ;
				msg.m_pt.x		= msg.m_pt.y	= 0 ;
			}
		} 
		if (! bPushEvent (&msg) || 
			! _bTick ()) {
			/*	̓NeBJȁc󋵂Ȃ̂ŁAB
			 */
			vClear () ;
			goto	exit_func ;
		}
	}

	pBuffer	= pGetCurrentBuffer () ;
	if (pBuffer->bInputByCodeOrMenuModep () ||
		pBuffer->bShowHenkanCandidatesModep () ||
		bRecursiveEditp () ||
		m_pPC != CImeDoc::LM_iMessageLoop) {
		vClear () ;
		goto	exit_func ;
	}
	/*	̒iK topbuffer ɓBĂ锤B
	 *	m肵ĂA̓[h̒킹B
	 */
	pBuffer	= pGetCurrentBuffer () ;
	nConversionMode		= pBuffer->iGetConversionMode  () ;
	if (! bWhereIsInternal (NFUNC_SKK_KAKUTEI, &msg)) {
		msg.m_nMessage	= WM_LM_COMMAND ;
		msg.m_nTime		= GetTickCount () ;
		msg.m_wParam	= NFUNC_SKK_KAKUTEI ;
		msg.m_lParam	= 0 ;
		msg.m_pt.x		= msg.m_pt.y	= 0 ;
	}
	if (! bPushEvent (&msg) || 
		! _bTick ()) {
		vClear () ;
		goto	exit_func ;
	}
	pBuffer->bSetConversionMode (nConversionMode) ;

	/*	̏Ԃ buffer-string 𓾂āAꂩ clear ΗǂB
	 */
exit_func:
	return	TRUE ;
}

void
CImeDoc::vSetUpdateFlag (
	unsigned int		uFlag)
{
	if ((uFlag & IMEDOC_CLOSE_CANDIDATELIST) != 0 && (uFlag & IMEDOC_CREATE_CANDIDATELIST) != 0) {
		/* 쐬Ɣj͓ɂ͂łȂB*/
		uFlag	&= ~IMEDOC_CREATE_CANDIDATELIST ;
	}
	if ((uFlag & IMEDOC_CLOSE_CANDIDATELIST) != 0) {
		/*	ŏII CANDIDATELIST ĵȂAr CREATE ͈ӖȂȂB
		 */
		m_dwUpdateFlag	&= ~(IMEDOC_CREATE_CANDIDATELIST | IMEDOC_UPDATE_CANDIDATELIST) ;
	}

	/*	ɂ CREATE_CANDIDATELIST  CLOSE_CANDIDATELIST ɑ݂̂
	 *	CLOSE -> CREATE ̏ԂŎsꂽłB
	 */
	m_dwUpdateFlag	|= uFlag ;
	return ;
}

void
CImeDoc::vClearUpdateFlag ()
{
	m_dwUpdateFlag	= 0 ;
	return ;
}

unsigned int
CImeDoc::uGetUpdateFlag ()
{
	return	m_dwUpdateFlag ;
}

BOOL
CImeDoc::bShowHenkanCandidatesModep () const
{
	if (m_pCurBuffer == NULL)
		return	FALSE ;
	return	m_pCurBuffer->bShowHenkanCandidatesModep  () ;
}

BOOL
CImeDoc::bInputByCodeOrMenuModep (
	BOOL*				pbMenu1p) const
{
	if (m_pCurBuffer == NULL)
		return	FALSE ;
	if (! m_pCurBuffer->bInputByCodeOrMenuModep  ())
		return	FALSE ;
	if (pbMenu1p != NULL) {
		/* MenuJump or Menu1Jump */
		*pbMenu1p	= m_pCurBuffer->bInputByCodeOrMenu1Modep  () ;
	}
	return	TRUE ;
}

/*	܂g͂ȂBB
 *	̓́AoݒɂĕωBoƁA}[NȂɂ͓Ȃ
 *	ȂȂʂɂB
 */
BOOL
CImeDoc::bSetReconvertText (
	LPCWSTR				wstrText,
	int					nTextLen)
{
	vClear () ;
	return	m_rpBuffer [0]->bSetReconvertText (wstrText, nTextLen) ;
}

IMECANDIDATES*
CImeDoc::pGetCandidateInfo ()
{
	return	&m_CandInfo ;
}

void
CImeDoc::vClearCandidateInfo ()
{
	m_CandInfo.vClear () ;
	return ;
}

int
CImeDoc::iGetCandidateText (
	LPWSTR					strDest,
	int						nDest,
	TEXTREGION*				pTextHead,
	int*					piNumText,
	const IMECANDIDATES*	pMyCand)
{
	LPWSTR		pDest ;
	int			iCount, nKeys, iNumText, nTotalCandLen ;
	UINT		iPageStart, iNextPage, iMaxPage, i ;
	TEXTREGION*	pText ;
	TEXTREGION*	pTextLast ;
	const UINT*	piPageIndex ;
	WCHAR		ch ;
	LPCWSTR		pCandKeyTbl ;
	LPCWSTR		pwCandBase ;
	const UINT*	puCandIndexBase ;
	const UINT*	puAnnoIndexBase ;

	pDest		= strDest ;
	iCount		= 0 ;
	piPageIndex	= pMyCand->m_vbufPageIndex.pGetBuffer () ;
	iMaxPage	= pMyCand->m_vbufPageIndex.iGetUsage () ;
	iPageStart	= piPageIndex [pMyCand->m_iCurrentPage] ;
	iNextPage	= ((pMyCand->m_iCurrentPage + 1) < iMaxPage)? (piPageIndex [pMyCand->m_iCurrentPage + 1]) : pMyCand->m_iCount ;
	i			= iPageStart ;
	pCandKeyTbl	= ImeConfig_pGetSkkHenkanShowCandidatesKeys (&nKeys) ;

	iNumText	= (piNumText != NULL)? *piNumText : 0 ;
	pText		= pTextHead ;
	if (piNumText != NULL && pText != NULL) {
		pTextLast	= pText + iNumText ;
	} else {
		pTextLast	= NULL ;
	}
	
#define	SAFE_COPY(dest,ndest,src,nsrc)	{ \
	LPWSTR	_pDest	= (dest) ;	\
	int		_nDest	= (ndest) ; \
	int		_nSrc	= (nsrc) ; \
	int		_n		= (_nDest < _nSrc)? _nDest : _nSrc ; \
	if (_n > 0) { \
		wcsncpy (_pDest, (src), _n) ;	\
		_pDest	+= _n ;	\
		_nDest	-= _n ;	\
		(dest)	= _pDest ;	\
		(ndest)	= _nDest ;	\
	}	\
}

	pwCandBase		= pMyCand->m_vbufCandidate.pGetBuffer () ;
	nTotalCandLen	= pMyCand->m_vbufCandidate.iGetUsage () ;
	puCandIndexBase	= pMyCand->m_vbufCandidateIndex.pGetBuffer () ;
	puAnnoIndexBase	= pMyCand->m_vbufAnnotationIndex.pGetBuffer () ;
	/* dwPageStart ŕ\₩ dwPageSize \B*/
	while (i < iNextPage && i < pMyCand->m_iCount && iCount < nKeys && nDest > 0) {
		LPCWSTR		pwCandidate ;
		int			nAnnotationIndex, nCandLen ;

		ch			= *(pCandKeyTbl + iCount) ;
		SAFE_COPY (pDest, nDest, &ch, 1) ;
		SAFE_COPY (pDest, nDest, L":", 1) ;

		pwCandidate	= pwCandBase + *(puCandIndexBase + i) ;
		nCandLen	= lstrlenW (pwCandidate) ;
		nAnnotationIndex	= (i < pMyCand->m_vbufAnnotationIndex.iGetUsage ())? *(puAnnoIndexBase + i) : 0 ;

		/*	 annotation ݂ꍇB*/
		if (pText != NULL && pText < pTextLast) {
			if (nDest > 0 && nAnnotationIndex > 0 && nAnnotationIndex < nTotalCandLen && lstrlenW (pwCandBase + nAnnotationIndex) > 0) {
				pText->m_nCandidate	= i ;
				pText->m_nOffset	= pDest - strDest ;
				pText->m_nLength	= (nDest < nCandLen)? nDest : nCandLen ;
			} else {
				pText->m_nCandidate	= 0 ;
				pText->m_nOffset	= 0 ;
				pText->m_nLength	= 0 ;
			}
			pText	++ ;
		}
		SAFE_COPY (pDest, nDest, pwCandidate, nCandLen) ;
		SAFE_COPY (pDest, nDest, L" ", 1) ;
		iCount	++ ;
		i		++ ;
	}
	if (nDest > 0) {
		WCHAR	buffer [32] ;
		int		n ;

		/*	+ \t悤ɂȂ̂ǂ͖B
		 */
		n	= wnsprintfW (buffer, ARRAYSIZE (buffer), L" [c %d]", pMyCand->m_iCount - i) ;
		SAFE_COPY (pDest, nDest, buffer, n) ;
	}
	if (pText != NULL && piNumText != NULL)
		*piNumText	= pText - pTextHead ;

#undef	SAFE_COPY
	return	(pDest - strDest) ;
}

int
CImeDoc::iGetCodeMenuJumpText (
	LPWSTR					strDest,
	int						nDest,
	const IMECANDIDATES*	pMyCand)
{
	LPWSTR		pDest ;
	int			iCount, nKeys ;
	UINT		iPageStart, iNextPage, iMaxPage, i ;
	const UINT*	piPageIndex ;
	WCHAR		ch ;
	LPCWSTR		pCandKeyTbl ;

	pDest		= strDest ;
	iCount		= 0 ;
	piPageIndex	= pMyCand->m_vbufPageIndex.pGetBuffer () ;
	iMaxPage	= pMyCand->m_vbufPageIndex.iGetUsage () ;
	iPageStart	= piPageIndex [pMyCand->m_iCurrentPage] ;
	iNextPage	= ((pMyCand->m_iCurrentPage + 1) < iMaxPage)? (piPageIndex [pMyCand->m_iCurrentPage + 1]) : pMyCand->m_iCount ;
	i			= iPageStart ;
	pCandKeyTbl	= ImeConfig_pGetSkkInputByCodeMenuKeys1 (&nKeys) ;
	
#define	SAFE_COPY(dest,ndest,src,nsrc)	{ \
	LPWSTR	_pDest	= (dest) ;	\
	int		_nDest	= (ndest) ; \
	int		_nSrc	= (nsrc) ; \
	int		_n		= (_nDest < _nSrc)? _nDest : _nSrc ; \
	if (_n > 0) { \
		wcsncpy (_pDest, (src), _n) ;	\
		_pDest	+= _n ;	\
		_nDest	-= _n ;	\
		(dest)	= _pDest ;	\
		(ndest)	= _nDest ;	\
	}	\
}

	/* dwPageStart ŕ\₩ dwPageSize \B*/
	while (i < iNextPage && i < pMyCand->m_iCount && iCount < nKeys && nDest > 0) {
		LPCWSTR		pwCandidate ;
		int			nCandLen ;

		ch			= *(pCandKeyTbl + iCount) ;
		SAFE_COPY (pDest, nDest, &ch, 1) ;
		SAFE_COPY (pDest, nDest, L":", 1) ;

		pwCandidate	= (LPCWSTR)pMyCand->m_vbufCandidate.pGetBuffer () + *(pMyCand->m_vbufCandidateIndex.pGetBuffer () + i) ;
		nCandLen	= lstrlenW (pwCandidate) ;

		SAFE_COPY (pDest, nDest, pwCandidate, nCandLen) ;
		SAFE_COPY (pDest, nDest, L" ", 1) ;
		iCount	++ ;
		i		++ ;
	}
#undef	SAFE_COPY
	return	(pDest - strDest) ;
}

int
CImeDoc::iGetCodeMenu1Text (
	LPWSTR					strDest,
	int						nDest,
	const IMECANDIDATES*	pMyCand)
{
	LPWSTR		pDest ;
	int			iCount, nKeys ;
	UINT		iPageStart, iNextPage, iMaxPage, i ;
	const UINT*	piPageIndex ;
	WCHAR		ch ;
	LPCWSTR		pCandKeyTbl ;

	pDest		= strDest ;
	iCount		= 0 ;
	piPageIndex	= pMyCand->m_vbufPageIndex.pGetBuffer () ;
	iMaxPage	= pMyCand->m_vbufPageIndex.iGetUsage () ;
	iPageStart	= piPageIndex [pMyCand->m_iCurrentPage] ;
	iNextPage	= ((pMyCand->m_iCurrentPage + 1) < iMaxPage)? (piPageIndex [pMyCand->m_iCurrentPage + 1]) : pMyCand->m_iCount ;
	i			= iPageStart ;
	pCandKeyTbl	= ImeConfig_pGetSkkInputByCodeMenuKeys2 (&nKeys) ;
	
#define	SAFE_COPY(dest,ndest,src,nsrc)	{ \
	LPWSTR	_pDest	= (dest) ;	\
	int		_nDest	= (ndest) ; \
	int		_nSrc	= (nsrc) ; \
	int		_n		= (_nDest < _nSrc)? _nDest : _nSrc ; \
	if (_n > 0) { \
		wcsncpy (_pDest, (src), _n) ;	\
		_pDest	+= _n ;	\
		_nDest	-= _n ;	\
		(dest)	= _pDest ;	\
		(ndest)	= _nDest ;	\
	}	\
}

	/* dwPageStart ŕ\₩ dwPageSize \B*/
	while (i < iNextPage && i < pMyCand->m_iCount && iCount < nKeys && nDest > 0) {
		LPCWSTR		pwCandidate ;
		int			nCandLen ;

		ch			= *(pCandKeyTbl + iCount) ;
		SAFE_COPY (pDest, nDest, &ch, 1) ;
		SAFE_COPY (pDest, nDest, L":", 1) ;

		pwCandidate	= (LPCWSTR)pMyCand->m_vbufCandidate.pGetBuffer () + *((const int*)pMyCand->m_vbufCandidateIndex.pGetBuffer () + i) ;
		nCandLen	= lstrlenW (pwCandidate) ;

		SAFE_COPY (pDest, nDest, pwCandidate, nCandLen) ;
		SAFE_COPY (pDest, nDest, L" ", 1) ;
		iCount	++ ;
		i		++ ;
	}
#undef	SAFE_COPY
	return	(pDest - strDest) ;
}

BOOL
CImeDoc::bHaveMessagep () const
{
	return	m_nbufMessage > 0 ;
}

LPCWSTR
CImeDoc::pGetMessage (int* piLength) const 
{
	if (piLength != NULL)
		*piLength	= m_nbufMessage ;
	return	m_bufMessage ;
}

/*========================================================================
 *	public functions (ImeDoc internal)
 */
BOOL
CImeDoc::bSkkPreviousCandidateCharp	(
	int						nCH)
{
	struct TMSG		msg ;
	int				nFunc ;

	msg.m_nMessage	= WM_CHAR ;
	msg.m_wParam	= nCH ;
	msg.m_lParam	= 0 ;
	msg.m_pvData	= NULL ;
	msg.m_nTime		= GetTickCount () ;
	msg.m_pt.x		= 0 ;
	msg.m_pt.y		= 0 ;

	if (bLookupKeymap (&msg, &nFunc))
		return	(nFunc == NFUNC_SKK_PREVIOUS_CANDIDATE) ;

	return	FALSE ;
}

/*========================================================================
 *	public functions (ImeDoc internal)
 */
void
CImeDoc::vSetCurrentBuffer (
	CImeBuffer*		pBuffer)
{
	m_pCurBuffer	= pBuffer ;
	return ;
}

CImeBuffer*
CImeDoc::pGetCurrentBuffer ()
{
	return	m_pCurBuffer ;
}

BOOL
CImeDoc::bSetMessage (
	LPCWSTR					wstrMessage)
{
	if (wstrMessage == NULL)
		return	FALSE ;
	return	bSetMessageN (wstrMessage, lstrlenW (wstrMessage)) ;
}

BOOL
CImeDoc::bSetMessageN (
	LPCWSTR					wstrMessage,
	int						nstrMessage)
{
	ASSERT (wstrMessage != NULL || nstrMessage == 0) ;

	m_nbufMessage	= (nstrMessage > MSGBUFSIZE)? MSGBUFSIZE : nstrMessage ;
	if (wstrMessage != NULL)
		memcpy (m_bufMessage, wstrMessage, m_nbufMessage * sizeof (WCHAR)) ;

	return	TRUE ;
}

void
CImeDoc::vClearMessage ()
{
	m_nbufMessage		= 0 ;
	return ;
}

BOOL
CImeDoc::bGetLastCommandChar (
	WCHAR*						pwCH)
{
	int		iCH ;

	/* printable łȂꍇɂ false ԂB*/
	iCH	= iGetCharFromEvent (&m_LastEvent) ;
	if (! (0x20 <= iCH && iCH < 0x7F)) {
		if (! ImeConfig_bSkkSpecialMidashiCharp		(iCH) && 
			! ImeConfig_bSkkSetHenkanPointKeyp		(iCH) &&
			! ImeConfig_bSkkStartHenkanCharp		(iCH) &&
			! ImeConfig_bSkkTryCompletionCharp		(iCH) &&
			! ImeConfig_bSkkNextCompletionCharp		(iCH) &&
			! ImeConfig_bSkkPreviousCompletionCharp	(iCH)) {
			return	FALSE ;
		}
	}
	if (pwCH != NULL)
		*pwCH	= (WCHAR) iCH ;
	return	TRUE ;
}

BOOL
CImeDoc::bUnprocessEvent (
	const struct TMSG*			pMsg)
{
	if (pMsg != NULL) {
		m_UnprocessEvent	= *pMsg ;
	} else {
		m_UnprocessEvent.m_nMessage	= WM_NULL ;
	}
	return	TRUE ;
}

BOOL
CImeDoc::bPushEvent (
	const struct TMSG*		pMSG)
{
	if (pMSG == NULL)
		return	FALSE ;
	if (m_nUnreadMessages >= MYARRAYSIZE (m_rMessages)) {
		return	FALSE ;
	}
	m_rMessages [m_nUnreadMessages ++]	= *pMSG ;
	return	TRUE ;
}

int
CImeDoc::nGetSignal () const
{
	return	m_nSignal ;
}

void
CImeDoc::vSetSignal (
	int							nSignal)
{
	m_nSignal	= nSignal ;
	return ;
}

BOOL
CImeDoc::bSignalp () const
{
	return	(m_nSignal != LMSIGNAL_NONE) ;
}

void
CImeDoc::vSetSignalError ()
{
	m_nSignal	= LMSIGNAL_ERROR ;
	return ;
}

BOOL
CImeDoc::bSignalMinibuffp () const
{
	/* minibuffer 𔲂 signal ĂΎ芸 TRUE ԂB*/
	return	(m_nSignal == LMSIGNAL_EXIT_MINIBUFFER || m_nSignal == LMSIGNAL_ABORT_RECURSIVE_EDIT) ;
}

BOOL
CImeDoc::bSignalAbortMinibuffp () const
{
	return	(m_nSignal == LMSIGNAL_ABORT_RECURSIVE_EDIT) ;
}

void
CImeDoc::vClearSignals ()
{
	m_nSignal			= LMSIGNAL_NONE ;
	return ;
}

void
CImeDoc::vJump (
	PLMFUNC					pNextState)
{
	m_pPC					= pNextState ;
	return ;
}

BOOL
CImeDoc::bCall (
	PLMFUNC					pNextState,
	PLMFUNC					pReturnState)
{
	int		nPos ;

	if (m_nRetPC >= MYARRAYSIZE (m_rpRetPC))
		return	FALSE ;
	nPos				= m_nRetPC ++ ;
	m_rpRetPC    [nPos]	= pReturnState ;
	m_rnStackPos [nPos]	= m_nStackHead ;
	m_pPC				= pNextState ;
	return	TRUE ;
}

void*
CImeDoc::pAlloca (
	int						nSize,
	int						nAlign) 
{
	LPBYTE		pbAddress ;
	LPBYTE		pbNextAddress ;
	void*		pvRetval ;
	int			nSkip ;

	if (nSize <= 0 || nAlign <= 0)
		return	NULL ;

	pbAddress			= m_pbStackArea + m_nStackHead ;
#pragma warning (push)
#pragma warning (disable : 4311)
	/*	BYTE*  DWORD ɐ؂l߂Ă̂́AɎvȂ̂ŁcB
	 */
	nSkip				= nAlign - ((DWORD) pbAddress & (nAlign - 1)) ;
#pragma warning (pop)
	pvRetval			= pbAddress + nSkip ;
	pbNextAddress		= pbAddress + nSkip + nSize ;
	if (pbNextAddress > (m_pbStackArea + m_nStackAreaSize)) 
		return	NULL ;
	m_nStackHead	+= nSkip + nSize ;
	return	pvRetval ;
}

void
CImeDoc::vSetRegBool (
	int						nReg,
	BOOL					bValue)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	m_rREGS [nReg].m_nType		= LMREGTYPE_BOOL ;
	m_rREGS [nReg].m_value.m_bool	= bValue ;
	return ;
}

void
CImeDoc::vSetRegString (
	int						nReg,
	LPWSTR					pwString,
	int						nStringLen)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	m_rREGS [nReg].m_nType	= LMREGTYPE_STRING ;
	m_rREGS [nReg].m_value.m_string.m_pointer	= pwString ;
	m_rREGS [nReg].m_value.m_string.m_length	= nStringLen ;
	return ;
}


void
CImeDoc::vSetRegConstString (
	int						nReg,
	LPCWSTR					pwString,
	int						nStringLen)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	m_rREGS [nReg].m_nType	= LMREGTYPE_CONST_STRING ;
	m_rREGS [nReg].m_value.m_cstring.m_pointer	= pwString ;
	m_rREGS [nReg].m_value.m_cstring.m_length	= nStringLen ;
	return ;
}

void
CImeDoc::vSetRegInteger (
	int						nReg,
	int						nValue)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	m_rREGS [nReg].m_nType	= LMREGTYPE_INTEGER ;
	m_rREGS [nReg].m_value.m_integer	= nValue ;
	return ;
}

void
CImeDoc::vSetRegPointer (
	int						nReg,
	void*					pvValue)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	m_rREGS [nReg].m_nType	= LMREGTYPE_POINTER ;
	m_rREGS [nReg].m_value.m_pointer	= pvValue ;
	return ;
}

void
CImeDoc::vSetRegConstPointer (
	int						nReg,
	const void*				pvValue)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	m_rREGS [nReg].m_nType	= LMREGTYPE_CONST_POINTER ;
	m_rREGS [nReg].m_value.m_const_pointer	= pvValue ;
	return ;
}

void
CImeDoc::vSetRegMsg (
	int						nReg,
	const struct TMSG*		pMsg)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	if (pMsg == NULL)
		return ;
	m_rREGS [nReg].m_nType			= LMREGTYPE_MESSAGE ;
	m_rREGS [nReg].m_value.m_msg	= *pMsg ;
	return ;
}

void
CImeDoc::vSetRegNil (
	int						nReg)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	m_rREGS [nReg].m_nType		= LMREGTYPE_NIL ;
	return ;
}

struct CImeDoc::LMREG*
CImeDoc::pGetRegValue (
	int						nReg)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;
	return	m_rREGS + nReg ;
}

BOOL
CImeDoc::bGetRegBool (
	int						nReg,
	BOOL*					pfRetval)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;
	if (m_rREGS [nReg].m_nType == LMREGTYPE_BOOL) {
		if (pfRetval != NULL)
			*pfRetval	= m_rREGS [nReg].m_value.m_bool ;
		return	TRUE ;
	} else {
		return	FALSE ;
	}
}

BOOL
CImeDoc::bGetRegInteger (
	int						nReg,
	int*					piRetval)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;
	if (m_rREGS [nReg].m_nType == LMREGTYPE_INTEGER) {
		if (piRetval != NULL)
			*piRetval	= m_rREGS [nReg].m_value.m_integer ;
		return	TRUE ;
	} else {
		return	FALSE ;
	}
}

BOOL
CImeDoc::bGetRegString (
	int						nReg,
	LPWSTR*					ppwString,
	int*					pnStringLen)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;
	if (m_rREGS [nReg].m_nType == LMREGTYPE_STRING) {
		if (ppwString != NULL)
			*ppwString	= m_rREGS [nReg].m_value.m_string.m_pointer ;
		if (pnStringLen != NULL)
			*pnStringLen	= m_rREGS [nReg].m_value.m_string.m_length ;
		return	TRUE ;
	} else {
		return	FALSE ;
	}
}

BOOL
CImeDoc::bGetRegConstString (
	int						nReg,
	LPCWSTR*				ppwString,
	int*					pnStringLen)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;
	if (m_rREGS [nReg].m_nType == LMREGTYPE_CONST_STRING) {
		if (ppwString != NULL)
			*ppwString	= m_rREGS [nReg].m_value.m_cstring.m_pointer ;
		if (pnStringLen != NULL)
			*pnStringLen	= m_rREGS [nReg].m_value.m_cstring.m_length ;
		return	TRUE ;
	} else {
		return	FALSE ;
	}
}

BOOL
CImeDoc::bGetRegConstStringPair (
	int						nReg,
	LPCWSTR*				ppwLeft,
	int*					pnLeftLen,
	LPCWSTR*				ppwRight,
	int*					pnRightLen)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	if (m_rREGS [nReg].m_nType != LMREGTYPE_CONST_STRINGPAIR) 
		return	FALSE ;
	if (ppwLeft != NULL)
		*ppwLeft	= m_rREGS [nReg].m_value.m_cstringpair.m_left ;
	if (pnLeftLen != NULL)
		*pnLeftLen	= m_rREGS [nReg].m_value.m_cstringpair.m_nleft ;
	if (ppwRight != NULL)
		*ppwRight	= m_rREGS [nReg].m_value.m_cstringpair.m_right ;
	if (pnRightLen != NULL)
		*pnRightLen	= m_rREGS [nReg].m_value.m_cstringpair.m_nright ;
	return	TRUE ;
}

BOOL
CImeDoc::bGetRegPointer (
	int						nReg,
	void**					ppvRetval)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	if (m_rREGS [nReg].m_nType == LMREGTYPE_POINTER) {
		if (ppvRetval != NULL)
			*ppvRetval	= m_rREGS [nReg].m_value.m_pointer ;
		return	TRUE ;
	} else {
		return	FALSE ;
	}
}

BOOL
CImeDoc::bGetRegConstPointer (
	int						nReg,
	const void**			ppvRetval)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	if (m_rREGS [nReg].m_nType == LMREGTYPE_CONST_POINTER) {
		if (ppvRetval != NULL)
			*ppvRetval	= m_rREGS [nReg].m_value.m_const_pointer ;
		return	TRUE ;
	} else {
		return	FALSE ;
	}
}

BOOL
CImeDoc::bGetRegMsg (
	int						nReg,
	struct TMSG*			pMsg)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	if (m_rREGS [nReg].m_nType == LMREGTYPE_MESSAGE) {
		if (pMsg != NULL)
			*pMsg	= m_rREGS [nReg].m_value.m_msg ;
		return	TRUE ;
	} else {
		return	FALSE ;
	}
}

BOOL
CImeDoc::bIsRegNilp (
	int						nReg) const
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	return	(m_rREGS [nReg].m_nType == LMREGTYPE_NIL) ;
}

BOOL
CImeDoc::bPushReg (
	int						nReg)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;

	if (m_nRegStackHead >= NSIZE_REGSTACK)
		return	FALSE ;
	m_pRegStack [m_nRegStackHead ++]	= m_rREGS [nReg] ;
	return	TRUE ;
}

void
CImeDoc::vPopReg (
	int						nReg)
{
	ASSERT (0 <= nReg && nReg < MAX_REGS) ;
	ASSERT (m_nRegStackHead > 0) ;

	m_rREGS [nReg]	= m_pRegStack [-- m_nRegStackHead] ;
	return ;
}

LPWSTR
CImeDoc::pGetReturnBuffer (
	int*					pnRetbufSize)
{
	if (pnRetbufSize != NULL)
		*pnRetbufSize	= MYARRAYSIZE (m_bufRETURN) ;
	return	m_bufRETURN ;
}

/*========================================================================
 *		state machine mainloop
 */
/*
 */
BOOL
CImeDoc::_bTick ()
{
	int		nRetval ;

	while (m_pPC != NULL) {
		nRetval		= (*m_pPC)(this) ;
		switch (nRetval) {
		case	LMR_STOP:
			goto	exit_loop ;
		case	LMR_RETURN:
			if (m_nRetPC <= 0) {
				/* fatal error */
				return	FALSE ;
			}
			-- m_nRetPC ;
			m_nStackHead	= m_rnStackPos	[m_nRetPC] ;
			m_pPC			= m_rpRetPC		[m_nRetPC] ;
			break ;
		case	LMR_ERROR:
			/* fatal error */
			return	FALSE ;
		case	LMR_CONTINUE:
		default:
			break ;
		}
	}
exit_loop:
	return	TRUE ;
}

/*========================================================================
 *	
 */
int
CImeDoc::iGetLastCommand () const
{
	return	m_iLastCommand ; 
}

int
CImeDoc::iGetThisCommand () const
{
	return	m_iThisCommand ;
}

BOOL
CImeDoc::bSetThisCommand (
	int							nFuncNo)
{
	m_iThisCommand	= nFuncNo ;
	return	TRUE ;
}

BOOL
CImeDoc::bSetLastCommand (
	int							nFuncNo)
{
	m_iLastCommand	= nFuncNo ;
	return	TRUE ;
}

int
CImeDoc::iGetLastCommandChar ()
{
	return	m_iLastCommandChar ;
}

BOOL
CImeDoc::bSetLastCommandChar (
	int							nCH)
{
	m_iLastCommandChar		= nCH ;

	/* last-command-event 𓯊cŗǂ̂H */
	m_LastEvent.m_nMessage	= WM_CHAR ;
	m_LastEvent.m_wParam	= nCH ;
	m_LastEvent.m_lParam	= 0 ;
	m_LastEvent.m_pvData	= NULL ;
	m_LastEvent.m_nTime		= GetTickCount () ;
	m_LastEvent.m_pt.x		= 0 ;
	m_LastEvent.m_pt.y		= 0 ;
	return	TRUE ;
}

const struct TMSG*
CImeDoc::pGetLastCommandEvent () const
{
	return	&m_LastEvent ;
}

BOOL
CImeDoc::bSetUnreadCommandChar (
	int							ch)
{
	struct TMSG		msg ;

	msg.m_nMessage	= WM_CHAR ;
	msg.m_wParam	= ch ;
	msg.m_lParam	= 0 ;
	msg.m_pvData	= NULL ;
	msg.m_nTime		= GetTickCount () ;
	msg.m_pt.x		= 0 ;
	msg.m_pt.y		= 0 ;
	return	bPushEvent (&msg) ;
}

BOOL
CImeDoc::bSetUnreadCommandEvent (
	const struct TMSG*			pMsg)
{
	if (pMsg == NULL)
		return	FALSE ;

	return	bPushEvent (pMsg) ;
}

BOOL
CImeDoc::bLookupKeymap (
	const struct TMSG*			pEvent,
	int*						pnFuncNo)
{
	CImeBuffer*	pCurBuffer ;
	int		nFuncNo ;

	ASSERT (pEvent != NULL) ;

	/*	@\𒼐ڎw肳Ă܂ꍇɂ́c
	 */
	if (pEvent->m_nMessage == WM_LM_COMMAND) {
		*pnFuncNo	= (int) pEvent->m_wParam ;
		return	TRUE ;
	}

	pCurBuffer	= pGetCurrentBuffer () ;
	if (pCurBuffer == NULL || pnFuncNo == NULL)
		return	FALSE ;

	nFuncNo		= NFUNC_INVALID_CHAR ;
	/* minor-mode  keymap  check */
	if (nFuncNo == NFUNC_INVALID_CHAR && pCurBuffer->bJModep  ()) 
		_bLookupKeymap (ImeConfig_pGetSkkJModeMap (), pEvent, &nFuncNo) ;
	if (nFuncNo == NFUNC_INVALID_CHAR && pCurBuffer->bLatinModep  ()) 
		_bLookupKeymap (ImeConfig_pGetSkkLatinModeMap (), pEvent, &nFuncNo) ;
	if (nFuncNo == NFUNC_INVALID_CHAR && pCurBuffer->bJisx0208LatinModep  ()) 
		_bLookupKeymap (ImeConfig_pGetSkkJisx0208LatinModeMap (), pEvent, &nFuncNo) ;
	if (nFuncNo == NFUNC_INVALID_CHAR && pCurBuffer->bAbbrevModep  ()) 
		_bLookupKeymap (ImeConfig_pGetSkkAbbrevModeMap (), pEvent, &nFuncNo) ;
	if (nFuncNo == NFUNC_INVALID_CHAR && bRecursiveEditp ()) 
		_bLookupKeymap (ImeConfig_pGetMinibufferMinorModeMap (), pEvent, &nFuncNo) ;
	/* major-mode  keymap  check */
	if (nFuncNo == NFUNC_INVALID_CHAR) 
		_bLookupKeymap (ImeConfig_pGetMajorModeMap (), pEvent, &nFuncNo) ;
	*pnFuncNo	= nFuncNo ;
	return	TRUE ;
}

BOOL
CImeDoc::bWhereIsInternal (
	int							nFuncNo,
	struct TMSG*				pEvent)
{
	CImeBuffer*	pCurBuffer ;

	if (m_pCurBuffer == NULL)
		return	FALSE ;
	pCurBuffer	= m_pCurBuffer ;
	if (pCurBuffer->bJModep  () &&
		_bWhereIsInternal (ImeConfig_pGetSkkJModeMap (), nFuncNo, pEvent))
		return	TRUE ;
	if (pCurBuffer->bLatinModep  () &&
		_bWhereIsInternal (ImeConfig_pGetSkkLatinModeMap (), nFuncNo, pEvent))
		return	TRUE ;
	if (pCurBuffer->bJisx0208LatinModep  () &&
		_bWhereIsInternal (ImeConfig_pGetSkkJisx0208LatinModeMap (), nFuncNo, pEvent))
		return	TRUE ;
	if (pCurBuffer->bAbbrevModep  () &&
		_bWhereIsInternal (ImeConfig_pGetSkkAbbrevModeMap (), nFuncNo, pEvent))
		return	TRUE ;
	if (bRecursiveEditp () &&
		_bWhereIsInternal (ImeConfig_pGetMinibufferMinorModeMap (), nFuncNo, pEvent))
		return	TRUE ;
	if (_bWhereIsInternal (ImeConfig_pGetMajorModeMap (), nFuncNo, pEvent))
		return	TRUE ;
	return	FALSE ;
}

BOOL
CImeDoc::bLookupMajorKeymap (
	const struct TMSG*			pEvent,
	int*						pnFuncNo)
{
	CImeBuffer*	pCurBuffer ;
	int			nFuncNo ;

	/*	@\𒼐ڎw肳Ă܂ꍇɂ́c
	 */
	if (pEvent->m_nMessage == WM_LM_COMMAND) {
		*pnFuncNo	= (int) pEvent->m_wParam ;
		return	TRUE ;
	}

	pCurBuffer	= pGetCurrentBuffer () ;
	if (pCurBuffer == NULL || pnFuncNo == NULL)
		return	FALSE ;

	if (! _bLookupKeymap (ImeConfig_pGetMajorModeMap (), pEvent, &nFuncNo)) {
		*pnFuncNo	= NFUNC_INVALID_CHAR ;
		return	TRUE ;
	}
	*pnFuncNo	= nFuncNo ;
	return	TRUE ;
}

PLMFUNC
CImeDoc::pLookupLMState (
	int						nFuncNo)
{
	PLMFUNC	pPC	= NULL ;

	switch (nFuncNo) {
	case	NFUNC_SELF_INSERT_CHARACTER:
		pPC	= CImeDoc::LM_bSelfInsertCharacter ;
		break ;
	case	NFUNC_BACKWARD_CHAR:
		pPC	= CImeDoc::LM_bBackwardChar ;
		break ;
	case	NFUNC_FORWARD_CHAR:
		pPC	= CImeDoc::LM_bForwardChar ;
		break ;
	case	NFUNC_DELETE_CHAR:
		pPC	= CImeDoc::LM_bDeleteChar ;
		break ;
	case	NFUNC_BACKWARD_DELETE_CHAR:
		pPC	= CImeDoc::LM_bBackwardDeleteChar ;
		break ;
	case	NFUNC_TRANSPOSE_CHARS:
		pPC	= CImeDoc::LM_bTransposeChars ;
		break ;
	case	NFUNC_END_OF_LINE:
		pPC	= CImeDoc::LM_bEndOfLine ;
		break ;
	case	NFUNC_BEGINNING_OF_LINE:
		pPC	= CImeDoc::LM_bBeginningOfLine ;
		break ;
	case	NFUNC_SET_MARK_COMMAND:
		pPC	= CImeDoc::LM_bSetMarkCommand ;
		break ;
	case	NFUNC_KEYBOARD_QUIT:
		pPC	= CImeDoc::LM_bKeyboardQuit ;
		break ;
	case	NFUNC_SKK_INSERT:
		pPC	= CImeDoc::LM_bSkkInsert ;
		break ;
	case	NFUNC_SKK_PREVIOUS_CANDIDATE:
		pPC	= CImeDoc::LM_bSkkPreviousCandidate ;
		break ;
	case	NFUNC_EXIT_RECURSIVE_EDIT:
		pPC	= CImeDoc::LM_bExitRecursiveEdit ;
		break ;
	case	NFUNC_ABORT_RECURSIVE_EDIT:
		pPC	= CImeDoc::LM_bAbortRecursiveEdit ;
		break ;
	case	NFUNC_SKK_MODE:
		pPC	= CImeDoc::LM_bSkkMode ;
		break ;
	case	NFUNC_SKK_KAKUTEI:
		pPC	= CImeDoc::LM_bSkkKakutei ;
		break ;
	case	NFUNC_SKK_START_HENKAN:
		pPC	= CImeDoc::LM_bSkkStartHenkan ;
		break ;
	case	NFUNC_SKK_SET_HENKAN_POINT_SUBR:
		pPC	= CImeDoc::LM_bSkkSetHenkanPointSubr ;
		break ;
	case	NFUNC_SKK_LATIN_MODE:
		pPC	= CImeDoc::LM_bSkkLatinMode ;
		break ;
	case	NFUNC_SKK_JISX0208_LATIN_MODE:
		pPC	= CImeDoc::LM_bSkkJisx0208LatinMode ;
		break ;
	case	NFUNC_SKK_ABBREV_MODE:
		pPC	= CImeDoc::LM_bSkkAbbrevMode ;
		break ;
	case	NFUNC_SKK_DELETE_BACKWARD_CHAR:
		pPC	= CImeDoc::LM_bSkkDeleteBackwardChar ;
		break ;
	case	NFUNC_NEWLINE:
		pPC	= CImeDoc::LM_bNewline ;
		break ;
	case	NFUNC_SKK_INPUT_BY_CODE_OR_MENU:
		pPC	= CImeDoc::LM_bSkkInputByCodeOrMenu ;
		break ;
	case	NFUNC_SKK_TRY_COMPLETION:
		pPC	= CImeDoc::LM_bSkkTryCompletion ;
		break ;
	case	NFUNC_SKK_ABBREV_COMMA:
		pPC	= CImeDoc::LM_bSkkAbbrevComma ;
		break ;
	case	NFUNC_SKK_ABBREV_PERIOD:
		pPC	= CImeDoc::LM_bSkkAbbrevPeriod ;
		break ;
	case	NFUNC_SKK_JISX0208_LATIN_INSERT:
		pPC	= CImeDoc::LM_bSkkJisx0208LatinInsert ;
		break ;
	case	NFUNC_SKK_TOGGLE_CHARACTERS:
		pPC	= CImeDoc::LM_bSkkToggleCharacters ;
		break ;
	case	NFUNC_SKK_TOGGLE_KUTOUTEN:
		pPC	= CImeDoc::LM_bSkkToggleKutouten ;
		break ;
	case	NFUNC_SKK_PURGE_FROM_JISYO:
		pPC	= CImeDoc::LM_bSkkPurgeFromJisyo ;
		break ;

		/*----------------------------------------------------------------
		 *	 interactive function ł͂Ȃ̂ŁAĂяöʒu (message-loopH)
		 *	ɒӂ邱ƁB
		 */
	case	NFUNC_SKK_CURRENT_KUTEN:
		pPC	= CImeDoc::LM_bSkkCurrentKuten ;
		break ;
	case	NFUNC_SKK_CURRENT_TOUTEN:
		pPC	= CImeDoc::LM_bSkkCurrentTouten ;
		break ;
		/*
		 *	܂
		 *----------------------------------------------------------------
		 */
	case	NFUNC_KILL_REGION:
		pPC	= CImeDoc::LM_bKillRegion ;
		break ;
	case	NFUNC_KILL_RING_SAVE:
		pPC	= CImeDoc::LM_bKillRingSave ;
		break ;
	case	NFUNC_YANK:
		pPC	= CImeDoc::LM_bYank ;
		break ;

	case	NFUNC_MOUSE_DRAG_REGION:
		pPC	= CImeDoc::LM_bMouseDragRegion ;
		break ;
	case	NFUNC_MOUSE_DRAG_REGION_END:
		pPC	= CImeDoc::LM_bMouseDragRegionEnd ;
		break ;
	case	NFUNC_MOUSE_PASTE:
		pPC	= CImeDoc::LM_bMousePaste ;
		break ;
	case	NFUNC_MOUSE_CUT:
		pPC	= CImeDoc::LM_bMouseCut ;
		break ;
	case	NFUNC_MOUSE_COPY:
		pPC	= CImeDoc::LM_bMouseCopy ;
		break ;
	case	NFUNC_MOUSE_DELETE:
		pPC	= CImeDoc::LM_bMouseDelete ;
		break ;

	default:
		/* default ̓B*/
		pPC	= CImeDoc::LM_bUnprecessEvent ;
		break ;
	}
	return	pPC ;
}

BOOL
CImeDoc::bIgnoreMessagep (
	const struct TMSG*				pEvent)
{
	if (pEvent == NULL)
		return	TRUE ;

	switch (pEvent->m_nMessage) {
	case	WM_KEYDOWN:
		{
			int		nVkCode			= (int) pEvent->m_wParam ;

			if (nVkCode == VK_SHIFT || nVkCode == VK_CONTROL || nVkCode == VK_RSHIFT || nVkCode == VK_RCONTROL)
				return	TRUE ;
		}
		break ;
	case	WM_IME_KEYDOWN:
		{
			/*
			 *		wParam
			 *			 16 rbgɉꂽ̕R[hA
			 *			 16 rbgɉꂽL[̉zL[R[hĂ܂B
			 *		lParam
			 *			WM_KEYDOWN ɗ^ lParam ƓłB
			 *		lpbKeyState
			 *			݂̃L[{[h̏Ԃ܂256oCg̔zւ̃|C^łB
			 *			IME ͂̓eύXĂ͂܂B
			 */
			//int		nKeyCode		= HIWORD (pEvent->m_wParam) ;
			int		nVkCode			= LOWORD (pEvent->m_wParam) ;


			if (nVkCode == VK_SHIFT || nVkCode == VK_CONTROL || nVkCode == VK_RSHIFT || nVkCode == VK_RCONTROL)
				return	TRUE ;
		}
		break ;
	default:
		break ;
	}
	return	FALSE ;
}

/*	WPARAM  WCHAR 쐬BWCHAR ɕϊ邽߂ WIN32API 
 *	ToAscii 𗘗pĂB
 *	ToAscii ŕϊłȂɑ΂ẮAꕔ VK_XXX E
 *	128 ȏ̃R[h蓖Ă邱ƂƂĂB
 */
int
CImeDoc::iGetCharFromEvent (
	const struct TMSG*				pEvent)
{
	static BYTE	byKeyState [256] = { 0 } ;
	int		iCH ;

	if (pEvent == NULL)
		return	-1 ;

	switch (pEvent->m_nMessage) {
	case	WM_KEYDOWN:
		if (pEvent->m_wParam < 0x100) {
			WORD					rwCH [2] ;

			byKeyState [VK_CONTROL]	= (GetKeyState (VK_CONTROL) & 0x0800)? 0x80 : 0 ;
			byKeyState [VK_SHIFT]	= (GetKeyState (VK_SHIFT)   & 0x0800)? 0x80 : 0 ;
			rwCH [0]	= 0 ;
		
			/*	܂݂͌ keyboard āAϊłȂǂ`FbN
			 *	B
			 */
			if (ToAscii ((UINT)pEvent->m_wParam, 0x80000000, byKeyState, rwCH, 0) == 0) {
				rwCH [0]	= (WCHAR) -1 ;
			} else {
				/*	keyboard ĕϊłꍇ̏B
				 */
				if (rwCH [0] == TEXT(' ') && (byKeyState [VK_CONTROL] & 0x80))
					rwCH [0]	= 0 ;
			}
			return	rwCH [0] ;
		} else {
			return	-1 ;
		}
		break ;
	case	WM_IME_KEYDOWN:
		{
			/*
			 *		wParam
			 *			 16 rbgɉꂽ̕R[hA
			 *			 16 rbgɉꂽL[̉zL[R[hĂ܂B
			 *		lParam
			 *			WM_KEYDOWN ɗ^ lParam ƓłB
			 *		lpbKeyState
			 *			݂̃L[{[h̏Ԃ܂256oCg̔zւ̃|C^łB
			 *			IME ͂̓eύXĂ͂܂B
			 */
			//int		nKeyCode		= HIWORD (pEvent->m_wParam) ;
			int		nVkCode			= LOWORD (pEvent->m_wParam) ;
			const BYTE*	pbKeyState	= reinterpret_cast <const BYTE*> (pEvent->m_pvData) ;
			WORD		rwCH [2] ;

			rwCH [0]	= 0 ;

			/*	where-is-internal pɁA pbKeyState ̎weRs[
			 *	ׂȂ̂낤H
			 */
			if (ToAscii (nVkCode, 0x80000000, pbKeyState, rwCH, 0) == 0) {
				rwCH [0]	= (WCHAR) -1 ;
			} else {
				/*	keyboard ĕϊłꍇ̏B
				 */
				/*	Control-Space  Code 0 ƂĈB
				 */
				if (rwCH [0] == TEXT(' ') && pbKeyState != NULL && (pbKeyState [VK_CONTROL] & 0x80) != 0)
					rwCH [0]	= 0 ;
			}
			return	rwCH [0] ;
		}
		break ;
	case	WM_CHAR:
		return	(UINT)pEvent->m_wParam ;
	default:
		iCH	= -1 ;
		break ;
	}
	return	iCH ;
}

BOOL
CImeDoc::bEventToCharacter (
	const struct TMSG*		pEvent,
	int*					pCH)
{
	int		nCH ;

	if (pEvent == NULL)
		return	FALSE ;

	nCH	= iGetCharFromEvent (pEvent) ;
	if (nCH < 0)
		return	FALSE ;
	if (pCH != NULL)
		*pCH	= nCH ;
	return	TRUE ;
}

BOOL
CImeDoc::bProcessEventp ()
{
	CImeBuffer*	pCurBuffer ;
	int		nText ;

	/*	ĂяõAvP[VɃCxg𓊂ԂׂǂfB
	 *	Ԃׂł͂ȂƔfꍇɂ́ACxg͐HׂĂ܂B
	 *
	 *	Ԃׂۂ̔f́A
	 *	- minibuffer  recursive-edit Ă => no
	 *	- ڃR[h̓[hȂϊI[h => no
	 *	- prefix݂ (ImeBuffer_bHavePrefixp  TRUE Ԃ) => no
	 *	- composition string ݂ (Anext-line ̂悤ɁAm肵Ă event 
	 *	  eɕԂ悤Ȃ݂̂BOɒӂ邱) => no
	 *	- 
	 */
	if (bRecursiveEditp ())
		return	TRUE ;

	pCurBuffer	= pGetCurrentBuffer () ;
	if (pCurBuffer == NULL)
		return	FALSE ;
	if (pCurBuffer->bShowHenkanCandidatesModep  () || pCurBuffer->bInputByCodeOrMenuModep  ())
		return	TRUE ;
	if (pCurBuffer->bHavePrefixp  ())
		return	TRUE ;
	(void) pCurBuffer->pBufferString (&nText) ;
	if (nText > 0)
		return	TRUE ;
	return	FALSE ;
}

/*========================================================================
 *	buffer interfaces
 */
CImeBuffer*
CImeDoc::pCreateBuffer (
	LPCWSTR					wstrMessage,
	int						nstrMessage,
	BOOL					bSkkModeOn)
{
	CImeBuffer*	pBuffer ;
	int			nIndex ;
	BOOL		fInc ;

	/*	ȏċAłȂH */
	if (m_nBuffer >= MAXIMEBUFFER) {
		for (nIndex = 0 ; nIndex < MAXIMEBUFFER ; nIndex ++)
			if (m_rpBuffer [nIndex] == NULL)
				break ;
		if (nIndex >= MAXIMEBUFFER)
			return	NULL ;
		fInc	= FALSE ;
	} else {
		nIndex	= m_nBuffer ;
		fInc	= TRUE ;
	}
	pBuffer	= new CImeBuffer (this) ;
	if (pBuffer == NULL) 
		return	NULL ;
	if (! pBuffer->bInit (wstrMessage, nstrMessage, bSkkModeOn)) {
		delete	pBuffer ;
		return	NULL ;
	}
	m_rpBuffer [nIndex]	= pBuffer ;
	if (fInc)
		m_nBuffer	++ ;
	return	pBuffer ;
}

void
CImeDoc::vDeleteUnusedBuffer ()
{
	int	i ;

	for (i = 0 ; i < m_nUnusedBuffer ; i ++) {
		delete	m_rpUnusedBuffer [i] ;
		m_rpUnusedBuffer [i]	= NULL ;
	}
	m_nUnusedBuffer	= 0 ;
	for (i = m_nBuffer - 1 ; i >= 0 && m_rpBuffer [i] == NULL ; i --) 
		;
	m_nBuffer	= i + 1 ;
	return ;
}

BOOL
CImeDoc::bDeleteBuffer (
	CImeBuffer*		pUnusedBuffer)
{
	int			i, nIndex ;

	nIndex = -1 ;
	for (i = 0 ; i < m_nBuffer ; i ++) {
		if (m_rpBuffer [i] == pUnusedBuffer) {
			nIndex	= i ;
			continue ;
		}
	}
	if (nIndex < 0)
		return	FALSE ;

	m_rpUnusedBuffer [m_nUnusedBuffer ++]	= pUnusedBuffer ;
	m_rpBuffer [nIndex]	= NULL ;
	if (nIndex == (m_nBuffer - 1))
		m_nBuffer	-- ;
	return	TRUE ;
}

/*========================================================================
 *	private methods
 */
BOOL
CImeDoc::_bInitConfig ()
{
	return	TRUE ;
}

void
CImeDoc::_vUninitConfig ()
{
	return ;
}

BOOL
CImeDoc::_bWhereIsInternal (
	const struct CImeKeymap*	pKeymap,
	int							nFuncNo,
	struct TMSG*				pEvent)
{
	const struct CImeKeyBind*	pBind ;
	const struct CImeKeyBind*	pCurBind ;
	int		i, nBind, nCurScore ;
	unsigned int	uKeyMask, uKeyMaskW ;

	for (i = 0 ; i < SIZE_IMEDOC_KEYMAP ; i ++) {
		if (nFuncNo == pKeymap->m_rbyBaseMap [i]) {
			if (pEvent != NULL) {
				pEvent->m_nMessage	= WM_CHAR ;
				pEvent->m_wParam	= (WPARAM) i ;
				pEvent->m_lParam	= 0 ;
				pEvent->m_nTime		= GetTickCount () ;
				pEvent->m_pt.x		= 0 ;
				pEvent->m_pt.y		= 0 ;
			}
			return	TRUE ;
		}
	}

	uKeyMask	= 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_LCONTROL))?	KEYMASK_LCONTROL	: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_RCONTROL))?	KEYMASK_RCONTROL	: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_LMENU))?		KEYMASK_LMENU		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_RMENU))?		KEYMASK_RMENU		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_LSHIFT))?		KEYMASK_LSHIFT		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_RSHIFT))?		KEYMASK_RSHIFT		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_SCROLL))?		KEYMASK_SCROLL		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_NUMLOCK))?		KEYMASK_NUMLOCK		: 0 ;

	uKeyMaskW	= uKeyMask ;
	uKeyMaskW	|= (uKeyMask & (KEYMASK_LCONTROL | KEYMASK_RCONTROL))? (KEYMASK_LCONTROL | KEYMASK_RCONTROL) : 0 ;
	uKeyMaskW	|= (uKeyMask & (KEYMASK_LMENU    | KEYMASK_RMENU))?    (KEYMASK_LMENU    | KEYMASK_RMENU)    : 0 ;
	uKeyMaskW	|= (uKeyMask & (KEYMASK_RSHIFT   | KEYMASK_LSHIFT))?   (KEYMASK_RSHIFT   | KEYMASK_LSHIFT)   : 0 ;

	pBind		= pKeymap->m_pKeyBinds ;
	nBind		= pKeymap->m_nKeyBinds ;
	pCurBind	= NULL ;
	nCurScore	= -1 ;
	while (nBind > 0) {
		if (pBind->m_nKeyFunction == nFuncNo) {
			if ((pBind->m_uKeyMask & uKeyMaskW) == pBind->m_uKeyMask) {
				int		nScore ;
				/* match xɑ΂ score 肷B*/
				nScore	= _iGetKeyMatchScore (pBind->m_uKeyMask, uKeyMaskW, uKeyMask) ;
				if (nScore > nCurScore) {
					pCurBind	= pBind ;
					nCurScore	= nScore ;
					if (nCurScore == 100)
						break ;
				}
			}
		}
		pBind	++ ;
		nBind	-- ;
	}
	if (pCurBind != NULL) {
		if (pEvent != NULL) {
			/*	L[}XNɈˑĂ܂̂ŁAƂȂc */
			pEvent->m_nMessage	= WM_KEYDOWN ;
			pEvent->m_wParam	= (WPARAM) pCurBind->m_nKeyCode ;
			pEvent->m_lParam	= 0 ;
			pEvent->m_nTime		= GetTickCount () ;
			pEvent->m_pt.x		= 0 ;
			pEvent->m_pt.y		= 0 ;
		}
		return	TRUE ;
	}
	return	FALSE ;
}

BOOL
CImeDoc::_bProcessFunctionp (
	int							nFuncNo)
{
	/*	topbuffer ŉ͂ĂȂ filter 邩ۂ̃e[uB
	 *	V`G[VƂāu蓾Ȃv́uǂ瑤ɓ|v̂ǂcB
	 */
	switch (nFuncNo) {
	case	NFUNC_SKK_MODE:
	case	NFUNC_SKK_LATIN_MODE:
	case	NFUNC_SKK_JISX0208_LATIN_MODE:
	case	NFUNC_SKK_SET_HENKAN_POINT_SUBR:
	case	NFUNC_SKK_INPUT_BY_CODE_OR_MENU:
	case	NFUNC_SKK_ABBREV_COMMA:
	case	NFUNC_SKK_ABBREV_PERIOD:
	case	NFUNC_SKK_INSERT:
	case	NFUNC_SKK_KAKUTEI:
	case	NFUNC_SKK_JISX0208_LATIN_INSERT:
	case	NFUNC_SKK_TODAY:
	case	NFUNC_TOGGLE_IME:
	case	NFUNC_SKK_TOGGLE_CHARACTERS:
	case	NFUNC_SKK_TOGGLE_KUTOUTEN:
	case	NFUNC_SKK_PREVIOUS_CANDIDATE:
		return	TRUE ;
	case	NFUNC_INVALID_CHAR:
	default:
		return	FALSE ;
	}

/*	ۗF
	NFUNC_SELF_INSERT_CHARACTER,
	NFUNC_SET_MARK_COMMAND,
	NFUNC_BEGINNING_OF_LINE,
	NFUNC_BACKWARD_CHAR,
	NFUNC_MODE_SPECIFIC_COMMAND_PREFIX,
	NFUNC_DELETE_CHAR,
	NFUNC_END_OF_LINE,
	NFUNC_FORWARD_CHAR,
	NFUNC_KEYBOARD_QUIT,
	NFUNC_BACKWARD_DELETE_CHAR,
	NFUNC_INDENT_FOR_TAB_COMMAND,
	NFUNC_NEWLINE_AND_INDENT,
	NFUNC_KILL_LINE,
	NFUNC_RECENTER,
	NFUNC_NEWLINE,
	NFUNC_NEXT_LINE,
	NFUNC_OPEN_LINE,
	NFUNC_PREVIOUS_LINE,
	NFUNC_QUATED_INSERT,
	NFUNC_ISEARCH_BACKWARD,
	NFUNC_ISEARCH_FORWARD,
	NFUNC_TRANSPOSE_CHARS,
	NFUNC_UNIVERSAL_ARGUMENT,
	NFUNC_SCROLL_UP,
	NFUNC_KILL_REGION,
	NFUNC_CONTROL_X_PREFIX,
	NFUNC_YANK,
	NFUNC_SCROLL_DOWN,
	NFUNC_PREFIX_COMMAND,
	NFUNC_EXIT_RECURSIVE_EDIT,
	NFUNC_ABORT_RECURSIVE_EDIT,
	NFUNC_UNDO,
	NFUNC_SKK_START_HENKAN,
	NFUNC_SKK_TRY_COMPLETION,
	NFUNC_SKK_DELETE_BACKWARD_CHAR,
	NFUNC_SKK_PREVIOUS_CANDIDATE,
	NFUNC_SKK_TOGGLE_CHARACTERS,
	NFUNC_SKK_PURGE_FROM_JISYO,
	NFUNC_SKK_ABBREV_MODE,
	NFUNC_SKK_DISPLAY_CODE_FOR_CHAR_AT_POINT,
	NFUNC_SKK_CURRENT_KUTEN,
	NFUNC_SKK_UNDO,
	NFUNC_SKK_COMP_DO,
	NFUNC_SKK_KAKUTEI_HENKAN,
	NFUNC_SKK_UNDO_KAKUTEI_HENKAN,
	NFUNC_SKK_LATIN_HENKAN,
	*/
}

BOOL
CImeDoc::_bLookupKeymap (
	const struct CImeKeymap*	pKeymap,
	const struct TMSG*			pEvent,
	int*						pnFuncNo)
{
	const struct CImeKeyBind*	pBind ;
	int		nFuncNo, iCH, nBind, nKeyCode, nCurScore ;
	unsigned int	uKeyMask, uKeyMaskW ;

	if (pKeymap == NULL || pEvent == NULL)
		return	FALSE ;

	iCH	= iGetCharFromEvent (pEvent) ; 
	if (0 <= iCH && iCH < SIZE_IMEDOC_KEYMAP) {
		if (pnFuncNo != NULL)
			*pnFuncNo	= pKeymap->m_rbyBaseMap [iCH] ;
		return	TRUE ;
	}
	switch (pEvent->m_nMessage) {
	case	WM_KEYDOWN:
		nKeyCode	= (short) pEvent->m_wParam ;
		break ;
	case	WM_LBUTTONDOWN:
		nKeyCode	= VK_LBUTTON ;
		break ;
	case	WM_RBUTTONDOWN:
		nKeyCode	= VK_RBUTTON ;
		break ;
	default:
		return	FALSE ;
	}

	uKeyMask	= 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_LCONTROL))?	KEYMASK_LCONTROL	: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_RCONTROL))?	KEYMASK_RCONTROL	: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_LMENU))?		KEYMASK_LMENU		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_RMENU))?		KEYMASK_RMENU		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_LSHIFT))?		KEYMASK_LSHIFT		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_RSHIFT))?		KEYMASK_RSHIFT		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_SCROLL))?		KEYMASK_SCROLL		: 0 ;
	uKeyMask	|= (GetAsyncKeyState (VK_NUMLOCK))?		KEYMASK_NUMLOCK		: 0 ;

	uKeyMaskW	= uKeyMask ;
	uKeyMaskW	|= (uKeyMask & (KEYMASK_LCONTROL | KEYMASK_RCONTROL))? (KEYMASK_LCONTROL | KEYMASK_RCONTROL) : 0 ;
	uKeyMaskW	|= (uKeyMask & (KEYMASK_LMENU    | KEYMASK_RMENU))?    (KEYMASK_LMENU    | KEYMASK_RMENU)    : 0 ;
	uKeyMaskW	|= (uKeyMask & (KEYMASK_RSHIFT   | KEYMASK_LSHIFT))?   (KEYMASK_RSHIFT   | KEYMASK_LSHIFT)   : 0 ;

	/*	eXg̏Ԃ͎̒ʂF
	 *	1. keymask ̊SvA
	 *	2. keymask  L  R 𖳎rł̈vA
	 *	3. keymask ̕vA
	 *	1->2->3 ̏ŃeXgāAtΏIB
	 *	tȂ΁AŃAEgB
	 *
	 *	 L  R 𖳎rACONTROL, MENU, SHIFT ȂǕ
	 *	Ăꍇɂ́A
	 */
	pBind	= pKeymap->m_pKeyBinds ;
	nBind	= pKeymap->m_nKeyBinds ;
	nFuncNo	= NFUNC_INVALID_CHAR ;
	nCurScore	= -1 ;
	while (nBind > 0) {
		if (pBind->m_nKeyCode == nKeyCode) {
			if ((pBind->m_uKeyMask & uKeyMaskW) == pBind->m_uKeyMask) {
				int		nScore ;

				/* match xɑ΂ score 肷B*/
				nScore	= _iGetKeyMatchScore (pBind->m_uKeyMask, uKeyMaskW, uKeyMask) ;
				if (nScore > nCurScore) {
					nFuncNo		= pBind->m_nKeyFunction ;
					nCurScore	= nScore ;

					if (nCurScore == 100)
						break ;
				}
			}
		}
		pBind	++ ;
		nBind	-- ;
	}
	if (nCurScore < 0)
		return	FALSE ;
	if (pnFuncNo != NULL)
		*pnFuncNo	= nFuncNo ;
	return	TRUE ;
}

int
CImeDoc::_iGetKeyMatchScore (
	unsigned int				uMask,
	unsigned int				uExactMask,
	unsigned int				uRoughMask)
{
	unsigned int	uValue ;
	int				nScore, i ;

	if (uMask == uRoughMask)
		return	100 ;	/* 100% */

	/*	ۂɉꂽL[牟ĂȂƍ邱Ƃ͂Ă̋t͂ȂB
	 *	(ɉƂɂ͂Ȃ)
	 *	܂AL[̉߂A
	 *		SHIFT + F.1 => F.1 
	 *	͂̕Ă
	 *		F.1 => SHIFT + F.1
	 *	ɂ͊g債ȂB
	 */
	if ((uRoughMask & uMask) != uMask)
		return	0 ;

	/*	vĂ鐔XRAȂBALR ̈Ⴂňvꍇ
	 *	SvXRAႢB
	 */
	uValue	= uExactMask & uMask ;
	nScore	= 0 ;
	for (i = 0 ; i < NBITS_KEYMASK ; i ++)
		if (uValue & (1 << i))
			nScore	++ ;
	uValue	= uRoughMask & uMask ;
	for (i = 0 ; i < NBITS_KEYMASK ; i ++)
		if (uValue & (1 << i))
			nScore	++ ;
	return	nScore ;
}

BOOL
CImeDoc::_bUpdateCandidateInfoForStatusText ()
{
	IMECANDIDATES*	pMyCand ;

	pMyCand				= &m_MessageInfo ;
	if (m_rpBuffer [0] != m_pCurBuffer) {
		LPWSTR			pwDest ;
		LPWSTR			pCandidateStr ;
		int				nText ;
		LPCWSTR			pwText ;
		int				iLength, iCursor ;
		CTMarker*		pmkPoint	= NULL ;
		UINT*			pPageIndex ;
		UINT*			pCandidateOffset ;

		/*	Message ꍇɂ͐擪ɁuIvǉcB
		 *	ƁAO StatusLine `łȂꍇɁAJ[\Ӗ
		 *	ĂB``|'' ŗǂB
		 */
		pwText	= m_pCurBuffer->pBufferRawString (&iLength) ;
		if (m_pCurBuffer->bGetMarker (CImeBuffer::MARKER_POINT, &pmkPoint) && pmkPoint != NULL) {
			iCursor	= pmkPoint->iGetPosition () ;
		} else {
			iCursor	= -1 ;
		}
		nText	= iLength + 1/*``|''̕*/ + 1/*nul̕*/ ;  

		pMyCand->m_vbufCandidate.vClear () ;
		pMyCand->m_vbufCandidateIndex.vClear () ;
		pMyCand->m_vbufPageIndex.vClear () ;
		pMyCand->m_vbufAnnotationIndex.vClear () ;
		if (! pMyCand->m_vbufCandidate.bRequire (nText) ||
			! pMyCand->m_vbufCandidateIndex.bRequire (1) ||
			! pMyCand->m_vbufPageIndex.bRequire (1))
			return	FALSE ;

		pMyCand->m_iStyle		= IMECANDSTYLE_MINIBUFFERTEXT ;
		pMyCand->m_iCount		= 1 ;
		pMyCand->m_iSelection	= 0 ;
		pMyCand->m_iCurrentPage	= 0 ;
		pCandidateOffset		= pMyCand->m_vbufCandidateIndex.pGetBuffer () ;
		pCandidateOffset [0]	= 0 ;
		pPageIndex				= pMyCand->m_vbufPageIndex.pGetBuffer () ;
		pPageIndex [0]			= 0 ;
		pCandidateStr			= pMyCand->m_vbufCandidate.pGetBuffer () ;

		pwDest		= pCandidateStr ;
		if (0 < iCursor) {
			memcpy (pwDest, pwText, iCursor * sizeof (WCHAR)) ;
			pwDest	+= iCursor ;
		}
		*pwDest ++	= L'|' ;
		if (iCursor < iLength) {
			memcpy (pwDest, pwText + iCursor, (iLength - iCursor) * sizeof (WCHAR)) ;
			pwDest	+= (iLength - iCursor) ;
		}
		*pwDest = L'\0' ;
	} else {
		pMyCand->vClear () ;
	}
	return	TRUE ;
}



