/*	config.cpp
 *
 *	ITfFnConfigure implementation
 */
#include "globals.h"
#include <shlwapi.h>
#include "skimic.h"
#include "skkimmgr.h"
#include "resource.h"
#include "IME\TJisyoUpdateSession.h"
#include "IME\ImeConfig.h"
#include "IME\RuleTreeNode.h"

#define	SKKWORD_OKURIARI	(IME_REGWORD_STYLE_USER_FIRST+0)
#define SKKWORD_OKURINASI	(IME_REGWORD_STYLE_USER_FIRST+1)
#define	MAX_KEYWORD			(256)

static	BOOL	bWalkRuleTreeNode (HWND, const CSkkRuleTreeNode*) ;

class	CRegisterWordContainer {
private:
	WCHAR		_wszKey  [MAX_KEYWORD] ;
	WCHAR		_wszWord [MAX_KEYWORD] ;
	WCHAR		_chOkurigana ;
	BOOL		_fOkuri ;

public:
	CRegisterWordContainer () {
		_wszKey [0]		= L'\0' ;
		_wszWord [0]	= L'\0' ;
		_chOkurigana	= L'\0' ;
		_fOkuri			= FALSE ;
	}

	CRegisterWordContainer (LPCWSTR wstrKey) {
		if (wstrKey != NULL) {
			lstrcpynW (_wszKey, wstrKey, MAX_KEYWORD) ;
		} else {
			_wszKey [0]	= L'\0' ;
		}
		_wszWord [0]	= L'\0' ;
		_chOkurigana	= L'\0' ;
		_fOkuri			= FALSE ;
		return ;
	}

	LPCWSTR		GetKeyword () const {
		return	_wszKey ;
	}
	LPCWSTR		GetWord () const {
		return	_wszWord ;
	}
	BOOL		GetOkuri () const {
		return	_fOkuri ;
	}
	BOOL		SetKeyword (LPCWSTR wstrKey) {
		lstrcpynW (_wszKey, wstrKey, MAX_KEYWORD) ;
		return	TRUE ;
	}
	BOOL		SetWord (LPCWSTR wstrWord) {
		lstrcpynW (_wszWord, wstrWord, MAX_KEYWORD) ;
		return	TRUE ;
	}
	BOOL		SetOkuri (BOOL fOkuri) {
		_fOkuri	= fOkuri ;
		return	TRUE ;
	}
	BOOL		SetOkurigana (WCHAR ch) {
		_chOkurigana	= ch ;
	}
	WCHAR		GetOkurigana () const {
		return	_chOkurigana ;
	}
} ;

HRESULT
CSkkImeTextService::Show (
	HWND		hwndParent,
	LANGID		langid,
	REFGUID		rguidProfile,
	BSTR		bstrRegistered)
{
	CRegisterWordContainer	container (bstrRegistered) ;
	int		nRetval ;

	DEBUGPRINTF ((TEXT ("CSkkImeTextService::Show ()/ITFnConfigureRegisterWord\n"))) ;

	if (langid != SKKIME_LANGID)
		return	E_FAIL ;

	nRetval		= DialogBoxParam (
		g_hInst,
		MAKEINTRESOURCE (IDD_DIALOG_REGISTER_WORD),
		hwndParent,
		&_DialogProc_RegisterWord,
		(LPARAM) &container) ;
	if (nRetval == IDOK) {
		CTRecordSession*	pSession		= new CTRecordSession ;
		LPCWSTR				wstrHenkanKey	= container.GetKeyword () ;
		LPCWSTR				wstrResult		= container.GetWord () ;
		WCHAR				chOkurigana		= container.GetOkurigana () ;
		BOOL				bOkurigana		= container.GetOkuri ()? TRUE : FALSE ;

		if (pSession != NULL) {
			/*	艼̐ݒ肪͓KȂ̂ŃRpCʂ悤ɂȂ猩(Sat Mar 31 20:40:53 2007)
			 *---
			 *	艼ɂ́u/r(r->)vȂuvBđ҂āc͕svȂ̂H
			 */
			pSession->bUpdate (wstrHenkanKey, lstrlenW (wstrHenkanKey), wstrResult, lstrlenW (wstrResult), &chOkurigana, 1, bOkurigana) ;
			delete	pSession ;
		}
	} else if (nRetval == -1) {
		DWORD	dwError	= GetLastError () ;
		DEBUGPRINTF ((TEXT ("DialogBoxParam () failed (0x%lx)\n"), dwError)) ;
	}
	return	S_OK ;
	UNREFERENCED_PARAMETER (rguidProfile) ;
}

INT_PTR	CALLBACK
CSkkImeTextService::_DialogProc_RegisterWord (
	HWND		hwndDlg,
    UINT		uMsg,
    WPARAM		wParam,
    LPARAM		lParam)
{
	CRegisterWordContainer*	pContainer	= NULL ;
	HWND					hwndItem ;

	switch (uMsg) {
	case	WM_INITDIALOG:
	{
		pContainer	= (CRegisterWordContainer*) lParam ;

		SetWindowLongPtr (hwndDlg, DWLP_USER, (LONG_PTR) pContainer) ;
		if (pContainer != NULL) {
			SetDlgItemTextW (hwndDlg, IDC_EDIT_REGISTER_KEY, pContainer->GetKeyword ()) ;
			SetDlgItemTextW (hwndDlg, IDC_EDIT_REGISTER_WORD, pContainer->GetWord ()) ;
			CheckRadioButton (hwndDlg, IDC_RADIO_OKURI_ARI, IDC_RADIO_OKURI_NASHI, IDC_RADIO_OKURI_NASHI) ;
		}
		hwndItem	= GetDlgItem (hwndDlg, IDC_COMBO_OKURI_MOJI) ;
		if (hwndItem != NULL) {
			const CSkkRuleTreeNode*	pRoot	= ImeConfig_pGetSkkRuleTree () ;

			/*	``/a'' ̂悤Ȍ`ɂȂB邪cB
			 */
			(void) bWalkRuleTreeNode (hwndItem, ImeConfig_pGetSkkRuleTree ()) ;
			EnableWindow (hwndItem, FALSE) ;
		}
		return	TRUE ;
	}
	case WM_COMMAND: 
	{
		WCHAR					wszBuffer [MAX_KEYWORD] ;

		pContainer	= (CRegisterWordContainer *)GetWindowLongPtr (hwndDlg, DWLP_USER) ;
		switch (LOWORD (wParam)) { 
		case	IDC_RADIO_OKURI_ARI:
		case	IDC_RADIO_OKURI_NASHI:
			if (HIWORD (wParam) == BN_CLICKED) {
				CheckRadioButton (hwndDlg, IDC_RADIO_OKURI_ARI, IDC_RADIO_OKURI_NASHI, LOWORD (wParam)) ;
				hwndItem	= GetDlgItem (hwndDlg, IDC_COMBO_OKURI_MOJI) ;
				if (hwndItem != NULL) 
					EnableWindow (hwndItem, (LOWORD (wParam) == IDC_RADIO_OKURI_ARI)) ;
				return	TRUE ; 
			}
			break ;
			
		case	IDOK: 
		{
			BOOL	fOkuri	= FALSE ;
			int		n, ch ;

			/*	ϊ̓o^ǂ`FbNB
			 */
			if (IsDlgButtonChecked (hwndDlg, IDC_RADIO_OKURI_ARI)) {
				int		nIndex ;

				hwndItem	= GetDlgItem (hwndDlg, IDC_COMBO_OKURI_MOJI) ;
				if (hwndItem == NULL)
					break ;

				nIndex	= SendMessage (hwndItem, CB_GETCURSEL, 0, 0) ;
				if (nIndex == CB_ERR) {
					MessageBox (hwndDlg, TEXT ("蕶w肵ĉ"), TEXT ("x"), MB_OK) ;
					break ;
				}
				if (SendMessageW (hwndItem, CB_GETLBTEXT, (WPARAM) nIndex, (LPARAM) buf) == CB_ERR) {
					MessageBox (hwndDlg, TEXT ("蕶K؂ł͂܂"), TEXT ("x"), MB_OK) ;
					break ;
				}
				chOkuri			= buf [0] ;	/* Okuri ƂĒǉB*/
				chOkuriAlpha	= buf [2] ;	/* HenkanKey ̍ŌɒǉB*/
				if (buf [1] != L'/' || chOkuriAlpha < L'a' || chOkuriAlpha > L'z') {
					MessageBox (hwndDlg, TEXT ("蕶K؂ł͂܂"), TEXT ("x"), MB_OK) ;
					break ;
				}
				fOkuri	= TRUE ;
			}
			/*	o^镶𓾂B
			 */
			n	= GetDlgItemTextW (hwndDlg, IDC_EDIT_REGISTER_KEY, wszBuffer, MAX_KEYWORD - 2) ;
			if (n <= 0) {
				MessageBox (hwndDlg, TEXT ("o^P(ϊO)ł"), TEXT ("x"), MB_OK) ;
				break ;
			}
			/*	ϊ̓o^Ȃ瑗蕶ǉB
			 */
			if (fOkuri) {
				wszBuffer [n ++]	= (WCHAR)chOkuriAlpha ;
				pContainer->SetOkurigana (chOkuri) ;
			}

			wszBuffer [n]	= L'\0' ;
			if (wcspbrk (wszBuffer, L" \t\n\r") != NULL) {
				MessageBox (hwndDlg, TEXT ("o^P(ϊO)ɎgpłȂ(󔒁A^u)܂܂Ă܂"), TEXT ("x"), MB_OK) ;
				break ;
			}
			pContainer->SetOkuri (fOkuri) ;

			/*	ϊʂǉB
			 */
			if (pContainer != NULL)
				pContainer->SetKeyword (wszBuffer) ;
			n	= GetDlgItemTextW (hwndDlg, IDC_EDIT_REGISTER_WORD, wszBuffer, MAX_KEYWORD - 1) ;
			if (n <= 0) {
				MessageBox (hwndDlg, TEXT ("ϊʂł"), TEXT ("x"), MB_OK) ;
				break ;
			}
			wszBuffer [n]	= L'\0' ;
			if (pContainer != NULL)
				pContainer->SetWord (wszBuffer) ;

			// Fall through. 
		}
		case	IDCANCEL: 
			EndDialog (hwndDlg, wParam) ; 
			return	TRUE ; 
		} 
    } 
	default:
		break ;
	}
	return	FALSE ;
}

/*
pRoot	= ImeConfig_pGetSkkRuleTree () ;
*/
BOOL
bWalkRuleTreeNode (
	HWND					hDlg,
	const CSkkRuleTreeNode*	pRoot)
{
	const CSkkRuleTreeNode*	pNode ;

	pNode	= pRoot ;
	while (pNode != NULL) {
		const CSkkRuleTreeNodeOutput*	pOutput ;

		pOutput	= pNode->pGetOutput () ;
		if (pOutput != NULL && pOutput->iGetType () == CSkkRuleTreeNodeOutput::NTYPE_STRINGPAIR) {
			LPCWSTR	wstrPrefix,	wstrHira ;
			int		nPrefixLen,	nHiraLen ;

			wstrPrefix	= CSkkRuleTreeNode::pGetPrefix (pNode, &nPrefixLen) ;
			wstrHira	= pOutput->pGetHiragana (&nHiraLen) ;
			if (nPrefixLen > 0 && wstrPrefix != NULL && nHiraLen > 0 && wstrHira != NULL) {
				WCHAR	buf [256] ;
				int		nText, nIndex ;

				nText	= wnsprintfW (buf, MYARRAYSIZE (buf) - 1, L"%c/%c", wstrHira [0], wstrPrefix [0]) ;
				buf [nText]	= L'\0' ;
				nIndex	= SendMessageW (hDlg, CB_FINDSTRINGEXACT, (WPARAM) 0, (LPARAM) buf) ;
				if (nIndex == CB_ERR) {
					/* ݂ȂΒǉB*/
					SendMessageW (hDlg, CB_INSERTSTRING, (WPARAM)-1, (LPARAM) buf) ;
				}
			}
		}
		pChild		= pRoot->pGetChild () ;
		if (pChild != NULL) {
			if (! bWalkRuleTreeNode (hDlg, pChild))
				return	FALSE ;
		}
		pNode	= pNode->pGetBrother () ;
	}
	return	TRUE ;
}


