/* # SKKIME (Simple Kana-Kanji Input Method Editor for Windows'98)
 * Config.c
 * This file is part of SKKIME.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "commctrl.h"
#include "skki1_0.h"
#include <limits.h>
#include "conf\skimcnfp.h"
#include "resource.h"

#define	REGPATH_INPUTVECTOR			MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\InputVector")
#define	REGPATH_ZENKAKUVECTOR		MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\ZenkakuVector")
#define	MAX_LENGTH_INPUTVECTOR			(128)

typedef struct {
	int				m_nKeyCode ;
	LPMYSTR			m_strInputVector ;
}	TDEFINPUTVECTOR ;

typedef struct {
	int				m_nCurSel ;
	LPMYSTR*		m_pInputVector ;
	MYCHAR			m_strValue [MAX_LENGTH_INPUTVECTOR] ;		/* 128 ȏ͑ʖځB*/
}	TDlgEditInputVectorArg ;

/*	Prototype 
 */
static	INT_PTR	skkimeConfig_dlgInputVectorOnInitDialog (HWND, WPARAM, LPARAM, int) ;
static	INT_PTR	skkimeConfig_dlgInputVectorOnCommand    (HWND, WPARAM, LPARAM, int) ;
static	INT_PTR	skkimeConfig_dlgInputVectorOnNotify     (HWND, WPARAM, LPARAM, int) ;
static	BOOL				skkimeConfig_doEditInputVector	(HWND, int, int, int, LPMYSTR*) ;
static	INT_PTR CALLBACK	skkimeConfig_dlgEditInputVector	(HWND, UINT, WPARAM, LPARAM) ;
static	int CALLBACK		skkimeConfig_compareInputVector (LPARAM, LPARAM, LPARAM) ;
static	BOOL	skkimeConfig_initInputVector			(TINYMEMMGR*, LPCMYSTR, LPMYSTR*, TDEFINPUTVECTOR*, int) ;
static	BOOL	skkimeConfig_updateInputVector			(LPCMYSTR, LPMYSTR*) ;

/*	static global variables
 */
static	TDEFINPUTVECTOR		srSkkinputDefaultInputVector []	= {
	{ '!',	MYTEXT ("I"), },	{ ',',	MYTEXT ("A"), },
	{ '-',	MYTEXT ("["), },	{ '.',	MYTEXT ("B"), },
	{ ':',	MYTEXT ("F"), },	{ ';',	MYTEXT ("G"), },
	{ '?',	MYTEXT ("H"), },	{ '[',	MYTEXT ("u"), },
	{ ']',	MYTEXT ("v"), },
} ;

static	TDEFINPUTVECTOR		srSkkinputDefaultZenkakuVector []	= {
	{ 0x20,	MYTEXT ("@"), },	{ 0x21,	MYTEXT ("I"), },
	{ 0x22,	MYTEXT ("h"), },	{ 0x23,	MYTEXT (""), },
	{ 0x24,	MYTEXT (""), },	{ 0x25,	MYTEXT (""), },
	{ 0x26,	MYTEXT (""), },	{ 0x27,	MYTEXT ("f"), },
	{ 0x28,	MYTEXT ("i"), },	{ 0x29,	MYTEXT ("j"), },
	{ 0x2A,	MYTEXT (""), },	{ 0x2B,	MYTEXT ("{"), },
	{ 0x2C,	MYTEXT ("C"), },	{ 0x2D,	MYTEXT ("|"), },
	{ 0x2E,	MYTEXT ("D"), },	{ 0x2F,	MYTEXT ("^"), },
	{ 0x30,	MYTEXT ("O"), },	{ 0x31,	MYTEXT ("P"), },
	{ 0x32,	MYTEXT ("Q"), },	{ 0x33,	MYTEXT ("R"), },
	{ 0x34, MYTEXT ("S"), },	{ 0x35,	MYTEXT ("T"), },
	{ 0x36,	MYTEXT ("U"), },	{ 0x37,	MYTEXT ("V"), }, 
	{ 0x38, MYTEXT ("W"), },	{ 0x39,	MYTEXT ("X"), },
	{ 0x3A,	MYTEXT ("F"), },	{ 0x3B,	MYTEXT ("G"), },
	{ 0x3C, MYTEXT (""), },	{ 0x3D,	MYTEXT (""), },
	{ 0x3E,	MYTEXT (""), },	{ 0x3F,	MYTEXT ("H"), },
	{ 0x40, MYTEXT (""), },	{ 0x41,	MYTEXT ("`"), },
	{ 0x42,	MYTEXT ("a"), },	{ 0x43,	MYTEXT ("b"), },
	{ 0x44, MYTEXT ("c"), },	{ 0x45,	MYTEXT ("d"), },
	{ 0x46,	MYTEXT ("e"), },	{ 0x47,	MYTEXT ("f"), },
	{ 0x48, MYTEXT ("g"), },	{ 0x49,	MYTEXT ("h"), },
	{ 0x4A,	MYTEXT ("i"), },	{ 0x4B,	MYTEXT ("j"), },
	{ 0x4C, MYTEXT ("k"), },	{ 0x4D,	MYTEXT ("l"), },
	{ 0x4E,	MYTEXT ("m"), },	{ 0x4F,	MYTEXT ("n"), }, 
	{ 0x50, MYTEXT ("o"), },	{ 0x51,	MYTEXT ("p"), },
	{ 0x52,	MYTEXT ("q"), },	{ 0x53,	MYTEXT ("r"), },
	{ 0x54, MYTEXT ("s"), },	{ 0x55,	MYTEXT ("t"), },
	{ 0x56,	MYTEXT ("u"), },	{ 0x57,	MYTEXT ("v"), },  
	{ 0x58, MYTEXT ("w"), },	{ 0x59,	MYTEXT ("x"), },
	{ 0x5A,	MYTEXT ("y"), },	{ 0x5B,	MYTEXT ("m"), },
	{ 0x5C, MYTEXT ("_"), },	{ 0x5D,	MYTEXT ("n"), },
	{ 0x5E,	MYTEXT ("O"), },	{ 0x5F,	MYTEXT ("Q"), },  
	{ 0x60, MYTEXT ("e"), },	{ 0x61,	MYTEXT (""), },
	{ 0x62,	MYTEXT (""), },	{ 0x63,	MYTEXT (""), },
	{ 0x64, MYTEXT (""), },	{ 0x65,	MYTEXT (""), },
	{ 0x66,	MYTEXT (""), },	{ 0x67,	MYTEXT (""), },  
	{ 0x68, MYTEXT (""), },	{ 0x69,	MYTEXT (""), },
	{ 0x6A,	MYTEXT (""), },	{ 0x6B,	MYTEXT (""), },
	{ 0x6C, MYTEXT (""), },	{ 0x6D,	MYTEXT (""), },
	{ 0x6E,	MYTEXT (""), },	{ 0x6F,	MYTEXT (""), },  
	{ 0x70, MYTEXT (""), },	{ 0x71,	MYTEXT (""), },
	{ 0x72,	MYTEXT (""), },	{ 0x73,	MYTEXT (""), },
	{ 0x74, MYTEXT (""), },	{ 0x75,	MYTEXT (""), },
	{ 0x76,	MYTEXT (""), },	{ 0x77,	MYTEXT (""), },  
	{ 0x78, MYTEXT (""), },	{ 0x79,	MYTEXT (""), },
	{ 0x7A,	MYTEXT (""), },	{ 0x7B,	MYTEXT ("o"), },
	{ 0x7C, MYTEXT ("b"), },	{ 0x7D,	MYTEXT ("p"), },
	{ 0x7E,	MYTEXT ("`"), },
} ;

/*========================================================================
 *	public functions
 */
BOOL
skkimeConfig_InitializeInputVector (
	register TINYMEMMGR*	pMemMgr,
	register LPMYSTR*		ppInputVector)
{
	if (pMemMgr == NULL || ppInputVector == NULL)
		return	FALSE ;

	return	skkimeConfig_initInputVector (pMemMgr, REGPATH_INPUTVECTOR, ppInputVector, srSkkinputDefaultInputVector, ARRAYSIZE (srSkkinputDefaultInputVector)) ;
}

BOOL
skkimeConfig_InitializeZenkakuVector (
	register TINYMEMMGR*	pMemMgr,
	register LPMYSTR*		ppZenkakuVector)
{
	if (pMemMgr == NULL || ppZenkakuVector == NULL)
		return	FALSE ;

	return	skkimeConfig_initInputVector (pMemMgr, REGPATH_ZENKAKUVECTOR, ppZenkakuVector, srSkkinputDefaultZenkakuVector, ARRAYSIZE (srSkkinputDefaultZenkakuVector)) ;
}

BOOL
skkimeConfig_UpdateInputVector (
	register LPMYSTR*		ppInputVector)
{
	if (ppInputVector == NULL)
		return	FALSE ;
	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateInputVector (%p)\n"), ppInputVector)) ;
	return	skkimeConfig_updateInputVector (REGPATH_INPUTVECTOR, ppInputVector) ;
}

BOOL
skkimeConfig_UpdateZenkakuVector (
	register LPMYSTR*		ppZenkakuVector)
{
	if (ppZenkakuVector == NULL)
		return	FALSE ;
	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateZenkakuVector (%p)\n"), ppZenkakuVector)) ;
	return	skkimeConfig_updateInputVector (REGPATH_ZENKAKUVECTOR, ppZenkakuVector) ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgInputVectorProc (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	DEBUGPRINTF ((MYTEXT ("skkimeConfig_dlgInputVectorProc (%x, %d, %x, %x)\n"),
				  hDlg, uMsg, wParam, lParam)) ;
	switch (uMsg) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgInputVectorOnInitDialog (hDlg, wParam, lParam, IDC_LIST_INPUTVECTOR) ;

	case	WM_COMMAND:
		return	skkimeConfig_dlgInputVectorOnCommand (hDlg, wParam, lParam, IDC_LIST_INPUTVECTOR) ;

	case	WM_NOTIFY:
	{
		register LPPROPSHEETPAGE	pPropSheet ;
		register SkkimeConfigArg*	pConfArg ;
		register NMHDR	FAR*	pNMHDR	= (NMHDR *) lParam ;

		if (pNMHDR->idFrom == IDC_LIST_INPUTVECTOR)
			return	skkimeConfig_dlgInputVectorOnNotify (hDlg, wParam, lParam, IDC_LIST_INPUTVECTOR) ;

		pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
		pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

		switch (pNMHDR->code){
		case PSN_SETACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
			return	TRUE ;

		case PSN_KILLACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, FALSE) ;
			return	TRUE ;

		case PSN_APPLY:
		{
			if (skkimeConfig_UpdateInputVector (pConfArg->m_rInputVector)) {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
			} else {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
			}
			return	TRUE ;
		}
		case PSN_RESET:
		case PSN_HELP:
			break ;

		default:
			return FALSE ;
		}
		return	TRUE ;
	}
	default:
		break ;
	}
	return	FALSE ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgZenkakuVectorProc (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	DEBUGPRINTF ((MYTEXT ("skkimeConfig_dlgZenkakuVectorProc (%x, %d, %x, %x)\n"),
				 hDlg, uMsg, wParam, lParam)) ;
	switch (uMsg) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgInputVectorOnInitDialog (hDlg, wParam, lParam, IDC_LIST_ZENKAKUVECTOR) ;

	case	WM_COMMAND:
		return	skkimeConfig_dlgInputVectorOnCommand (hDlg, wParam, lParam, IDC_LIST_ZENKAKUVECTOR) ;

	case	WM_NOTIFY:
	{
		register NMHDR	FAR*	pNMHDR	= (NMHDR *) lParam ;
		register LPPROPSHEETPAGE	pPropSheet ;
		register SkkimeConfigArg*	pConfArg ;

		if (pNMHDR->idFrom == IDC_LIST_ZENKAKUVECTOR)
			return	skkimeConfig_dlgInputVectorOnNotify (hDlg, wParam, lParam, IDC_LIST_ZENKAKUVECTOR) ;

		pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
		pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

		switch (pNMHDR->code){
		case PSN_SETACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
			return	TRUE ;

		case PSN_KILLACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, FALSE) ;
			return	TRUE ;

		case PSN_APPLY:
		{
			if (skkimeConfig_UpdateZenkakuVector (pConfArg->m_rZenkakuVector)) {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
			} else {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
			}
			return	TRUE ;
		}
		case PSN_RESET:
		case PSN_HELP:
			break ;
		default:
			return FALSE ;
		}
		return	TRUE ;
	}
	default:
		break ;
	}
	return	FALSE ;
}

/*========================================================================
 *	private functions
 */
INT_PTR
skkimeConfig_dlgInputVectorOnInitDialog (
	register HWND		hDlg,
	register WPARAM		wParam,
	register LPARAM		lParam,
	register int		nResourceId)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register TINYMEMMGR*		pMemMgr ;
	register HWND				hwndInputVectorList ;
	register BOOL				fZenkaku	= (nResourceId == IDC_LIST_ZENKAKUVECTOR) ;
	register LPMYSTR*			pInputVector ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_dlgInputVectorOnInitDialog (%lx)\n"), hDlg)) ;

	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR)lParam) ;
	pPropSheet			= (LPPROPSHEETPAGE) lParam ;
	pConfArg			= (SkkimeConfigArg *) pPropSheet->lParam ;
	hwndInputVectorList	= GetDlgItem (hDlg, nResourceId) ;
	if (hwndInputVectorList != NULL && pConfArg != NULL) {
		register int	i, nLength, nItem ;
		MYLVITEM		lvI ;
		MYCHAR			rszBuffer [256] ;

		pMemMgr			= &pConfArg->m_MemMgr ;
		pInputVector	= (fZenkaku)? pConfArg->m_rZenkakuVector : pConfArg->m_rInputVector ;

		MyListView_DeleteAllItems (hwndInputVectorList) ;
		MyListView_SetExtendedListViewStyle (hwndInputVectorList, LVS_EX_FULLROWSELECT) ;

		lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
		lvI.state		= 0 ; 
		lvI.stateMask	= 0 ;

		nItem			= 0 ;
		for (i = 0 ; i < MAX_INPUTVECTOR ; i ++) {
			if (pInputVector  [i] != NULL) {
				lvI.iItem		= nItem ;
				lvI.iImage		= 0 ;
				lvI.iSubItem	= 0 ;
				lvI.lParam		= (LPARAM) i ;
				nLength			= Mylstrlen (pInputVector [i]) ;
				/* 128 ȂB*/
				if (nLength < MAX_LENGTH_INPUTVECTOR) {
					Mywsprintf (rszBuffer, MYTEXT ("[%2x] `%c' = `%s'"), i, i, pInputVector [i]) ;
					lvI.pszText		= rszBuffer ;
					if (MyListView_InsertItem (hwndInputVectorList, &lvI) != -1) {
						nItem	++ ;
					} else {
						DEBUGPRINTF ((MYTEXT ("dlgInputVectorProc: MyListView_InsertItem () failed (%d).\n"), i)) ;
					}
				} else {
					TinyMalloc_Free (pMemMgr, pInputVector [i]) ;
					pInputVector [i]	= NULL ;
				}
			}
		}
	} else {
		DEBUGPRINTF ((MYTEXT ("IDC_LIST_INPUTVECTOR: error (%ld)\n"), GetLastError ())) ;
	}
	return	FALSE ;

	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgInputVectorOnNotify (
	register HWND		hDlg,
	register WPARAM		wParam,
	register LPARAM		lParam,
	register int		nResourceId)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register NMHDR	FAR*		pNMHDR		= (NMHDR *) lParam ;
	register BOOL				fZenkaku	= (nResourceId == IDC_LIST_ZENKAKUVECTOR) ;
	register LPMYSTR*			pInputVector ;
	static TMYMENUITEM			rmi []	= {
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ǉ"), IDC_BUTTON_INSERT, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ύX"), IDC_BUTTON_EDIT, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("폜"), IDC_BUTTON_DELETE, },
	} ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;
	if (pConfArg != NULL && pNMHDR->code == NM_RCLICK) {
		register NMITEMACTIVATE*	pNM 	= (NMITEMACTIVATE*) lParam ;
		register int			n ;

		pInputVector	= (fZenkaku)? pConfArg->m_rZenkakuVector : pConfArg->m_rInputVector ;

		n	= skkimeConfig_PopupMenu (hDlg, rmi, ARRAYSIZE (rmi)) ;
		DEBUGPRINTF ((MYTEXT ("NM: item(%d), lParam(%d)\n"), pNM->iItem, pNM->lParam)) ;
		if (n > 0) 
			skkimeConfig_doEditInputVector (hDlg, nResourceId, ((n != IDC_BUTTON_INSERT)? pNM->iItem : -1), n, pInputVector) ;
		DEBUGPRINTF ((MYTEXT ("TrackPopupMenu returns %d\n"), n)) ;
	}
	return	FALSE ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgInputVectorOnCommand (
	register HWND		hDlg,
	register WPARAM		wParam,
	register LPARAM		lParam,
	register int		nResourceId)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register WORD	wNotifyCode ;
	register WORD	wID ;
	register HWND	hwndCtl ;
	register BOOL	fZenkaku ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;
	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;
	fZenkaku	= (nResourceId == IDC_LIST_ZENKAKUVECTOR) ;

	if (pConfArg == NULL)
		return	FALSE ;

	if ((wID == IDC_BUTTON_INSERT || wID == IDC_BUTTON_EDIT || wID == IDC_BUTTON_DELETE) &&
		wNotifyCode == BN_CLICKED) {
		register HWND		hwndLV	= GetDlgItem (hDlg, nResourceId) ;
		register int		nMark	= -1 ;
		register LPMYSTR*	pInputVector ;

		if (hwndLV == NULL)
			return	FALSE ;
		if (wID != IDC_BUTTON_INSERT) {
			nMark	= MyListView_GetSelectionMark (hwndLV) ;
			if (nMark < 0)
				return	FALSE ;
		}
		pInputVector	= (fZenkaku)? pConfArg->m_rZenkakuVector : pConfArg->m_rInputVector ;
		skkimeConfig_doEditInputVector (hDlg, nResourceId, nMark, wID, pInputVector) ;
	}
	return	TRUE ;
}

BOOL
skkimeConfig_doEditInputVector (
	register HWND		hDlg,
	register int		nResourceId,
	register int		nItem,
	register int		nControl,
	register LPMYSTR*	pInputVector)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register TINYMEMMGR*		pMemMgr ;
	register HWND				hwndLV ;
	MYLVFINDINFO				lvfi ;
	MYLVITEM					lvI ;
	register int				nCurSel ;
	MYCHAR						rszBuffer [256] ;

	pPropSheet		= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg		= (SkkimeConfigArg *) pPropSheet->lParam ;
	hwndLV			= GetDlgItem (hDlg, nResourceId) ;
	if (hwndLV == NULL || pConfArg == NULL)
		return	FALSE ;
	pMemMgr			= &pConfArg->m_MemMgr ;

	if (nItem >= 0) {
		lvI.mask		= LVIF_PARAM ;
		lvI.state		= 0 ;
		lvI.stateMask	= 0 ;
		lvI.iItem		= nItem ;
		lvI.iSubItem	= 0 ;
		lvI.pszText		= 0 ;
		lvI.cchTextMax	= 0 ;
		if (! MyListView_GetItem (hwndLV, &lvI)) {
			nCurSel		= -1 ;
		} else {
			nCurSel		= lvI.lParam ;
		}
	} else {
		/* INSERT ̎͑IĂeȂB*/
		nCurSel	= -1 ;
	}

	switch (nControl) {
	case	IDC_BUTTON_EDIT:
		if (nCurSel < 0 || nCurSel >= MAX_INPUTVECTOR)
			return	FALSE ;
		/*	fall through */

	case	IDC_BUTTON_INSERT:
	{
		TDlgEditInputVectorArg	arg ;
		register int			nRetval ;

		arg.m_nCurSel		= nCurSel ;
		arg.m_pInputVector	= pInputVector ;
		nRetval	= DialogBoxParam (pConfArg->m_hInst, MAKEINTRESOURCE (IDD_DIALOG_INPUTVECTOR), hDlg, skkimeConfig_dlgEditInputVector, (LPARAM) &arg) ;
		if (nRetval == IDOK) {
			nCurSel	= arg.m_nCurSel ;

			if (0 <= nCurSel && nCurSel < MAX_INPUTVECTOR) {
				register LPMYSTR	pDest ;
				register int		nLength ;

				if (pInputVector [nCurSel]	!= NULL) {
					lvfi.flags		= LVFI_PARAM ;
					lvfi.lParam		= (LPARAM) nCurSel ;
					nItem			= MyListView_FindItem (hwndLV, -1, &lvfi) ;

					TinyMalloc_Free (pMemMgr, pInputVector [nCurSel]) ;
					pInputVector [nCurSel]	= NULL ;
				} else {
					nItem			= -1 ;
				}
				nLength	= Mylstrlen (arg.m_strValue) ;
				pDest	= TinyMalloc_Malloc (pMemMgr, sizeof (MYCHAR) * (nLength + 1)) ;
				if (pDest != NULL) 
					memcpy (pDest, arg.m_strValue, sizeof (MYCHAR) * (nLength + 1)) ;
				pInputVector [nCurSel]	= pDest ;

				lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
				lvI.state		= 0 ; 
				lvI.stateMask	= 0 ;
				lvI.iImage		= 0 ;
				lvI.iSubItem	= 0 ;
				lvI.lParam		= (LPARAM) nCurSel ;
				Mywsprintf (rszBuffer, MYTEXT ("[%2x] `%c' = `%s'"), nCurSel, nCurSel, pInputVector [nCurSel]) ;
				lvI.pszText		= rszBuffer ;
				if (nItem >= 0) {
					lvI.iItem	= nItem ;
					MyListView_SetItem (hwndLV, &lvI) ;
				} else {
					lvI.iItem	= MyListView_GetItemCount (hwndLV) ;
					MyListView_InsertItem (hwndLV, &lvI) ;
					MyListView_SortItems  (hwndLV, skkimeConfig_compareInputVector, 0) ;
				}
				PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
			}
		}
		break ;
	}
	case	IDC_BUTTON_DELETE:
		if (0 <= nCurSel && nCurSel < MAX_INPUTVECTOR) {
			if (pInputVector [nCurSel]	!= NULL) {
				TinyMalloc_Free (pMemMgr, pInputVector [nCurSel]) ;
				pInputVector [nCurSel]	= NULL ;
			}
			if (! MyListView_DeleteItem (hwndLV, nItem)) {
				DEBUGPRINTF ((MYTEXT ("MyListView_DeleteItem (%lx, %d) failed\n"), 
							  hwndLV, nItem)) ;
			}
			PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		}
		break ;
	default:
		return	FALSE ;
	}
	return	TRUE ;
}

/*
 */
INT_PTR CALLBACK
skkimeConfig_dlgEditInputVector (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register HWND		hwndKey, hwndEdit ;
	MYCHAR				szBuffer [32] ;
	register LPMYSTR*	pInputVector ;
	register TDlgEditInputVectorArg*	pArg ;

	switch (uMsg) {
	case	WM_INITDIALOG:
	{
		register int	i, nCurSel ;

		SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;

		pArg			= (TDlgEditInputVectorArg*) lParam ;
		if (pArg != NULL && 0 <= pArg->m_nCurSel && pArg->m_nCurSel < MAX_INPUTVECTOR) {
			nCurSel			= pArg->m_nCurSel ;
		} else {
			nCurSel			= -1 ;
		}
		hwndKey	= GetDlgItem (hDlg, IDC_COMBO_CHARTOBIND) ;
		if (hwndKey != NULL) {
			/*	̗̈́uvƂĈB*/
			for (i = 0 ; i < MAX_INPUTVECTOR ; i ++) {
				skkimeConfig_key2string (szBuffer, i) ;
				MySendMessage (hwndKey, CB_ADDSTRING, 0, (LPARAM) szBuffer) ;
			}
			if (nCurSel >= 0)
				MySendMessage (hwndKey, CB_SETCURSEL, (WPARAM) nCurSel, (LPARAM) 0) ;
		}
		hwndEdit	= GetDlgItem (hDlg, IDC_EDIT_CHARVALUE) ;
		if (hwndEdit != NULL) {
			pInputVector	= pArg->m_pInputVector ;
			MySendMessage (hwndEdit, EM_LIMITTEXT, (WPARAM) MAX_LENGTH_INPUTVECTOR - 1, 0) ;
			if (0 <= nCurSel && nCurSel < MAX_INPUTVECTOR &&
				pInputVector != NULL && pInputVector [nCurSel] != NULL) 
				MySetWindowText (hwndEdit, pInputVector [nCurSel]) ;
		}
		return	TRUE ;
	}
	case	WM_COMMAND:
		/*	[ށBԂl global ϐoR邵Ȃ... */
		switch (LOWORD (wParam)) {
		case	IDOK:
		{
			register int	nCH ;

			pArg	= (TDlgEditInputVectorArg *)GetWindowLongPtr (hDlg, DWLP_USER) ;
			if (pArg != NULL) {
				hwndKey		= GetDlgItem (hDlg, IDC_COMBO_CHARTOBIND) ;
				hwndEdit	= GetDlgItem (hDlg, IDC_EDIT_CHARVALUE) ;
				nCH				= MySendMessage (hwndKey,        CB_GETCURSEL, 0, 0) ;
				MyGetWindowText (hwndEdit, pArg->m_strValue, MAX_LENGTH_INPUTVECTOR) ;
				pArg->m_strValue [MAX_LENGTH_INPUTVECTOR - 1]	= MYTEXT ('\0') ;
				if (nCH != CB_ERR) {
					pArg->m_nCurSel			= nCH ;
				} else {
					/*	ƒlĂȂ CANCEL ƈB*/
					pArg->m_nCurSel			= -1 ;
					pArg->m_strValue [0]	= TEXT ('\0') ;
					wParam					= IDCANCEL ;
				}
			}
			/*	fall through */
		}
		case	IDCANCEL:
			EndDialog (hDlg, wParam) ;
			return	TRUE ;
		default:
			break ;
		}
		/*	fall through */
	default:
		break ;
	}
	return	FALSE ;
}

int CALLBACK
skkimeConfig_compareInputVector (
	register LPARAM		lParam1,
	register LPARAM		lParam2, 
	register LPARAM		lParamSort)
{
	return	(lParam1 - lParam2) ;
	UNREFERENCED_PARAMETER (lParamSort) ;
}

BOOL
skkimeConfig_initInputVector (
	register TINYMEMMGR*		pMemMgr,
	register LPCMYSTR			pstrSubKey,
	register LPMYSTR*			ppVectorTable,
	register TDEFINPUTVECTOR*	pDefaultVector,
	register int				nDefaultVector)
{
	HKEY				hSubKey ;
	MYCHAR				szRegPathInfo [MAX_PATH] ;
	DWORD				dwRegPathInfo, dwType, cbData ;
	long				lValue ;
	register int		nIndex ;
	register LONG		lResult ;
	register BOOL		fRetval	= TRUE ;
	register int		i ;
	register LPMYSTR	pDest ;

	for (i = 0 ; i < MAX_INPUTVECTOR ; i++)
		*(ppVectorTable + i)	= NULL ;
	
	if (MyRegOpenKeyEx (HKEY_CURRENT_USER, pstrSubKey, 0, KEY_READ, &hSubKey) != ERROR_SUCCESS) {
		register TDEFINPUTVECTOR*	pDefVector ;
		register LPCMYSTR			strVector ;
		register int				nCode, nDest ;

		pDefVector	= pDefaultVector ;
		for (i = 0 ; i < nDefaultVector ; i ++) {
			nCode		= pDefVector->m_nKeyCode ;
			strVector	= pDefVector->m_strInputVector ;
			ASSERT (0 <= nCode && nCode < MAX_INPUTVECTOR) ;
			ASSERT (strVector != NULL) ;

			nDest		= Mylstrlen (strVector) + 1 ;
			pDest		= TinyMalloc_Malloc (pMemMgr, nDest * sizeof (MYCHAR)) ;
			if (pDest == NULL)
				return	FALSE ;

			/*	TCHAR ̊i[ nDest B*/
			Mylstrcpy (pDest, strVector) ;
			DEBUGPRINTF ((MYTEXT ("InputVector(%d):\"%s\"\n"), nCode, pDest)) ;
			*(ppVectorTable + nCode)	= pDest ;
			pDefVector	++ ;
		}
		return	TRUE ;
	}

	for (nIndex = 0 ; ; nIndex ++) {
		dwRegPathInfo	= ARRAYSIZE (szRegPathInfo) ;
		cbData			= 0 ;
		lResult	= MyRegEnumValue (hSubKey, (DWORD)nIndex, szRegPathInfo, &dwRegPathInfo, NULL, NULL, NULL, NULL) ;
		if (lResult != ERROR_SUCCESS)
			break ;
		lResult	= MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, &dwType, NULL, &cbData) ;
		if (lResult != ERROR_SUCCESS)
			continue ;
		/* 傷f[^͎̂ĂB*/
		if (dwType != REG_SZ) 
			continue ;
		/* vector[] = f[^ ƂB*/
		lValue	= Myatol (szRegPathInfo) ;
		if (lValue < 0 || lValue >= MAX_INPUTVECTOR)
			continue ;
		pDest	= TinyMalloc_Malloc (pMemMgr, cbData) ;
		if (pDest == NULL) {
			fRetval	= FALSE ;
			break ;
		}
		(void) MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, NULL, (BYTE*)pDest, &cbData) ;
		*(ppVectorTable + lValue)	= pDest ;
	}
	MyRegCloseKey (hSubKey) ;
	return	fRetval ;
}

BOOL
skkimeConfig_updateInputVector (
	register LPCMYSTR	pstrSubKey,
	register LPMYSTR*	ppInputVector)
{
	HKEY				hSubKey ;
	MYCHAR				szRegPathInfo [MAX_PATH] ;
	register LPMYSTR*	pTable ;
	register int		i ;
	
	if (! skkimeConfig_CreateKey (pstrSubKey, TRUE, &hSubKey)) {
		DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateInputVector (): create-key(\"%s\") failed\n"), pstrSubKey)) ;
		return	FALSE ;
	}

	/*	̒lo^B*/
	pTable	= ppInputVector ;
	for (i = 0 ; i < MAX_INPUTVECTOR ; i ++) {
		if (*pTable != NULL && **pTable != MYTEXT ('\0')) {
			Mywsprintf (szRegPathInfo, MYTEXT ("%d"), i) ;
			/*	REG_SZ œo^ꍇɂ́AI[ NUL ܂߂TCY cbdata 
			 *	nȂ΂ȂȂBvӁB
			 */
			if (MyRegSetValueEx (hSubKey, szRegPathInfo, 0, REG_SZ, (BYTE *)*pTable, sizeof (MYCHAR) * (Mylstrlen (*pTable) + 1)) != ERROR_SUCCESS) {
				break ;
			}
		}
		pTable	++ ;
	}
	MyRegCloseKey (hSubKey) ;
	return	(i < MAX_INPUTVECTOR)? FALSE : TRUE ;
}

