/* # SKKIME (Simple Kana-Kanji Input Method Editor for Windows'98)
 * Config.c
 * This file is part of SKKIME.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "commctrl.h"
#include "shlwapi.h"
#include "skki1_0.h"
#include <limits.h>
#include "conf\skimcnfp.h"
#include "resource.h"

#define	REGPATH_ROMAKANARULELIST	MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\RomaKanaRuleList")
#define	MAX_LENGTH_ROMAKANARULE_STATE	(64)
#define	MAX_LENGTH_ROMAKANARULE_HOUTPUT	(64)
#define	MAX_LENGTH_ROMAKANARULE_KOUTPUT	(64)

typedef struct tagSKKROMAKANARULE {
	LPMYSTR			state ;
	LPMYSTR			next ;
	LPMYSTR			houtput ;
	LPMYSTR			koutput ;
}	SKKROMAKANARULE, NEAR *PSKKROMAKANARULE, FAR *LPSKKROMAKANARULE ;

typedef struct {
	MYCHAR		m_strState     [MAX_LENGTH_ROMAKANARULE_STATE] ;
	MYCHAR		m_strNextState [MAX_LENGTH_ROMAKANARULE_STATE] ;
	MYCHAR		m_strHoutput   [MAX_LENGTH_ROMAKANARULE_HOUTPUT] ;
	MYCHAR		m_strKoutput   [MAX_LENGTH_ROMAKANARULE_KOUTPUT] ;
}	TDlgEditRomaKanaRuleArg ;

static	INT_PTR	skkimeConfig_dlgRomaKanaRuleOnInitDialog	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgRomaKanaRuleOnNotify		(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR	skkimeConfig_dlgRomaKanaRuleOnCommand		(HWND, UINT, WPARAM, LPARAM) ;

static	BOOL	skkimeConfig_editRomaKanaRule				(HWND, int, int) ;
static	INT_PTR CALLBACK		skkimeConfig_dlgEditRomaKanaRule	(HWND, UINT, WPARAM, LPARAM) ;

static	BOOL	skkimeConfig_initRomaKanaRuleList	(TINYMEMMGR*, LPCMYSTR, SKKROMAKANARULENODE**, SKKROMAKANARULE*, int) ;
static	BOOL	skkimeConfig_updateRomaKanaRuleList	(LPCMYSTR, SKKROMAKANARULENODE*) ;
static	SKKROMAKANARULENODE*	skkimeConfig_newRomaKanaRuleNode	(TINYMEMMGR*, int, int, int, int) ;
static	SKKROMAKANARULENODE*	skkimeConfig_findRomaKanaRuleNode	(SKKROMAKANARULENODE**, LPCMYSTR) ;
static	BOOL	skkimeConfig_registerRomaKanaRuleNode	(SKKROMAKANARULENODE**, SKKROMAKANARULENODE*) ;
static	BOOL	skkimeConfig_unregisterRomaKanaRuleNode	(SKKROMAKANARULENODE**, SKKROMAKANARULENODE*) ;

static	SKKROMAKANARULE		srDefaultRomaKanaRuleList [] = {
	{ MYTEXT ("nn"),	MYTEXT (""),	MYTEXT (""),	MYTEXT ("")},
	{ MYTEXT ("n'"),	MYTEXT (""),	MYTEXT (""),	MYTEXT ("")},
} ;

/*========================================================================
 *	public functions
 */
BOOL
skkimeConfig_InitializeRomaKanaRuleList (
	register TINYMEMMGR*			pMemMgr,
	register SKKROMAKANARULENODE**	plstTop)
{
	if (pMemMgr == NULL || plstTop == NULL)
		return	FALSE ;

	return	skkimeConfig_initRomaKanaRuleList (pMemMgr, REGPATH_ROMAKANARULELIST, plstTop, srDefaultRomaKanaRuleList, ARRAYSIZE (srDefaultRomaKanaRuleList)) ;
}

BOOL
skkimeConfig_UpdateRomaKanaRuleList (
	register SKKROMAKANARULENODE*	lstTop)
{
	if (lstTop == NULL)
		return	FALSE ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_UpdateRomaKanaRuleList (%p)\n"), lstTop)) ;
	return	skkimeConfig_updateRomaKanaRuleList (REGPATH_ROMAKANARULELIST, lstTop) ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgRomaKanaRuleProc (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	switch (message) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgRomaKanaRuleOnInitDialog (hDlg, message, wParam, lParam) ;
	case	WM_NOTIFY:
		return	skkimeConfig_dlgRomaKanaRuleOnNotify (hDlg, message, wParam, lParam) ;
	case	WM_COMMAND:
		return	skkimeConfig_dlgRomaKanaRuleOnCommand (hDlg, message, wParam, lParam) ;
	default:
		break ;
	}
	return	FALSE ;

	UNREFERENCED_PARAMETER (wParam) ;
}

/*========================================================================
 *	private functions
 */
INT_PTR
skkimeConfig_dlgRomaKanaRuleOnInitDialog (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	static LPMYSTR	rszColumnTitle []	= {
		MYTEXT(""), MYTEXT(""), MYTEXT("o"), MYTEXT("Љo"),
	} ;
	static int		rnColumnWidth []	= {
		60, 60, 70, 80,
	} ;
	register LPPROPSHEETPAGE		pPropSheet ;
	register SkkimeConfigArg*		pConfArg ;
	register HWND					hwndRomaKanaRule ;
	register SKKROMAKANARULENODE*	pNode ;
	register int					n ;
	MYLV_COLUMN						lvColumn;
	MYLVITEM						lvI ;

	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;
	pPropSheet	= (LPPROPSHEETPAGE) lParam ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	hwndRomaKanaRule	= GetDlgItem (hDlg, IDC_LIST_ROMAKANARULE) ;
	if (hwndRomaKanaRule != NULL) {
		MyListView_DeleteAllItems (hwndRomaKanaRule) ;
		MyListView_SetExtendedListViewStyle (hwndRomaKanaRule, LVS_EX_FULLROWSELECT) ;

		lvColumn.mask	= LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM ;
		lvColumn.fmt	= LVCFMT_LEFT ;

		// Insert the five columns in the list view.
		for (n = 0; n < 4 ; n ++) {
			lvColumn.pszText	= rszColumnTitle [n] ;
			lvColumn.cx			= rnColumnWidth [n] ;
			MyListView_InsertColumn (hwndRomaKanaRule, n, &lvColumn) ;
		}

		lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
		lvI.state		= 0 ; 
		lvI.stateMask	= 0 ;
		pNode			= pConfArg->m_lstRomaKanaRule ;
		n				= 0 ;
		while (pNode != NULL) {
			register int	nItem ;
			lvI.iItem		= n ;
			lvI.iImage		= n ;
			lvI.iSubItem	= 0 ;
			lvI.lParam		= (LPARAM) pNode ;
			lvI.pszText		= pNode->m_strState ;
			nItem	= MyListView_InsertItem (hwndRomaKanaRule, &lvI) ;
			if (nItem != -1) {
				MyListView_SetItemText (hwndRomaKanaRule, nItem, 1, pNode->m_strNext) ;
				MyListView_SetItemText (hwndRomaKanaRule, nItem, 2, pNode->m_strHoutput) ;
				MyListView_SetItemText (hwndRomaKanaRule, nItem, 3, pNode->m_strKoutput) ;
			}
			pNode	= pNode->m_pNext ;
			n		++ ;
		}
	}
	return	TRUE ;
	UNREFERENCED_PARAMETER (message) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgRomaKanaRuleOnCommand (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register WORD	wNotifyCode ;
	register WORD	wID ;
	register HWND	hwndCtl ;

	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;

	DEBUGPRINTF ((MYTEXT ("HWND:(%lx), ID(%u), Notify(%u), CTLWND:(%lx)\n"),
				  hDlg, wID, wNotifyCode, hwndCtl)) ;

	switch (wID) {
	case	IDC_BUTTON_INSERT:
	case	IDC_BUTTON_EDIT:
	case	IDC_BUTTON_DELETE:
		if (wNotifyCode == BN_CLICKED) {
			register int	nItem	= -1 ;

			if (wID != IDC_BUTTON_INSERT) {
				register HWND	hwndLV ;

				hwndLV	= GetDlgItem (hDlg, IDC_LIST_ROMAKANARULE) ;
				if (hwndLV == NULL)
					return	1 ;
				nItem	= MyListView_GetSelectionMark (hwndLV) ;
			}
			skkimeConfig_editRomaKanaRule (hDlg, nItem, wID) ;
			return	0 ;
		}
		break ;
	default:
		break ;
	}
	return	1 ;
	UNREFERENCED_PARAMETER (message) ;
}

INT_PTR
skkimeConfig_dlgRomaKanaRuleOnNotify (
	register HWND		hDlg,
	register UINT		message,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register NMHDR	FAR*		pNMHDR	= (NMHDR *) lParam ;
	static TMYMENUITEM		rmi []	= {
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ǉ"), IDC_BUTTON_INSERT, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ύX"), IDC_BUTTON_EDIT, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("폜"), IDC_BUTTON_DELETE, },
	} ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	switch (pNMHDR->idFrom) {
	case	IDC_LIST_ROMAKANARULE:
	{
		if (pNMHDR->code == NM_RCLICK) {
			register NMITEMACTIVATE*	pNM 	= (NMITEMACTIVATE*) lParam ;
			register int	n ;

			n	= skkimeConfig_PopupMenu (hDlg, rmi, ARRAYSIZE (rmi)) ;
			DEBUGPRINTF ((MYTEXT ("NM: item(%d), lParam(%d)\n"), pNM->iItem, pNM->lParam)) ;
			if (n > 0) 
				skkimeConfig_editRomaKanaRule (hDlg, ((n != IDC_BUTTON_INSERT)? pNM->iItem : -1), n) ;
			DEBUGPRINTF ((MYTEXT ("TrackPopupMenu returns %d\n"), n)) ;
		}
		return	FALSE ;
	}
	default:
		switch (pNMHDR->code){
		case PSN_SETACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
			return	TRUE ;

		case PSN_KILLACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, FALSE) ;
			return	TRUE ;

		case PSN_APPLY:
		{
			if (skkimeConfig_UpdateRomaKanaRuleList (pConfArg->m_lstRomaKanaRule)) {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
			} else {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
			}
			return	TRUE ;
		}
		case PSN_RESET:
		case PSN_HELP:
			break ;

		default:
			return FALSE ;
		}
		return	TRUE ;
	}

	UNREFERENCED_PARAMETER (message) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

BOOL
skkimeConfig_editRomaKanaRule (
	register HWND		hDlg,
	register int		nItem,
	register int		nCommand)
{
	register LPPROPSHEETPAGE		pPropSheet ;
	register SkkimeConfigArg*		pConfArg ;
	register TINYMEMMGR*			pMemMgr ;
	register HWND					hwndLV ;
	register SKKROMAKANARULENODE**	plstTop ;
	register SKKROMAKANARULENODE*	pNode ;
	MYLVITEM						lvI ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;
	hwndLV		= GetDlgItem (hDlg, IDC_LIST_ROMAKANARULE) ;
	if (hwndLV == NULL || pConfArg == NULL)
		return	FALSE ;
	plstTop		= &pConfArg->m_lstRomaKanaRule ;
	pMemMgr		= &pConfArg->m_MemMgr ;

	if (nItem >= 0) {
		lvI.mask		= LVIF_PARAM ;
		lvI.state		= 0 ;
		lvI.stateMask	= 0 ;
		lvI.iItem		= nItem ;
		lvI.iSubItem	= 0 ;
		lvI.pszText		= 0 ;
		lvI.cchTextMax	= 0 ;
		if (! MyListView_GetItem (hwndLV, &lvI)) {
			pNode		= NULL ;
		} else {
			pNode		= (SKKROMAKANARULENODE*) lvI.lParam ;
		}
	} else {
		pNode	= NULL ;
	}
	switch (nCommand) {
	case	IDC_BUTTON_INSERT:
	case	IDC_BUTTON_EDIT:
	{
		register SKKROMAKANARULENODE*	pNewNode ;
		register int			nRetval, nStateLen, nNextLen, nHoutputLen, nKoutputLen ;
		register BOOL			f	= FALSE ;
		TDlgEditRomaKanaRuleArg	arg ;

		if (pNode != NULL) {
			Mylstrcpyn (arg.m_strState,     pNode->m_strState,   MAX_LENGTH_ROMAKANARULE_STATE) ;
			Mylstrcpyn (arg.m_strNextState, pNode->m_strNext,    MAX_LENGTH_ROMAKANARULE_STATE) ;
			Mylstrcpyn (arg.m_strHoutput,   pNode->m_strHoutput, MAX_LENGTH_ROMAKANARULE_HOUTPUT) ;
			Mylstrcpyn (arg.m_strKoutput,   pNode->m_strKoutput, MAX_LENGTH_ROMAKANARULE_KOUTPUT) ;
			arg.m_strState     [MAX_LENGTH_ROMAKANARULE_STATE - 1]	= MYTEXT ('\0') ;
			arg.m_strNextState [MAX_LENGTH_ROMAKANARULE_STATE - 1]	= MYTEXT ('\0') ;
			arg.m_strHoutput   [MAX_LENGTH_ROMAKANARULE_HOUTPUT - 1]	= MYTEXT ('\0') ;
			arg.m_strKoutput   [MAX_LENGTH_ROMAKANARULE_KOUTPUT - 1]	= MYTEXT ('\0') ;
		} else {
			arg.m_strState     [0]	= MYTEXT ('\0') ;
			arg.m_strNextState [0]	= MYTEXT ('\0') ;
			arg.m_strHoutput   [0]	= MYTEXT ('\0') ;
			arg.m_strKoutput   [0]	= MYTEXT ('\0') ;
		}
		nRetval	= DialogBoxParam (pConfArg->m_hInst, MAKEINTRESOURCE (IDD_DIALOG_EDIT_ROMAKANARULE), hDlg, skkimeConfig_dlgEditRomaKanaRule, (LPARAM) &arg) ;
		if (nRetval != IDOK) 
			break ;

		nStateLen	= Mylstrlen (arg.m_strState)     + 1 ;
		nNextLen	= Mylstrlen (arg.m_strNextState) + 1 ;
		nHoutputLen	= Mylstrlen (arg.m_strHoutput)   + 1 ;
		nKoutputLen	= Mylstrlen (arg.m_strKoutput)   + 1 ;
		if (nStateLen <= 1) {
			DEBUGPRINTF ((MYTEXT ("fatal: state is empty.\n"))) ;
			return	FALSE ;
		}
		pNewNode	= skkimeConfig_newRomaKanaRuleNode (pMemMgr, nStateLen, nNextLen, nHoutputLen, nKoutputLen) ;
		Mylstrcpy (pNewNode->m_strState,   arg.m_strState) ;
		Mylstrcpy (pNewNode->m_strNext,    arg.m_strNextState) ;
		Mylstrcpy (pNewNode->m_strHoutput, arg.m_strHoutput) ;
		Mylstrcpy (pNewNode->m_strKoutput, arg.m_strKoutput) ;

		/*	Ԃ𕡐邱Ƃ͂łȂB*/
		pNode	= skkimeConfig_findRomaKanaRuleNode (plstTop, arg.m_strState) ;

		lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
		lvI.state		= 0 ; 
		lvI.stateMask	= 0 ;
		lvI.iImage		= 0 ;
		lvI.iSubItem	= 0 ;
		lvI.lParam		= (LPARAM) pNewNode ;
		if (! skkimeConfig_registerRomaKanaRuleNode (plstTop, pNewNode)) {
			DEBUGPRINTF ((MYTEXT ("fatal: failed in registering roma-kana-rule-node(%p)\n"),
						  pNewNode)) ;
			TinyMalloc_Free (pMemMgr, pNewNode) ;
			return	FALSE ;
		}
		lvI.pszText		= pNewNode->m_strState ;
		if (pNode != NULL) {
			LVFINDINFO		lvfi ;

			lvfi.flags		= LVFI_PARAM ;
			lvfi.lParam		= (LPARAM) pNode ;
			nItem			= MyListView_FindItem (hwndLV, -1, &lvfi) ;
			if (nItem >= 0) {
				lvI.iItem		= nItem ;
				f	= MyListView_SetItem (hwndLV, &lvI) ;
			}
			skkimeConfig_unregisterRomaKanaRuleNode (plstTop, pNode) ;
			TinyMalloc_Free (pMemMgr, pNode) ;
		} else {
			nItem	= MyListView_GetItemCount (hwndLV) ;
			nItem	= MyListView_InsertItem (hwndLV, &lvI) ;
			f		= (nItem != -1) ;
		}
		if (f) {
			/*	eLXg̐ݒB*/
			MyListView_SetItemText (hwndLV, nItem, 1, (LPMYSTR) pNewNode->m_strNext) ;
			MyListView_SetItemText (hwndLV, nItem, 2, (LPMYSTR) pNewNode->m_strHoutput) ;
			MyListView_SetItemText (hwndLV, nItem, 3, (LPMYSTR) pNewNode->m_strKoutput) ;
		}
		PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		break ;
	}
	case	IDC_BUTTON_DELETE:
		if (pNode == NULL || nItem < 0) {
			DEBUGPRINTF ((MYTEXT ("editRomaKanaRule: no row is selected\n"))) ;
			return	FALSE ;
		}
		MyListView_DeleteItem (hwndLV, nItem) ;
		skkimeConfig_unregisterRomaKanaRuleNode (plstTop, pNode) ;
		TinyMalloc_Free (pMemMgr, pNode) ;
		PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		break ;
	default:
		return	FALSE ;
	}
	return	TRUE ;
}

INT_PTR CALLBACK
skkimeConfig_dlgEditRomaKanaRule (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register TDlgEditRomaKanaRuleArg*	pArg ;
	register HWND	hwndState, hwndNext, hwndHoutput, hwndKoutput ;

	switch (uMsg) {
	case	WM_INITDIALOG:
	{
		SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;

		pArg		= (TDlgEditRomaKanaRuleArg*) lParam ;
		hwndState	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_STATE) ;
		hwndNext	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_NEXTSTATE) ;
		hwndHoutput	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_HOUTPUT) ;
		hwndKoutput	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_KOUTPUT) ;

		if (pArg == NULL || hwndState == NULL || hwndNext == NULL ||
			hwndHoutput == NULL || hwndKoutput == NULL) {
			DEBUGPRINTF ((MYTEXT ("fatal: can't initialize dialog: IDD_DIALOG_EDIT_ROMAKANARULE\n"))) ;
			return	TRUE ;
		}
		MySendMessage (hwndState,   EM_LIMITTEXT, (WPARAM) MAX_LENGTH_ROMAKANARULE_STATE, 0) ;
		MySendMessage (hwndNext,    EM_LIMITTEXT, (WPARAM) MAX_LENGTH_ROMAKANARULE_STATE, 0) ;
		MySendMessage (hwndHoutput, EM_LIMITTEXT, (WPARAM) MAX_LENGTH_ROMAKANARULE_HOUTPUT, 0) ;
		MySendMessage (hwndKoutput, EM_LIMITTEXT, (WPARAM) MAX_LENGTH_ROMAKANARULE_KOUTPUT, 0) ;
		MySetWindowText (hwndState,   pArg->m_strState) ;
		MySetWindowText (hwndNext,    pArg->m_strNextState) ;
		MySetWindowText (hwndHoutput, pArg->m_strHoutput) ;
		MySetWindowText (hwndKoutput, pArg->m_strKoutput) ;
		return	TRUE ;
	}
	case	WM_COMMAND:
		switch (LOWORD (wParam)) {
		case	IDOK:
		{
			pArg		= (TDlgEditRomaKanaRuleArg*) GetWindowLongPtr (hDlg, DWLP_USER) ;
			hwndState	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_STATE) ;
			hwndNext	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_NEXTSTATE) ;
			hwndHoutput	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_HOUTPUT) ;
			hwndKoutput	= GetDlgItem (hDlg, IDC_EDIT_ROMAKANARULE_KOUTPUT) ;
			MyGetWindowText (hwndState,   pArg->m_strState,     MAX_LENGTH_ROMAKANARULE_STATE) ;
			MyGetWindowText (hwndNext,    pArg->m_strNextState, MAX_LENGTH_ROMAKANARULE_STATE) ;
			MyGetWindowText (hwndHoutput, pArg->m_strHoutput,   MAX_LENGTH_ROMAKANARULE_HOUTPUT) ;
			MyGetWindowText (hwndKoutput, pArg->m_strKoutput,   MAX_LENGTH_ROMAKANARULE_KOUTPUT) ;
			if (pArg->m_strState [0] == MYTEXT ('\0'))
				wParam	= IDCANCEL ;
		}
		/*	fall through */
		case	IDCANCEL:
			EndDialog (hDlg, wParam) ;
			return	TRUE ;
		default:
			break ;
		}
		break ;
	default:
		break ;
	}
	return	FALSE ;
}

/*	. . .   .
 *	RoMaKanaRule
 */
BOOL
skkimeConfig_initRomaKanaRuleList (
	register TINYMEMMGR*			pMemMgr,
	register LPCMYSTR				pstrSubKey,
	register SKKROMAKANARULENODE**	plstRuleListTop,
	register SKKROMAKANARULE*		pDefaultRuleList,
	register int					nDefaultRuleList)
{
	HKEY				hSubKey ;
	MYCHAR				szRegPathInfo [MAX_PATH] ;
	MYCHAR				rbyData [MAX_REGVALUE_SIZE] ;
	DWORD				dwRegPathInfo, dwType, cbData ;
	register SKKROMAKANARULENODE*	pNode ;
	register SKKROMAKANARULENODE*	pLastNode ;
	register int		nIndex, nStateLen, nNextLen, nHoutputLen, nKoutputLen ;
	register LONG		lResult ;

	pLastNode	= NULL ;
	pNode		= NULL ;
	if (MyRegOpenKeyEx (HKEY_CURRENT_USER, pstrSubKey, 0, KEY_READ, &hSubKey) == ERROR_SUCCESS) {
		register BOOL	fError	= FALSE ;

		/*	[}ϊe[uo^Ă̂ȂcB*/
		for (nIndex = 0 ; ; nIndex ++) {
			LPCMYSTR	strState, strNext, strHoutput, strKoutput ;

			dwRegPathInfo	= ARRAYSIZE (szRegPathInfo) ;
			cbData			= sizeof (rbyData) ;
			lResult	= MyRegEnumValue (hSubKey, (DWORD)nIndex, szRegPathInfo, &dwRegPathInfo, NULL, NULL, NULL, NULL) ;
			if (lResult != ERROR_SUCCESS)
				break ;

			/*	szRegPathInfo = "nn" ̂悤 szRegPathInfo uԁvɑB
			 *	R (, , o) Ȃ̂ŁAԂdȂ邱Ƃ͂
			 *	Ă͂ȂȂB
			 *	Ԃuv͂蓾B
			 */
			lResult	= MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, &dwType, NULL, &cbData) ;
			if (lResult != ERROR_SUCCESS || dwType != REG_MULTI_SZ)
				continue ;
			/*	Too big value. */
			if (cbData >= sizeof (rbyData)) 
				continue ;
			cbData	= sizeof (rbyData) ;
			(void) MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, NULL, (BYTE *)rbyData, &cbData) ;
			
			nKoutputLen	= 0 ;
			if (TEXT ('0') <= szRegPathInfo [0] && szRegPathInfo [0] <= TEXT ('9')) {
				nStateLen	= Mylstrlen ((LPCMYSTR) rbyData) ;
				nNextLen	= (nStateLen   > 0)? Mylstrlen ((LPCMYSTR) rbyData + nStateLen + 1) : 0 ;
				nHoutputLen	= (nNextLen    > 0)? Mylstrlen ((LPCMYSTR) rbyData + nStateLen + 1 + nNextLen + 1) : 0 ;
				if (nStateLen <= 0 || nNextLen < 1 || nHoutputLen < 1) {
					fError	= TRUE ;
					break ;
				}
				/*	nKoutputLen == 0 ̏ꍇɂ́A] assoc rule ̉\B
				 *	] (next, hira, kata)  3 Acase l
				 *	(now, next, hira, kata)  4 gɂ̂ŁB
				 */
				nKoutputLen	= (nHoutputLen > 0)? Mylstrlen ((LPCMYSTR) rbyData + nStateLen + 1 + nNextLen + 1 + nHoutputLen + 1) : 0 ;
				strState	= (LPCMYSTR) rbyData ;
				strNext		= (LPCMYSTR) rbyData + nStateLen + 1 ;
				strHoutput	= (LPCMYSTR) rbyData + nStateLen + 1 + nNextLen + 1 ;
				strKoutput	= (LPCMYSTR) rbyData + nStateLen + 1 + nNextLen + 1 + nHoutputLen + 1 ;
			}
			if (nKoutputLen <= 0) {
				nStateLen	= Mylstrlen ((LPCMYSTR) szRegPathInfo) ;
				nNextLen	= Mylstrlen ((LPCMYSTR) rbyData) ;
				nHoutputLen	= (nNextLen    > 0)? Mylstrlen ((LPCMYSTR) rbyData + nNextLen + 1) : 0 ;
				nKoutputLen	= (nHoutputLen > 0)? Mylstrlen ((LPCMYSTR) rbyData + nNextLen + 1 + nHoutputLen + 1) : 0 ;
				strState	= (LPCMYSTR) szRegPathInfo ;
				strNext		= (LPCMYSTR) rbyData ;
				strHoutput	= (LPCMYSTR) rbyData + nNextLen + 1 ;
				strKoutput	= (LPCMYSTR) rbyData + nNextLen + 1 + nHoutputLen + 1 ;
			}
			if (nStateLen <= 0 || nNextLen < 1 || nHoutputLen < 1 || nKoutputLen < 1) {
				fError	= TRUE ;
				break ;
			}

			pNode		= skkimeConfig_newRomaKanaRuleNode (pMemMgr, nStateLen, nNextLen, nHoutputLen, nKoutputLen) ;
			if (pNode == NULL) {
				fError	= TRUE ;
				break ;
			}
			Mylstrcpy (pNode->m_strState,	strState) ;
			Mylstrcpy (pNode->m_strNext,	strNext    + 1) ;
			Mylstrcpy (pNode->m_strHoutput,	strHoutput + 1) ;
			Mylstrcpy (pNode->m_strKoutput,	strKoutput + 1) ;

			pNode->m_pPrev	= pLastNode ;
			if (pLastNode != NULL)
				pLastNode->m_pNext	= pNode ;
			pLastNode		= pNode ;
		}
		MyRegCloseKey (hSubKey) ;
		if (fError)
			return	FALSE ;
	} else if (pDefaultRuleList != NULL && nDefaultRuleList > 0) {
		register SKKROMAKANARULE*	pRule ;

		pRule	= pDefaultRuleList ;
		for (nIndex = 0 ; nIndex < nDefaultRuleList ; nIndex ++) {
			nStateLen	= Mylstrlen (pRule->state) ;
			nNextLen	= Mylstrlen (pRule->next)    + 1 ;
			nHoutputLen	= Mylstrlen (pRule->houtput) + 1 ;
			nKoutputLen	= Mylstrlen (pRule->koutput) + 1 ;
			pNode		= skkimeConfig_newRomaKanaRuleNode (pMemMgr, nStateLen, nNextLen, nHoutputLen, nKoutputLen) ;
			if (pNode == NULL)
				return	FALSE ;
			Mylstrcpy (pNode->m_strState,	pRule->state) ;
			Mylstrcpy (pNode->m_strNext,	pRule->next) ;
			Mylstrcpy (pNode->m_strHoutput,	pRule->houtput) ;
			Mylstrcpy (pNode->m_strKoutput,	pRule->koutput) ;

			pNode->m_pPrev	= pLastNode ;
			if (pLastNode != NULL)
				pLastNode->m_pNext	= pNode ;
			pLastNode		= pNode ;

			pRule	++ ;
		}
	}
	if (pLastNode != NULL) {
		pNode	= pLastNode ;
		while (pNode->m_pPrev != NULL)
			pNode	= pNode->m_pPrev ;
		*plstRuleListTop	= pNode ;
	} else {
		*plstRuleListTop	= pNode ;
	}
	return	TRUE ;
}

BOOL
skkimeConfig_updateRomaKanaRuleList (
	register LPCMYSTR				pstrSubKey,
	register SKKROMAKANARULENODE*	lstRuleListTop)
{
	register SKKROMAKANARULENODE*	pNode ;
	register int	nStateLen, nNextLen, nHoutLen, nKoutLen, nTotal, n ;
	MYCHAR			szData [MAX_REGVALUE_SIZE] ;
	HKEY			hSubKey ;
	TCHAR			buf [16] ;

	if (! skkimeConfig_CreateKey (pstrSubKey, TRUE, &hSubKey))
		return	FALSE ;

	pNode	= lstRuleListTop ;
	n		= 1 ;
	while (pNode != NULL) {
		/* MULTI_SZ ƁB*/
		nStateLen	= Mylstrlen (pNode->m_strState) ;
		nNextLen	= Mylstrlen (pNode->m_strNext) ;
		nHoutLen	= Mylstrlen (pNode->m_strHoutput) ;
		nKoutLen	= Mylstrlen (pNode->m_strKoutput) ;
		nTotal		= nStateLen + nNextLen + nHoutLen + nKoutLen ;
		if (nStateLen > 0 && 0 <= nTotal && nTotal < (MAX_REGVALUE_SIZE - (4 + 3 + 1))) {
			register LPMYSTR	pDest	= szData ;
			DEBUGPRINTF ((MYTEXT ("State(%s), Next(%s), H(%s), K(%s)\n"),
						  pNode->m_strState, pNode->m_strNext,
						  pNode->m_strHoutput, pNode->m_strKoutput)) ;
			Mylstrcpy (pDest, pNode->m_strState) ;		pDest	+= nStateLen + 1 ;
			Mylstrcpy (pDest, MYTEXT ("_")) ;			pDest	++ ;
			Mylstrcpy (pDest, pNode->m_strNext) ;		pDest	+= nNextLen + 1 ;
			Mylstrcpy (pDest, MYTEXT ("_")) ;			pDest	++ ;
			Mylstrcpy (pDest, pNode->m_strHoutput) ;	pDest	+= nHoutLen + 1 ;
			Mylstrcpy (pDest, MYTEXT ("_")) ;			pDest	++ ;
			Mylstrcpy (pDest, pNode->m_strKoutput) ;	pDest	+= nKoutLen + 1 ;
			*pDest	= MYTEXT ('\0') ;
			wnsprintf (buf, ARRAYSIZE (buf) - 1, TEXT ("%d"), n) ;
			buf [ARRAYSIZE (buf) - 1]	= TEXT ('\0') ;
			if (MyRegSetValueEx (hSubKey, buf, 0, REG_MULTI_SZ, (BYTE *)szData, sizeof (MYCHAR) * (nTotal + 4 + 3 + 1)) != ERROR_SUCCESS) {
				DEBUGPRINTF ((MYTEXT ("RegSetValueEx: failed (\"%s\", \"%s\")\n"),
							  pNode->m_strState, szData)) ;
				break ;
			}
			n	++ ;
		}
		pNode	= pNode->m_pNext ;
	}
	MyRegCloseKey (hSubKey) ;
	return	pNode == NULL ;
}

SKKROMAKANARULENODE*
skkimeConfig_newRomaKanaRuleNode (
	register TINYMEMMGR*	pMemMgr,
	register int			nStateLen,
	register int			nNextLen,
	register int			nHoutputLen,
	register int			nKoutputLen)
{
	register SKKROMAKANARULENODE*	pNode ;

	pNode		= TinyMalloc_Malloc (pMemMgr, sizeof (SKKROMAKANARULENODE) + sizeof (MYCHAR) * (nStateLen + 1 + nNextLen + nHoutputLen + nKoutputLen)) ;
	pNode->m_strState		= (LPMYSTR)(pNode + 1) ;
	pNode->m_strNext		= pNode->m_strState   + nStateLen + 1 ;
	pNode->m_strHoutput		= pNode->m_strNext    + nNextLen ;
	pNode->m_strKoutput		= pNode->m_strHoutput + nHoutputLen ;
	pNode->m_pPrev			= NULL ;
	pNode->m_pNext			= NULL ;
	return	pNode ;
}

SKKROMAKANARULENODE*
skkimeConfig_findRomaKanaRuleNode (
	register SKKROMAKANARULENODE**	plstTop,
	register LPCMYSTR				strState)
{
	register SKKROMAKANARULENODE*	pNode ;

	if (plstTop == NULL || strState == NULL) 
		return	NULL ;

	pNode	= *plstTop ;
	while (pNode != NULL) {
		if (pNode->m_strState != NULL && !Mylstrcmp (pNode->m_strState, strState))
			return	pNode ;
		pNode	= pNode->m_pNext ;
	}
	return	NULL ;
}

BOOL
skkimeConfig_registerRomaKanaRuleNode (
	register SKKROMAKANARULENODE**	plstTop,
	register SKKROMAKANARULENODE*	pNode)
{
	register SKKROMAKANARULENODE*	pNextNode ;

	if (plstTop == NULL || pNode == NULL)
		return	FALSE ;

	pNextNode		= *plstTop ;
	pNode->m_pNext	= pNextNode ;
	pNode->m_pPrev	= NULL ;
	if (pNextNode != NULL)
		pNextNode->m_pPrev	= pNode ;
	*plstTop		= pNode ;
	return	TRUE ;
}

BOOL
skkimeConfig_unregisterRomaKanaRuleNode (
	register SKKROMAKANARULENODE**	plstTop,
	register SKKROMAKANARULENODE*	pNode)
{
	register SKKROMAKANARULENODE*	pPrevNode ;
	register SKKROMAKANARULENODE*	pNextNode ;

	if (pNode == NULL || plstTop == NULL)
		return	FALSE ;

	pPrevNode	= pNode->m_pPrev ;
	pNextNode	= pNode->m_pNext ;
	if (pPrevNode == NULL) {
		if (*plstTop != pNode) {
			DEBUGPRINTF ((MYTEXT ("fatal: invalid node? listtop(%p) != node(%p)\n"),
						  *plstTop, pNode)) ;
			return	FALSE ;
		}
		*plstTop	= pNextNode ;
	} else {
		pPrevNode->m_pNext	= pNextNode ;
	}
	if (pNextNode != NULL) {
		pNextNode->m_pPrev	= pPrevNode ;
	}
	pNode->m_pPrev	= NULL ;
	pNode->m_pNext	= NULL ;
	return	TRUE ;

}

