/* # SKKIME (Simple Kana-Kanji Input Method Editor for Windows'98)
 * Config.c
 * This file is part of SKKIME.
 * Copyright (C) 1999
 * Takashi SAKAMOTO (tatari_sakamoto@nifty.ne.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with skkinput; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "windows.h"
#include "tchar.h"
#include "immdev.h"
#include "commctrl.h"
#include "skki1_0.h"
#include <limits.h>
#include "conf\skimcnfp.h"
#include "resource.h"

#define	MAX_LENGTH_ROMAKANAPREFIX		(128)
#define	MAX_LENGTH_ROMAKANAKANA			(128)
#define	MAX_LENGTH_ROMAKANAKATA			(128)
#define	REGPATH_ROMAKANATABLE_A		MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\RomaKanaTable\\A")
#define	REGPATH_ROMAKANATABLE_I		MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\RomaKanaTable\\I")
#define	REGPATH_ROMAKANATABLE_U		MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\RomaKanaTable\\U")
#define	REGPATH_ROMAKANATABLE_E		MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\RomaKanaTable\\E")
#define	REGPATH_ROMAKANATABLE_O		MYTEXT("Software\\TT\\Windows\\CurrentVersion\\SKKIME\\1.0\\RomaKanaTable\\O")

typedef struct tagROMAKANATABLE {
	LPMYSTR			prefix ;
	LPMYSTR			kana ;
	LPMYSTR			katakana ;
}	ROMAKANATABLE, NEAR *PROMAKANATABLE, FAR *LPROMAKANATABLE ;

typedef struct {
	MYCHAR		m_strPrefix   [MAX_LENGTH_ROMAKANAPREFIX] ;
	MYCHAR		m_strKana     [MAX_LENGTH_ROMAKANAPREFIX] ;
	MYCHAR		m_strKatakana [MAX_LENGTH_ROMAKANAPREFIX] ;
}	TDlgEditRomaKanaTableArg ;

static	INT_PTR		skkimeConfig_dlgRomaKanaTblOnInitDialog	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR		skkimeConfig_dlgRomaKanaTblOnCommand	(HWND, UINT, WPARAM, LPARAM) ;
static	INT_PTR		skkimeConfig_dlgRomaKanaTblOnNotify		(HWND, UINT, WPARAM, LPARAM) ;
static	BOOL		skkimeConfig_initControlRomaKanaTableList	(SkkimeConfigArg*, HWND, BOOL) ;
static	BOOL		skkimeConfig_initControlRomaKanaOneTable	(HWND, BOOL, ROMAKANATABLENODE*) ;
static	BOOL		skkimeConfig_switchKanaKata				(HWND, BOOL) ;

static	BOOL		skkimeConfig_editRomaKanaTable			(HWND, int, int) ;
static	INT_PTR CALLBACK	skkimeConfig_dlgEditRomaKanaTbl	(HWND, UINT, WPARAM, LPARAM) ;
static	int	CALLBACK	skkimeConfig_compareRomaKanaTblNode	(LPARAM, LPARAM, LPARAM) ;
static	BOOL	skkimeConfig_registerRomaKanaTableNode	(SkkimeConfigArg*, ROMAKANATABLENODE*) ;
static	BOOL	skkimeConfig_unregisterRomaKanaTableNode	(SkkimeConfigArg*, ROMAKANATABLENODE*) ;
static	ROMAKANATABLENODE**	skkimeConfig_getRomaKanaTableListTop	(SkkimeConfigArg*, int) ;

static	BOOL	skkimeConfig_initRomaKanaTable		(TINYMEMMGR*, LPCMYSTR, ROMAKANATABLENODE**, ROMAKANATABLE*, int, int) ;
static	BOOL	skkimeConfig_updateRomaKanaTable	(LPCMYSTR, ROMAKANATABLENODE*) ;
static	ROMAKANATABLENODE*	skkimeConfig_newRomaKanaTableNode	(TINYMEMMGR*, int, int, int) ;
static	ROMAKANATABLENODE*	skkimeConfig_findRomaKanaTableNode	(SkkimeConfigArg*, LPCMYSTR, int) ;


/* [}Ȃ̕ϊe[uuv */
static	ROMAKANATABLE		srSkkDefaultRomaKanaA [] = {
	{ MYTEXT (""),		MYTEXT (""),  	MYTEXT ("A")   },
	{ MYTEXT ("b") ,	MYTEXT (""),  	MYTEXT ("o")   },
	{ MYTEXT ("by"),	MYTEXT ("т"),	MYTEXT ("r") },
	{ MYTEXT ("ch"),	MYTEXT (""),	MYTEXT ("`") },  
	{ MYTEXT ("cy"),	MYTEXT (""),	MYTEXT ("`") }, 
	{ MYTEXT ("d"), 	MYTEXT (""),  	MYTEXT ("_")   },
	{ MYTEXT ("dh"),	MYTEXT ("ł"),	MYTEXT ("f") },
	{ MYTEXT ("dy"),	MYTEXT (""),	MYTEXT ("a") },
	{ MYTEXT ("f"),		MYTEXT ("ӂ"),	MYTEXT ("t@") },
	{ MYTEXT ("fy"),	MYTEXT ("ӂ"),	MYTEXT ("t") },
	{ MYTEXT ("g"),		MYTEXT (""),  	MYTEXT ("K")   },
	{ MYTEXT ("gy"),	MYTEXT (""),	MYTEXT ("M") },
	{ MYTEXT ("h"),		MYTEXT (""),  	MYTEXT ("n")   },
	{ MYTEXT ("hy"),	MYTEXT ("Ђ"),	MYTEXT ("q") },
	{ MYTEXT ("j"),		MYTEXT (""),	MYTEXT ("W") },
	{ MYTEXT ("jy"),	MYTEXT (""),	MYTEXT ("W") },
	{ MYTEXT ("k"), 	MYTEXT (""),  	MYTEXT ("J")   },
	{ MYTEXT ("ky"),	MYTEXT (""),	MYTEXT ("L") },
	{ MYTEXT ("m"), 	MYTEXT (""),  	MYTEXT ("}")   },
	{ MYTEXT ("my"),	MYTEXT ("݂"),	MYTEXT ("~") },
	{ MYTEXT ("n"), 	MYTEXT (""),  	MYTEXT ("i")   },
	{ MYTEXT ("ny"),	MYTEXT ("ɂ"),	MYTEXT ("j") },
	{ MYTEXT ("p"), 	MYTEXT (""),  	MYTEXT ("p")   },
	{ MYTEXT ("py"),	MYTEXT ("҂"),	MYTEXT ("s") },
	{ MYTEXT ("r"), 	MYTEXT (""),  	MYTEXT ("")   },
	{ MYTEXT ("ry"),	MYTEXT (""),	MYTEXT ("") },
	{ MYTEXT ("s"), 	MYTEXT (""),  	MYTEXT ("T")   },
	{ MYTEXT ("sh"),	MYTEXT (""),	MYTEXT ("V") },
	{ MYTEXT ("sy"),	MYTEXT (""),	MYTEXT ("V") }, 
	{ MYTEXT ("t"), 	MYTEXT (""),  	MYTEXT ("^")   },
	{ MYTEXT ("th"),	MYTEXT ("Ă"),	MYTEXT ("e@") },
	{ MYTEXT ("ty"),	MYTEXT (""),	MYTEXT ("`") },
	{ MYTEXT ("v"), 	MYTEXT ("J"),	MYTEXT ("@") },
	{ MYTEXT ("w"), 	MYTEXT (""),  	MYTEXT ("") },
	{ MYTEXT ("x"), 	MYTEXT (""),  	MYTEXT ("@")   },
	{ MYTEXT ("xk"),	MYTEXT (""),  	MYTEXT ("") },
	{ MYTEXT ("xw"),	MYTEXT (""),  	MYTEXT ("")   },
	{ MYTEXT ("xy"),	MYTEXT (""),  	MYTEXT ("") },
	{ MYTEXT ("y"), 	MYTEXT (""),  	MYTEXT ("")   },
	{ MYTEXT ("z"), 	MYTEXT (""),  	MYTEXT ("U")   },
	{ MYTEXT ("zy"),	MYTEXT (""),	MYTEXT ("W") },
} ;

/* [}Ȃ̕ϊe[uuv */
static	ROMAKANATABLE		srSkkDefaultRomaKanaI [] = {
	{ MYTEXT (""),		MYTEXT (""),		MYTEXT ("C")   },
	{ MYTEXT ("b"),		MYTEXT (""),		MYTEXT ("r")   },
	{ MYTEXT ("by"),	MYTEXT ("т"),	MYTEXT ("rB") },
	{ MYTEXT ("ch"),	MYTEXT (""),  	MYTEXT ("`")   },
	{ MYTEXT ("cy"),	MYTEXT (""),	MYTEXT ("`B") },
	{ MYTEXT ("d"),		MYTEXT (""),		MYTEXT ("a")   },
	{ MYTEXT ("dh"),	MYTEXT ("ł"),	MYTEXT ("fB") },
	{ MYTEXT ("dy"),	MYTEXT (""),	MYTEXT ("aB") },
	{ MYTEXT ("f"),		MYTEXT ("ӂ"),	MYTEXT ("tB") },
	{ MYTEXT ("fy"),	MYTEXT ("ӂ"),	MYTEXT ("tB") },
	{ MYTEXT ("g"),		MYTEXT (""),		MYTEXT ("M")   },
	{ MYTEXT ("gy"),	MYTEXT (""),	MYTEXT ("MB") },
	{ MYTEXT ("h"),		MYTEXT (""),		MYTEXT ("q")   },
	{ MYTEXT ("hy"),	MYTEXT ("Ђ"),	MYTEXT ("qB") },
	{ MYTEXT ("j"),		MYTEXT (""),		MYTEXT ("W")   },
	{ MYTEXT ("jy"),	MYTEXT (""),	MYTEXT ("WB") },
	{ MYTEXT ("k"),		MYTEXT (""),		MYTEXT ("L")   },
	{ MYTEXT ("ky"),	MYTEXT (""),	MYTEXT ("LB") },
	{ MYTEXT ("m"),		MYTEXT (""),		MYTEXT ("~")   },
	{ MYTEXT ("my"),	MYTEXT ("݂"),	MYTEXT ("~B") },
	{ MYTEXT ("n"),		MYTEXT (""),		MYTEXT ("j")   },
	{ MYTEXT ("ny"),	MYTEXT ("ɂ"),	MYTEXT ("jB") },
	{ MYTEXT ("p"),		MYTEXT (""),		MYTEXT ("s")   },
	{ MYTEXT ("py"),	MYTEXT ("҂"),	MYTEXT ("sB") },
	{ MYTEXT ("r"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("ry"),	MYTEXT ("股"),	MYTEXT ("B") },
	{ MYTEXT ("s"),		MYTEXT (""),		MYTEXT ("V")   },
	{ MYTEXT ("sh"),	MYTEXT (""),		MYTEXT ("V")   },
	{ MYTEXT ("sy"),	MYTEXT (""),	MYTEXT ("VB") },
	{ MYTEXT ("t"),		MYTEXT (""),		MYTEXT ("`")   },
	{ MYTEXT ("th"),	MYTEXT ("Ă"),	MYTEXT ("eB") },
	{ MYTEXT ("ty"),	MYTEXT (""),	MYTEXT ("`B") },
	{ MYTEXT ("v"),		MYTEXT ("J"),	MYTEXT ("B") },
	{ MYTEXT ("w"),		MYTEXT (""),	MYTEXT ("EB") },
	{ MYTEXT ("x"),		MYTEXT (""),		MYTEXT ("B")   },
	{ MYTEXT ("xw"),	MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("z"),		MYTEXT (""),		MYTEXT ("W")   },
	{ MYTEXT ("zy"),	MYTEXT (""),	MYTEXT ("WB") },
} ;

/* [}Ȃ̕ϊe[uuv */
static	ROMAKANATABLE		srSkkDefaultRomaKanaU [] = {
	{ MYTEXT (""),		MYTEXT (""),		MYTEXT ("E")   },
	{ MYTEXT ("b"),		MYTEXT (""),		MYTEXT ("u")   },
	{ MYTEXT ("by"),	MYTEXT ("т"),	MYTEXT ("r") },
	{ MYTEXT ("ch"),	MYTEXT (""),	MYTEXT ("`") },
	{ MYTEXT ("cy"),	MYTEXT (""),	MYTEXT ("`") },
	{ MYTEXT ("d"),		MYTEXT (""),		MYTEXT ("d")   },
	{ MYTEXT ("dh"),	MYTEXT ("ł"),	MYTEXT ("f") },
	{ MYTEXT ("dy"),	MYTEXT (""),	MYTEXT ("a") },
	{ MYTEXT ("f"),		MYTEXT (""),		MYTEXT ("t")   },
	{ MYTEXT ("fy"),	MYTEXT ("ӂ"),	MYTEXT ("t") },
	{ MYTEXT ("g"),		MYTEXT (""),		MYTEXT ("O")   },
	{ MYTEXT ("gy"),	MYTEXT (""),	MYTEXT ("M") },
	{ MYTEXT ("h"),		MYTEXT (""),		MYTEXT ("t")   },
	{ MYTEXT ("hy"),	MYTEXT ("Ђ"),	MYTEXT ("q") },
	{ MYTEXT ("j"),		MYTEXT (""),	MYTEXT ("W") },
	{ MYTEXT ("jy"),	MYTEXT (""),	MYTEXT ("W") },
	{ MYTEXT ("k"),		MYTEXT (""),		MYTEXT ("N")   },
	{ MYTEXT ("ky"),	MYTEXT (""),	MYTEXT ("L") },
	{ MYTEXT ("m"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("my"),	MYTEXT ("݂"),	MYTEXT ("~") },
	{ MYTEXT ("n"),		MYTEXT (""),		MYTEXT ("k")   },
	{ MYTEXT ("ny"),	MYTEXT ("ɂ"),	MYTEXT ("j") },
	{ MYTEXT ("p"),		MYTEXT (""),		MYTEXT ("v")   },
	{ MYTEXT ("py"),	MYTEXT ("҂"),	MYTEXT ("s") },
	{ MYTEXT ("r"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("ry"),	MYTEXT (""),	MYTEXT ("") },
	{ MYTEXT ("s"),		MYTEXT (""),		MYTEXT ("X")   },
	{ MYTEXT ("sh"),	MYTEXT (""),	MYTEXT ("V") },
	{ MYTEXT ("sy"),	MYTEXT (""),	MYTEXT ("V") },
	{ MYTEXT ("t"),		MYTEXT (""),		MYTEXT ("c")   },
	{ MYTEXT ("th"),	MYTEXT ("Ă"),	MYTEXT ("e") },
	{ MYTEXT ("ts"),	MYTEXT (""),		MYTEXT ("c")   },
	{ MYTEXT ("ty"),	MYTEXT (""),	MYTEXT ("`") },
	{ MYTEXT ("v"),		MYTEXT ("J"),	MYTEXT ("")   },
	{ MYTEXT ("w"),		MYTEXT (""),		MYTEXT ("E")   },
	{ MYTEXT ("x"),		MYTEXT (""),		MYTEXT ("D")   },
	{ MYTEXT ("xt"),	MYTEXT (""),		MYTEXT ("b")   },
	{ MYTEXT ("xts"),	MYTEXT (""),		MYTEXT ("b")   },
	{ MYTEXT ("xy"),	MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("y"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("z"),		MYTEXT (""),		MYTEXT ("Y")   },
	{ MYTEXT ("zy"),	MYTEXT (""),	MYTEXT ("W") },
} ;

/* [}Ȃ̕ϊe[uuv */
static	ROMAKANATABLE		srSkkDefaultRomaKanaE [] = {
	{ MYTEXT (""),		MYTEXT (""),		MYTEXT ("G")   },
	{ MYTEXT ("b"),		MYTEXT (""),		MYTEXT ("x")   },
	{ MYTEXT ("by"),	MYTEXT ("т"),	MYTEXT ("rF") },
	{ MYTEXT ("ch"),	MYTEXT (""),	MYTEXT ("`F") },
	{ MYTEXT ("cy"),	MYTEXT (""),	MYTEXT ("`F") },
	{ MYTEXT ("d"),		MYTEXT (""),		MYTEXT ("f")   },
	{ MYTEXT ("dh"),	MYTEXT ("ł"),	MYTEXT ("fF") },
	{ MYTEXT ("dy"),	MYTEXT (""),	MYTEXT ("aF") },
	{ MYTEXT ("f"),		MYTEXT ("ӂ"),	MYTEXT ("tF") },
	{ MYTEXT ("fy"),	MYTEXT ("ӂ"),	MYTEXT ("tF") },
	{ MYTEXT ("g"),		MYTEXT (""),		MYTEXT ("Q")   },
	{ MYTEXT ("gy"),	MYTEXT (""),	MYTEXT ("MF") },
	{ MYTEXT ("h"),		MYTEXT (""),		MYTEXT ("w")   },
	{ MYTEXT ("hy"),	MYTEXT ("Ђ"),	MYTEXT ("qF") },
	{ MYTEXT ("j"),		MYTEXT (""),	MYTEXT ("WF") },
	{ MYTEXT ("jy"),	MYTEXT (""),	MYTEXT ("WF") },
	{ MYTEXT ("k"),		MYTEXT (""),		MYTEXT ("P")   },
	{ MYTEXT ("ky"),	MYTEXT (""),	MYTEXT ("LF") },
	{ MYTEXT ("m"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("my"),	MYTEXT ("݂"),	MYTEXT ("~F") },
	{ MYTEXT ("n"),		MYTEXT (""),		MYTEXT ("l")   },
	{ MYTEXT ("ny"),	MYTEXT ("ɂ"),	MYTEXT ("jF") },
	{ MYTEXT ("p"),		MYTEXT (""),		MYTEXT ("y")   },
	{ MYTEXT ("py"),	MYTEXT ("҂"),	MYTEXT ("sF") },
	{ MYTEXT ("r"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("ry"),	MYTEXT ("肥"),	MYTEXT ("F") },
	{ MYTEXT ("s"),		MYTEXT (""),		MYTEXT ("Z")   },
	{ MYTEXT ("sh"),	MYTEXT (""),	MYTEXT ("VF") },
	{ MYTEXT ("sy"),	MYTEXT (""),	MYTEXT ("VF") },
	{ MYTEXT ("t"),		MYTEXT (""),		MYTEXT ("e")   },
	{ MYTEXT ("th"),	MYTEXT ("Ă"),	MYTEXT ("eF") },
	{ MYTEXT ("ty"),	MYTEXT (""),	MYTEXT ("`F") },
	{ MYTEXT ("v"),		MYTEXT ("J"),	MYTEXT ("F") },
	{ MYTEXT ("w"),		MYTEXT (""),	MYTEXT ("EF") },
	{ MYTEXT ("x"),		MYTEXT (""),		MYTEXT ("F")   },
	{ MYTEXT ("xk"),	MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("xw"),	MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("y"),		MYTEXT (""),	MYTEXT ("CF") },
	{ MYTEXT ("z"),		MYTEXT (""),		MYTEXT ("[")   },
	{ MYTEXT ("zy"),	MYTEXT (""),	MYTEXT ("WF") },
} ;

/* [}Ȃ̕ϊe[uuv */
static	ROMAKANATABLE		srSkkDefaultRomaKanaO [] = {
	{ MYTEXT (""),		MYTEXT (""),		MYTEXT ("I")   },
	{ MYTEXT ("b"),		MYTEXT (""),		MYTEXT ("{")   },
	{ MYTEXT ("by"),	MYTEXT ("т"),	MYTEXT ("r") },
	{ MYTEXT ("ch"),	MYTEXT (""),	MYTEXT ("`") },
	{ MYTEXT ("cy"),	MYTEXT (""),	MYTEXT ("`") },
	{ MYTEXT ("d"),		MYTEXT (""),		MYTEXT ("h")   },
	{ MYTEXT ("dh"),	MYTEXT ("ł"),	MYTEXT ("f") },
	{ MYTEXT ("dy"),	MYTEXT (""),	MYTEXT ("a") },
	{ MYTEXT ("f"),		MYTEXT ("ӂ"),	MYTEXT ("tH") },
	{ MYTEXT ("fy"),	MYTEXT ("ӂ"),	MYTEXT ("t") },
	{ MYTEXT ("g"),		MYTEXT (""),		MYTEXT ("S")   },
	{ MYTEXT ("gy"),	MYTEXT (""),	MYTEXT ("M") },
	{ MYTEXT ("h"),		MYTEXT (""),		MYTEXT ("z")   },
	{ MYTEXT ("hy"),	MYTEXT ("Ђ"),	MYTEXT ("q") },
	{ MYTEXT ("j"),		MYTEXT (""),	MYTEXT ("W") },
	{ MYTEXT ("jy"),	MYTEXT (""),	MYTEXT ("W") },
	{ MYTEXT ("k"),		MYTEXT (""),		MYTEXT ("R")   },
	{ MYTEXT ("ky"),	MYTEXT (""),	MYTEXT ("L") },
	{ MYTEXT ("m"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("my"),	MYTEXT ("݂"),	MYTEXT ("~") },
	{ MYTEXT ("n"),		MYTEXT (""),		MYTEXT ("m")   },
	{ MYTEXT ("ny"),	MYTEXT ("ɂ"),	MYTEXT ("j") },
	{ MYTEXT ("p"),		MYTEXT (""),		MYTEXT ("|")   },
	{ MYTEXT ("py"),	MYTEXT ("҂"),	MYTEXT ("s") },
	{ MYTEXT ("r"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("ry"),	MYTEXT (""),	MYTEXT ("") },
	{ MYTEXT ("s"),		MYTEXT (""),		MYTEXT ("\")   },
	{ MYTEXT ("sh"),	MYTEXT (""),	MYTEXT ("V") },
	{ MYTEXT ("sy"),	MYTEXT (""),	MYTEXT ("V") },
	{ MYTEXT ("t"),		MYTEXT (""),		MYTEXT ("g")   },
	{ MYTEXT ("th"),	MYTEXT ("Ă"),	MYTEXT ("e") },
	{ MYTEXT ("ty"),	MYTEXT (""),	MYTEXT ("`") },
	{ MYTEXT ("v"),		MYTEXT ("J"),	MYTEXT ("H") },
	{ MYTEXT ("w"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("x"),		MYTEXT (""),		MYTEXT ("H")   },
	{ MYTEXT ("xy"),	MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("y"),		MYTEXT (""),		MYTEXT ("")   },
	{ MYTEXT ("z"),		MYTEXT (""),		MYTEXT ("]")   },
	{ MYTEXT ("zy"),	MYTEXT (""),	MYTEXT ("W") },
} ;

BOOL
skkimeConfig_InitializeRomaKanaTableA (
	register TINYMEMMGR*			pMemMgr,
	register ROMAKANATABLENODE**	plstRomaKanaA)
{
	if (pMemMgr == NULL || plstRomaKanaA == NULL)
		return	FALSE ;
	return	skkimeConfig_initRomaKanaTable (pMemMgr, REGPATH_ROMAKANATABLE_A, plstRomaKanaA, srSkkDefaultRomaKanaA, ARRAYSIZE (srSkkDefaultRomaKanaA), MYTEXT ('a')) ;
}

BOOL
skkimeConfig_InitializeRomaKanaTableI (
	register TINYMEMMGR*			pMemMgr,
	register ROMAKANATABLENODE**	plstRomaKanaI)
{
	if (pMemMgr == NULL || plstRomaKanaI == NULL)
		return	FALSE ;
	return	skkimeConfig_initRomaKanaTable (pMemMgr, REGPATH_ROMAKANATABLE_I, plstRomaKanaI, srSkkDefaultRomaKanaI, ARRAYSIZE (srSkkDefaultRomaKanaI), MYTEXT('i')) ;
}

BOOL
skkimeConfig_InitializeRomaKanaTableU (
	register TINYMEMMGR*			pMemMgr,
	register ROMAKANATABLENODE**	plstRomaKanaU)
{
	if (pMemMgr == NULL || plstRomaKanaU == NULL)
		return	FALSE ;
	return	skkimeConfig_initRomaKanaTable (pMemMgr, REGPATH_ROMAKANATABLE_U, plstRomaKanaU, srSkkDefaultRomaKanaU, ARRAYSIZE (srSkkDefaultRomaKanaU), MYTEXT('u')) ;
}

BOOL
skkimeConfig_InitializeRomaKanaTableE (
	register TINYMEMMGR*			pMemMgr,
	register ROMAKANATABLENODE**	plstRomaKanaE)
{
	if (pMemMgr == NULL || plstRomaKanaE == NULL)
		return	FALSE ;
	return	skkimeConfig_initRomaKanaTable (pMemMgr, REGPATH_ROMAKANATABLE_E, plstRomaKanaE, srSkkDefaultRomaKanaE, ARRAYSIZE (srSkkDefaultRomaKanaE), MYTEXT('e')) ;
}

BOOL
skkimeConfig_InitializeRomaKanaTableO (
	register TINYMEMMGR*			pMemMgr,
	register ROMAKANATABLENODE**	plstRomaKanaO)
{
	if (pMemMgr == NULL || plstRomaKanaO == NULL)
		return	FALSE ;
	return	skkimeConfig_initRomaKanaTable (pMemMgr, REGPATH_ROMAKANATABLE_O, plstRomaKanaO, srSkkDefaultRomaKanaO, ARRAYSIZE (srSkkDefaultRomaKanaO), MYTEXT('o')) ;
}

BOOL
skkimeConfig_UpdateRomaKanaTableA (
	register ROMAKANATABLENODE*		lstRomaKanaA)
{
	if (lstRomaKanaA == NULL)
		return	FALSE ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateRomaKanaTableA (%p)\n"), lstRomaKanaA)) ;
	return	skkimeConfig_updateRomaKanaTable (REGPATH_ROMAKANATABLE_A, lstRomaKanaA) ;
}

BOOL
skkimeConfig_UpdateRomaKanaTableI (
	register ROMAKANATABLENODE*		lstRomaKanaI)
{
	if (lstRomaKanaI == NULL)
		return	FALSE ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateRomaKanaTableI (%p)\n"), lstRomaKanaI)) ;
	return	skkimeConfig_updateRomaKanaTable (REGPATH_ROMAKANATABLE_I, lstRomaKanaI) ;
}

BOOL
skkimeConfig_UpdateRomaKanaTableU (
	register ROMAKANATABLENODE*		lstRomaKanaU)
{
	if (lstRomaKanaU == NULL)
		return	FALSE ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateRomaKanaTableU (%p)\n"), lstRomaKanaU)) ;
	return	skkimeConfig_updateRomaKanaTable (REGPATH_ROMAKANATABLE_U, lstRomaKanaU) ;
}

BOOL
skkimeConfig_UpdateRomaKanaTableE (
	register ROMAKANATABLENODE*		lstRomaKanaE)
{
	if (lstRomaKanaE == NULL)
		return	FALSE ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateRomaKanaTableE (%p)\n"), lstRomaKanaE)) ;
	return	skkimeConfig_updateRomaKanaTable (REGPATH_ROMAKANATABLE_E, lstRomaKanaE) ;
}

BOOL
skkimeConfig_UpdateRomaKanaTableO (
	register ROMAKANATABLENODE*		lstRomaKanaO)
{
	if (lstRomaKanaO == NULL)
		return	FALSE ;

	DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateRomaKanaTableO (%p)\n"), lstRomaKanaO)) ;
	return	skkimeConfig_updateRomaKanaTable (REGPATH_ROMAKANATABLE_O, lstRomaKanaO) ;
}

INT_PTR	CALLBACK
skkimeConfig_dlgRomaKanaTblProc (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	switch (uMsg) {
	case	WM_INITDIALOG:
		return	skkimeConfig_dlgRomaKanaTblOnInitDialog (hDlg, uMsg, wParam, lParam) ;
	case	WM_COMMAND:
		return	skkimeConfig_dlgRomaKanaTblOnCommand (hDlg, uMsg, wParam, lParam) ;
	case	WM_NOTIFY:
		return	skkimeConfig_dlgRomaKanaTblOnNotify (hDlg, uMsg, wParam, lParam) ;
	default:
		break ;
	}
	return	FALSE ;
}

/*========================================================================
 *	private functions
 */
INT_PTR
skkimeConfig_dlgRomaKanaTblOnInitDialog (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register HWND				hwndRomaKanaTblList, hwndRadio ;

	SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;
	pPropSheet	= (LPPROPSHEETPAGE) lParam ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	hwndRomaKanaTblList	= GetDlgItem (hDlg, IDC_LIST_ROMAKANATABLE) ;
	if (hwndRomaKanaTblList == NULL)
		return	TRUE ;
	MyListView_SetExtendedListViewStyle (hwndRomaKanaTblList, LVS_EX_FULLROWSELECT) ;
	skkimeConfig_initControlRomaKanaTableList (pConfArg, hwndRomaKanaTblList, FALSE) ;

	hwndRadio	= GetDlgItem (hDlg, IDC_RADIO_SHOW_ROMAKANATABLE_WITH_KANA) ;
	if (hwndRadio == NULL)
		return	TRUE ;
	SendMessage (hwndRadio, BM_SETCHECK, BST_CHECKED, 0) ;
	return	TRUE ;

	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

INT_PTR
skkimeConfig_dlgRomaKanaTblOnCommand (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register WORD	wNotifyCode ;
	register WORD	wID ;
	register HWND	hwndCtl ;
	register HWND	hwndLV ;

	wNotifyCode	= HIWORD (wParam) ;
	wID			= LOWORD (wParam) ;
	hwndCtl		= (HWND) lParam ;

	DEBUGPRINTF ((TEXT ("WM_COMMAND: (wID(%d), wNotifyCode(%d))\n"),
				  wID, wNotifyCode)) ;

	switch (wID) {
	case	IDC_RADIO_SHOW_ROMAKANATABLE_WITH_KANA:
	case	IDC_RADIO_SHOW_ROMAKANATABLE_WITH_KATA:
	{
		hwndLV	= GetDlgItem (hDlg, IDC_LIST_ROMAKANATABLE) ;
		if (hwndLV == NULL)
			return	TRUE ;	/* Message Ȃ 1 ԂB*/
		skkimeConfig_switchKanaKata (hwndLV, (wID == IDC_RADIO_SHOW_ROMAKANATABLE_WITH_KATA)) ;
		return	FALSE ;	/* Message ꂽ 0 ԂB*/
	}
	case	IDC_EDIT_SEARCH:
	{
		if (wNotifyCode != EN_CHANGE)
			return	TRUE ;

		wNotifyCode = BN_CLICKED ;
		/*	fall through */
	}
	case	IDC_BUTTON_SEARCH_NEXT:
	{
		MYLVFINDINFO				lvfi ;
		register HWND				hwndEdit ;
		register int				nItem, nStart ;
		MYCHAR						szBuffer [MAX_LENGTH_ROMAKANAPREFIX] ;

		if (wNotifyCode != BN_CLICKED)
			return	TRUE ;
		hwndEdit	= GetDlgItem (hDlg, IDC_EDIT_SEARCH) ;
		hwndLV		= GetDlgItem (hDlg, IDC_LIST_ROMAKANATABLE) ;
		if (hwndEdit == NULL || hwndLV == NULL)
			return	TRUE ;	/* Message Ȃ 1 ԂB*/

		MyGetWindowText (hwndEdit, szBuffer, MAX_LENGTH_ROMAKANAPREFIX) ;
		szBuffer [MAX_LENGTH_ROMAKANAPREFIX - 1]	= MYTEXT ('\0') ;
		if (wID == IDC_EDIT_SEARCH) {
			nStart	= 0 ;
		} else {
			nStart	= MyListView_GetSelectionMark (hwndLV) ;
			if (nStart < 0)
				nStart	= 0 ;
		}
		lvfi.psz	= szBuffer ;
		lvfi.flags	= LVFI_PARTIAL ;
		nItem		= MyListView_FindItem (hwndLV, nStart, &lvfi) ;
		if ((nItem == nStart || nItem < 0) && nStart != 0)
			nItem	= MyListView_FindItem (hwndLV, -1, &lvfi) ;
		if (nItem >= 0) {
			MyListView_SetSelectionMark (hwndLV, nItem) ;
			MyListView_SetItemState (hwndLV, nItem, LVIS_FOCUSED | LVIS_SELECTED, LVIS_FOCUSED | LVIS_SELECTED) ;
		}
		return	FALSE ;
	}
	case	IDC_BUTTON_EDIT:
	case	IDC_BUTTON_INSERT:
	case	IDC_BUTTON_DELETE:
	{
		register int	nItem ;

		if (wNotifyCode != BN_CLICKED) 
			return	TRUE ;

		if (wID != IDC_BUTTON_INSERT) {
			hwndLV	= GetDlgItem (hDlg, IDC_LIST_ROMAKANATABLE) ;
			if (hwndLV == NULL)
				return	TRUE ;	/* Message Ȃ 1 ԂB*/
			nItem	= MyListView_GetSelectionMark (hwndLV) ;
		} else {
			nItem	= -1 ;
		}
		skkimeConfig_editRomaKanaTable (hDlg, nItem, wID) ;
		return	FALSE ;
	}
	default:
		break ;
	}
	return 0 ;
	UNREFERENCED_PARAMETER (uMsg) ;
}

INT_PTR
skkimeConfig_dlgRomaKanaTblOnNotify (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register NMHDR	FAR*		pNMHDR	= (NMHDR *) lParam ;
	static TMYMENUITEM		rmi []	= {
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ǉ"), IDC_BUTTON_INSERT, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("ύX"), IDC_BUTTON_EDIT, },
		{ MIIM_ID | MIIM_STRING, 0,  TEXT("폜"), IDC_BUTTON_DELETE, },
	} ;

	pPropSheet	= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg	= (SkkimeConfigArg *) pPropSheet->lParam ;

	switch (pNMHDR->idFrom) {
	case	IDC_LIST_ROMAKANATABLE:
	{
		if (pNMHDR->code == NM_RCLICK) {
			register NMITEMACTIVATE*	pNM 	= (NMITEMACTIVATE*) lParam ;
			register int	n ;

			n	= skkimeConfig_PopupMenu (hDlg, rmi, ARRAYSIZE (rmi)) ;
			DEBUGPRINTF ((TEXT ("NM: item(%d), lParam(%d)\n"), pNM->iItem, pNM->lParam)) ;
			if (n > 0) 
				skkimeConfig_editRomaKanaTable (hDlg, ((n != IDC_BUTTON_INSERT)? pNM->iItem : -1), n) ;
			DEBUGPRINTF ((TEXT ("TrackPopupMenu returns %d\n"), n)) ;
		}
		return	FALSE ;	/* WM_NOTIFY  Return Value  ignore ̂
						 * ԂĂǂB*/
	}
	default:
		switch (pNMHDR->code){
		case PSN_SETACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, 0) ;
			return	TRUE ;

		case PSN_KILLACTIVE:
			SetWindowLongPtr (hDlg, DWLP_MSGRESULT, FALSE) ;
			return	TRUE ;

		case PSN_APPLY:
		{
			if (skkimeConfig_UpdateRomaKanaTableA (pConfArg->m_lstRomaKanaA) &&
				skkimeConfig_UpdateRomaKanaTableI (pConfArg->m_lstRomaKanaI) &&
				skkimeConfig_UpdateRomaKanaTableU (pConfArg->m_lstRomaKanaU) &&
				skkimeConfig_UpdateRomaKanaTableE (pConfArg->m_lstRomaKanaE) &&
				skkimeConfig_UpdateRomaKanaTableO (pConfArg->m_lstRomaKanaO)) {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_NOERROR) ;
			} else {
				SetWindowLongPtr (hDlg, DWLP_MSGRESULT, PSNRET_INVALID) ;
			}
			return	TRUE ;
		}
		case PSN_RESET:
		case PSN_HELP:
			break ;
		default:
			return FALSE ;
		}
		return	TRUE ;
	}

	UNREFERENCED_PARAMETER (uMsg) ;
	UNREFERENCED_PARAMETER (wParam) ;
}

BOOL
skkimeConfig_editRomaKanaTable (
	register HWND	hDlg,
	register int	nItem,
	register int	nCommand)
{
	register LPPROPSHEETPAGE	pPropSheet ;
	register SkkimeConfigArg*	pConfArg ;
	register HWND				hwndLV ;
	register TINYMEMMGR*		pMemMgr ;
	MYLVFINDINFO				lvfi ;
	MYLVITEM					lvI ;
	register int				nRetval ;
	register int				nPrefixLen, nKanaLen, nKataLen ;
	register int				nVowel, nSpace ;
	register ROMAKANATABLENODE*	pNode ;
	register ROMAKANATABLENODE*	pNewNode ;
	register BOOL				fKatakana ;
	TDlgEditRomaKanaTableArg	arg ;
	MYCHAR						rszBuffer [256 + 16] ;

	pPropSheet		= (LPPROPSHEETPAGE) GetWindowLongPtr (hDlg, DWLP_USER) ;
	pConfArg		= (SkkimeConfigArg *) pPropSheet->lParam ;
	hwndLV			= GetDlgItem (hDlg, IDC_LIST_ROMAKANATABLE) ;
	if (hwndLV == NULL || pConfArg == NULL)
		return	FALSE ;
	pMemMgr			= &pConfArg->m_MemMgr ;

	lvI.mask		= LVIF_PARAM ;
	lvI.state		= 0 ;
	lvI.stateMask	= 0 ;
	lvI.iItem		= nItem ;
	lvI.iSubItem	= 0 ;
	lvI.pszText		= 0 ;
	lvI.cchTextMax	= 0 ;
	if (! MyListView_GetItem (hwndLV, &lvI)) {
		pNode		= NULL ;
	} else {
		pNode		= (ROMAKANATABLENODE*) lvI.lParam ;
	}

	switch (nCommand) {
	case	IDC_BUTTON_EDIT:
	case	IDC_BUTTON_INSERT:
		arg.m_strPrefix   [0]	= MYTEXT ('\0') ;
		arg.m_strKana     [0]	= MYTEXT ('\0') ;
		arg.m_strKatakana [0]	= MYTEXT ('\0') ;
		if (pNode != NULL && pNode->m_strPrefix != NULL) {
			Mylstrcpyn (arg.m_strPrefix,   pNode->m_strPrefix, MAX_LENGTH_ROMAKANAPREFIX - 1) ;
			arg.m_strPrefix [MAX_LENGTH_ROMAKANAPREFIX - 2]	= MYTEXT ('\0') ;
			nPrefixLen	= Mylstrlen (arg.m_strPrefix) ;
			arg.m_strPrefix [nPrefixLen ++]	= (MYCHAR) pNode->m_chVowel ;
			arg.m_strPrefix [nPrefixLen]	= MYTEXT ('\0') ;

			if (pNode->m_strKana != NULL) {
				Mylstrcpyn (arg.m_strKana,     pNode->m_strKana,   MAX_LENGTH_ROMAKANAKANA) ;
				arg.m_strKana [MAX_LENGTH_ROMAKANAKANA - 1]	= MYTEXT ('\0') ;
			}
			if (pNode->m_strKatakana != NULL) {
				Mylstrcpyn (arg.m_strKatakana, pNode->m_strKatakana, MAX_LENGTH_ROMAKANAKATA) ;
				arg.m_strKatakana [MAX_LENGTH_ROMAKANAKATA - 1]	= MYTEXT ('\0') ;
			}
		}
		nRetval	= DialogBoxParam (pConfArg->m_hInst, MAKEINTRESOURCE (IDD_DIALOG_EDIT_ROMAKANATBL), hDlg, skkimeConfig_dlgEditRomaKanaTbl, (LPARAM) &arg) ;
		if (nRetval == IDOK) {
			nPrefixLen	= Mylstrlen (arg.m_strPrefix) ;
			if (nPrefixLen <= 1) {
				DEBUGPRINTF ((MYTEXT ("Fatal: The length of prefix(%d) is too short.\n"),
							  nPrefixLen)) ;
				return	FALSE ;
			}
			nVowel	= arg.m_strPrefix [nPrefixLen - 1] ;
			if (nVowel != MYTEXT ('a') && nVowel != MYTEXT ('i') &&
				nVowel != MYTEXT ('u') && nVowel != MYTEXT ('e') &&
				nVowel != MYTEXT ('o')) {
				DEBUGPRINTF ((MYTEXT ("Fatal: Vowel(%c) is not acceptable.\n"), nVowel)) ;
				return	FALSE ;
			}
			arg.m_strPrefix [nPrefixLen - 1]	= MYTEXT ('\0') ;
			nKanaLen	= Mylstrlen (arg.m_strKana) + 1 ;
			nKataLen	= Mylstrlen (arg.m_strKatakana) + 1 ;
			pNewNode	= skkimeConfig_newRomaKanaTableNode (pMemMgr, nPrefixLen, nKanaLen, nKataLen) ;
			if (pNewNode == NULL) {
				DEBUGPRINTF ((MYTEXT ("Fatal: Failed in creating new romakanatable-node,\n"))) ;
				return	FALSE ;
			}
			Mylstrcpy (pNewNode->m_strPrefix,   arg.m_strPrefix) ;
			Mylstrcpy (pNewNode->m_strKana,     arg.m_strKana) ;
			Mylstrcpy (pNewNode->m_strKatakana, arg.m_strKatakana) ;
			pNewNode->m_chVowel	= nVowel ;

			lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
			lvI.state		= 0 ; 
			lvI.stateMask	= 0 ;
			lvI.iImage		= 0 ;
			lvI.iSubItem	= 0 ;
			lvI.lParam		= (LPARAM) pNewNode ;
			fKatakana		= (IsDlgButtonChecked (hDlg, IDC_RADIO_SHOW_ROMAKANATABLE_WITH_KATA) == BST_CHECKED) ;
			nSpace			= Mylstrlen (pNewNode->m_strPrefix) ;
			nSpace			= (nSpace > 4)? 4 : nSpace ;
			Mywsprintf (rszBuffer, MYTEXT ("%s%c%s = %s"),
						pNewNode->m_strPrefix,
						pNewNode->m_chVowel,
						MYTEXT("    ") + nSpace,
						(fKatakana)? pNewNode->m_strKatakana : pNewNode->m_strKana) ;
			lvI.pszText		= rszBuffer ;

			pNode		= skkimeConfig_findRomaKanaTableNode (pConfArg, arg.m_strPrefix, nVowel) ;
			if (! skkimeConfig_registerRomaKanaTableNode (pConfArg, pNewNode)) {
				DEBUGPRINTF ((MYTEXT ("Fatal: Failed in registering new romakanatable-node,\n"))) ;
				TinyMalloc_Free (pMemMgr, pNewNode) ;	/* fatal */
				return	FALSE ;
			}

			if (pNode != NULL) {
				DEBUGPRINTF ((MYTEXT ("Change romakanatable-node(%p->%p)\n"),
							  pNode, pNewNode)) ;
				lvfi.flags		= LVFI_PARAM ;
				lvfi.lParam		= (LPARAM) pNode ;
				nItem			= MyListView_FindItem (hwndLV, -1, &lvfi) ;
				if (nItem >= 0) {
					lvI.iItem	= nItem ;
					MyListView_SetItem (hwndLV, &lvI) ;
				}
				skkimeConfig_unregisterRomaKanaTableNode (pConfArg, pNode) ;
				TinyMalloc_Free (pMemMgr, pNode) ;
			} else {
				lvI.iItem		= MyListView_GetItemCount (hwndLV) ;
				DEBUGPRINTF ((MYTEXT ("Insert new romakanatable-node(%p/%d)\n"),
							  pNewNode, lvI.iItem)) ;
				MyListView_InsertItem (hwndLV, &lvI) ;
				MyListView_SortItems  (hwndLV, skkimeConfig_compareRomaKanaTblNode, 0) ;
			}
			PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		}
		break ;

	case	IDC_BUTTON_DELETE:
		if (pNode == NULL || nItem < 0)
			return	FALSE ;
		MyListView_DeleteItem (hwndLV, nItem) ;
		skkimeConfig_unregisterRomaKanaTableNode (pConfArg, pNode) ;
		TinyMalloc_Free (pMemMgr, pNode) ;
		PropSheet_Changed (pConfArg->m_hPropSheetDlg, hDlg) ;
		break ;
	default:
		return	FALSE ;
	}
	return	TRUE ;
}

INT_PTR CALLBACK
skkimeConfig_dlgEditRomaKanaTbl (
	register HWND		hDlg,
	register UINT		uMsg,
	register WPARAM		wParam,
	register LPARAM		lParam)
{
	register TDlgEditRomaKanaTableArg*	pArg ;
	register HWND	hwndEditPrefix, hwndEditHira, hwndEditKata ;

	switch (uMsg) {
	case	WM_INITDIALOG:
	{
		SetWindowLongPtr (hDlg, DWLP_USER, (LONG_PTR) lParam) ;

		pArg			= (TDlgEditRomaKanaTableArg*) lParam ;
		hwndEditPrefix	= GetDlgItem (hDlg, IDC_EDIT_KANAPREFIX) ;
		hwndEditHira	= GetDlgItem (hDlg, IDC_EDIT_HIRAGANA) ;
		hwndEditKata	= GetDlgItem (hDlg, IDC_EDIT_KATAKANA) ;
		if (pArg != NULL) {
		}
		if (hwndEditPrefix != NULL) {
			MySendMessage (hwndEditPrefix, EM_LIMITTEXT, (WPARAM) MAX_LENGTH_ROMAKANAPREFIX - 1, 0) ;
			if (pArg != NULL)
				MySetWindowText (hwndEditPrefix, pArg->m_strPrefix) ;
		}
		if (hwndEditHira != NULL) {
			MySendMessage (hwndEditHira,   EM_LIMITTEXT, (WPARAM) MAX_LENGTH_ROMAKANAKANA - 1, 0) ;
			if (pArg != NULL)
				MySetWindowText (hwndEditHira, pArg->m_strKana) ;
		}
		if (hwndEditKata != NULL) {
			MySendMessage (hwndEditKata,   EM_LIMITTEXT, (WPARAM) MAX_LENGTH_ROMAKANAKATA - 1, 0) ;
			if (pArg != NULL)
				MySetWindowText (hwndEditKata, pArg->m_strKatakana) ;
		}
		return	TRUE ;
	}
	case	WM_COMMAND:
		/*	[ށBԂl global ϐoR邵Ȃ... */
		switch (LOWORD (wParam)) {
		case	IDOK:
		{
			/*	ݑIĂ item 𓾂BCB_ERR ̏ꍇ undefined 
			 *	Ǝv...ȁB*/
			pArg	= (TDlgEditRomaKanaTableArg *)GetWindowLongPtr (hDlg, DWLP_USER) ;
			if (pArg != NULL) {
				hwndEditPrefix	= GetDlgItem (hDlg, IDC_EDIT_KANAPREFIX) ;
				hwndEditHira	= GetDlgItem (hDlg, IDC_EDIT_HIRAGANA) ;
				hwndEditKata	= GetDlgItem (hDlg, IDC_EDIT_KATAKANA) ;
				pArg->m_strPrefix   [0]	= MYTEXT ('\0') ;
				pArg->m_strKana     [0]	= MYTEXT ('\0') ;
				pArg->m_strKatakana [0]	= MYTEXT ('\0') ;
				if (hwndEditPrefix != NULL)
					MyGetWindowText (hwndEditPrefix, pArg->m_strPrefix,   MAX_LENGTH_ROMAKANAPREFIX) ;
				if (hwndEditHira != NULL)
					MyGetWindowText (hwndEditHira,   pArg->m_strKana,     MAX_LENGTH_ROMAKANAKANA) ;
				if (hwndEditKata != NULL)
					MyGetWindowText (hwndEditKata,   pArg->m_strKatakana, MAX_LENGTH_ROMAKANAKATA) ;
				if (pArg->m_strPrefix [0] == MYTEXT ('\0')) 
					wParam					= IDCANCEL ;
			}
			/*	fall through */
		}
		case	IDCANCEL:
			EndDialog (hDlg, wParam) ;
			return	TRUE ;
		}
	default:
		break ;
	}
	return	FALSE ;
}

BOOL
skkimeConfig_initControlRomaKanaTableList (
	register SkkimeConfigArg*	pConfArg,
	register HWND				hwndLV,
	register BOOL				fKata)
{
	MyListView_DeleteAllItems (hwndLV) ;
	skkimeConfig_initControlRomaKanaOneTable (hwndLV, fKata, pConfArg->m_lstRomaKanaA) ;
	skkimeConfig_initControlRomaKanaOneTable (hwndLV, fKata, pConfArg->m_lstRomaKanaI) ;
	skkimeConfig_initControlRomaKanaOneTable (hwndLV, fKata, pConfArg->m_lstRomaKanaU) ;
	skkimeConfig_initControlRomaKanaOneTable (hwndLV, fKata, pConfArg->m_lstRomaKanaE) ;
	skkimeConfig_initControlRomaKanaOneTable (hwndLV, fKata, pConfArg->m_lstRomaKanaO) ;
	return	TRUE ;
}

BOOL
skkimeConfig_initControlRomaKanaOneTable (
	register HWND				hwndLV,
	register BOOL				fKatakana,
	register ROMAKANATABLENODE*	lstRomaKanaTbl)
{
	register ROMAKANATABLENODE*	pNode ;
	register int				n, nSpace ;
	MYLVITEM					lvI ;
	MYCHAR						rszBuffer [256 + 16] ;

	lvI.mask		= LVIF_TEXT | LVIF_PARAM | LVIF_STATE ;
	lvI.state		= 0 ; 
	lvI.stateMask	= 0 ;
	n				= MyListView_GetItemCount (hwndLV) ;
	pNode			= lstRomaKanaTbl ;
	while (pNode != NULL) {
		lvI.iItem		= n ;
		lvI.iImage		= 0 ;
		lvI.iSubItem	= 0 ;
		lvI.lParam		= (LPARAM) pNode ;
		nSpace			= Mylstrlen (pNode->m_strPrefix) ;
		nSpace			= (nSpace > 4)? 4 : nSpace ;
		Mywsprintf (rszBuffer, MYTEXT ("%s%c%s = %s"),
					pNode->m_strPrefix,
					pNode->m_chVowel,
					MYTEXT("    ") + nSpace,
					(fKatakana)? pNode->m_strKatakana : pNode->m_strKana) ;
		lvI.pszText		= rszBuffer ;
		if (MyListView_InsertItem (hwndLV, &lvI) == -1) {
			DEBUGPRINTF ((MYTEXT ("MyListView_InsertItem() error\n"))) ;
		}
		n		++ ;
		pNode	= pNode->m_pNext ;
	}
	return	TRUE ;
}

/*	uListView  DeleteAllItems ƃRXg傫B[Uɂ炩ɕB
 *	  item ̐ςȂ̂ȂAsetitemtext Ő؂ւ̂gvɏ]āA
 *	item text ̂ݕύXB
 */
BOOL
skkimeConfig_switchKanaKata (
	register HWND	hwndLV,
	register BOOL	fKatakana)
{
	register ROMAKANATABLENODE*	pNode ;
	register int				i, nSpace, nMaxItem ;
	MYLVITEM					lvI ;
	MYCHAR						rszBuffer [256 + 16] ;

	if (hwndLV == NULL)
		return	FALSE ;

	nMaxItem		= MyListView_GetItemCount (hwndLV) ;
	for (i = 0 ; i < nMaxItem ; i ++) {
		lvI.mask		= LVIF_PARAM ;
		lvI.iItem		= i ;
		lvI.iSubItem	= 0 ;
		if (MyListView_GetItem (hwndLV, &lvI) == TRUE &&
			lvI.lParam != 0) {
			pNode		= (ROMAKANATABLENODE *) lvI.lParam ;
			nSpace		= Mylstrlen (pNode->m_strPrefix) ;
			nSpace		= (nSpace > 4)? 4 : nSpace ;
			Mywsprintf (rszBuffer, MYTEXT ("%s%c%s = %s"),
						pNode->m_strPrefix,
						pNode->m_chVowel,
						MYTEXT("    ") + nSpace,
						(fKatakana)? pNode->m_strKatakana : pNode->m_strKana) ;
			MyListView_SetItemText (hwndLV, i, 0, rszBuffer) ;
		}
	}
	return	TRUE ;
}

int	CALLBACK
skkimeConfig_compareRomaKanaTblNode (
	register LPARAM		lParam1,
	register LPARAM		lParam2,
	register LPARAM		lParamSort)
{
	register ROMAKANATABLENODE*	pNode1	= (ROMAKANATABLENODE*) lParam1 ;
	register ROMAKANATABLENODE*	pNode2	= (ROMAKANATABLENODE*) lParam2 ;
	register LPCMYSTR	str1 ;
	register LPCMYSTR	str2 ;
	register int		nCompare ;

	str1	= (pNode1->m_strPrefix != NULL)? pNode1->m_strPrefix : MYTEXT ("") ;
	str2	= (pNode2->m_strPrefix != NULL)? pNode2->m_strPrefix : MYTEXT ("") ;

	nCompare	= Mylstrcmp (str1, str2) ;
	if (nCompare == 0) 
		return	(pNode1->m_chVowel - pNode2->m_chVowel) ;

	return	nCompare ;
	UNREFERENCED_PARAMETER (lParamSort) ;
}

/*========================================================================
 *	private functions
 */
BOOL
skkimeConfig_initRomaKanaTable (
	register TINYMEMMGR*			pMemMgr,
	register LPCMYSTR				pstrSubKey,
	register ROMAKANATABLENODE**	plstTableTop,
	register ROMAKANATABLE*			pDefaultTable,
	register int					nDefaultTable,
	register int					nChVowel)
{
	HKEY				hSubKey ;
	MYCHAR				szRegPathInfo [MAX_PATH] ;
	MYCHAR				rbyData [MAX_REGVALUE_SIZE] ;
	DWORD				dwRegPathInfo, dwType, cbData ;
	register int		nIndex ;
	register LONG		lResult ;
	register int		nPrefixLen, nKanaLen, nKataLen ;
	register ROMAKANATABLENODE*	pNode ; 
	register ROMAKANATABLENODE*	pLastNode ; 

	ASSERT (pstrSubKey   != NULL) ;
	ASSERT (plstTableTop != NULL) ;

	pLastNode	= NULL ;
	pNode		= NULL ;
	if (MyRegOpenKeyEx (HKEY_CURRENT_USER, pstrSubKey, 0, KEY_READ, &hSubKey) == ERROR_SUCCESS) {
		register BOOL	fError	= FALSE ;

		/*	[}ϊe[uo^Ă̂ȂcB*/
		for (nIndex = 0 ; nIndex < INT_MAX ; nIndex ++) {
			dwRegPathInfo	= ARRAYSIZE (szRegPathInfo) ;
			cbData			= ARRAYSIZE (rbyData) ;
			lResult	= MyRegEnumValue (hSubKey, (DWORD)nIndex, szRegPathInfo, &dwRegPathInfo, NULL, NULL, NULL, NULL) ;
			if (lResult != ERROR_SUCCESS)
				break ;
			/*	szRegPathInfo = "ky" ƂvtBNXɑB
			 *	Value = REG_MULTI_SZ ŁA, Љ̏ԁB񒷂 0 ̏ꍇɂ
			 *	``nul nul'' ɂȂċI[Ă܂...擪ꕶ]ɂB
			 *	\0Љ\0\0 ̍\AƁB
			 */
			lResult	= MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, &dwType, NULL, &cbData) ;
			if (lResult != ERROR_SUCCESS || dwType != REG_MULTI_SZ)
				continue ;
			/*	Too big value. */
			if (cbData >= sizeof (rbyData)) 
				continue ;
			cbData	= sizeof (rbyData) ;
			(void) MyRegQueryValueEx (hSubKey, szRegPathInfo, NULL, NULL, (BYTE*)rbyData, &cbData) ;

			nPrefixLen	= Mylstrlen (szRegPathInfo) ;
			nKanaLen	= Mylstrlen (rbyData) ;
			nKataLen	= Mylstrlen (rbyData + nKanaLen + 1) ;
			/*	擪 "_" tĂBꂪȂĂƂ̂͂B*/
			if (nKanaLen < 1 || nKataLen < 1)
				continue ;

			pNode		= skkimeConfig_newRomaKanaTableNode (pMemMgr, nPrefixLen, nKanaLen, nKataLen) ;
			if (pNode == NULL) {
				fError	= TRUE ;
				break ;
			}
			Mylstrcpy (pNode->m_strPrefix,	szRegPathInfo) ;
			/*	"_" Ƃ΂ăRs[B*/
			Mylstrcpy (pNode->m_strKana,	(LPCMYSTR)rbyData + 1) ;
			Mylstrcpy (pNode->m_strKatakana,(LPCMYSTR)rbyData + nKanaLen + 1 + 1) ;
			pNode->m_chVowel	= nChVowel ;

			pNode->m_pPrev	= pLastNode ;
			if (pLastNode != NULL)
				pLastNode->m_pNext	= pNode ;
			pLastNode		= pNode ;
		}
		RegCloseKey (hSubKey) ;
		if (fError)
			return	FALSE ;
	} else if (pDefaultTable != NULL && nDefaultTable > 0) {
		register ROMAKANATABLE*	pTable ;
		register int			nPrefixLen, nKanaLen, nKataLen ;

		pTable	= pDefaultTable ;
		/*	ftHg̏sB*/
		for (nIndex = 0 ; nIndex < nDefaultTable ; nIndex ++, pTable ++) {
			nPrefixLen	= Mylstrlen (pTable->prefix)   + 1 ;
			nKanaLen	= Mylstrlen (pTable->kana)     + 1 ;
			nKataLen	= Mylstrlen (pTable->katakana) + 1 ;

			pNode		= skkimeConfig_newRomaKanaTableNode (pMemMgr, nPrefixLen, nKanaLen, nKataLen) ;
			if (pNode == NULL)
				return	FALSE ;

			/*	MYSTR  TSTR ɕϊRs[BƂĂ̏I[͂ĂB*/
			Mylstrcpy (pNode->m_strPrefix,	 pTable->prefix) ;
			Mylstrcpy (pNode->m_strKana,	 pTable->kana) ;
			Mylstrcpy (pNode->m_strKatakana, pTable->katakana) ;
			pNode->m_chVowel	= nChVowel ;
			DEBUGPRINTF ((MYTEXT ("prefix\"%s\", Kana\"%s\", Kata\"%s\"\n"),
						  pNode->m_strPrefix, pNode->m_strKana, pNode->m_strKatakana)) ;
			pNode->m_pPrev	= pLastNode ;
			if (pLastNode != NULL)
				pLastNode->m_pNext	= pNode ;
			pLastNode		= pNode ;
		}
	}
	if (pLastNode != NULL) {
		pNode	= pLastNode ;
		while (pNode->m_pPrev != NULL)
			pNode	= pNode->m_pPrev ;
		*plstTableTop	= pNode ;
	} else {
		*plstTableTop	= pNode ;
	}
	return	TRUE ;
}

BOOL
skkimeConfig_updateRomaKanaTable (
	register LPCMYSTR				pstrSubKey,
	register ROMAKANATABLENODE*		lstTableTop)
{
	register ROMAKANATABLENODE*	pNode ; 
	register int	nKataLen, nKanaLen, nTotal ;
	MYCHAR	szData [MAX_REGVALUE_SIZE] ;
	HKEY	hSubKey ;

	if (! skkimeConfig_CreateKey (pstrSubKey, TRUE, &hSubKey)) {
		DEBUGPRINTF ((MYTEXT ("skkimeConfig_updateRomaKanaTable (): create-key(\"%s\") failed\n"),
					  pstrSubKey)) ;
		return	FALSE ;
	}

	pNode	= lstTableTop ;
	while (pNode != NULL) {
		nKanaLen	= (pNode->m_strKana != NULL)?     Mylstrlen (pNode->m_strKana) : 0 ;
		nKataLen	= (pNode->m_strKatakana != NULL)? Mylstrlen (pNode->m_strKatakana) : 0 ;
		nTotal		= nKanaLen + nKataLen ;
		if (0 <= nTotal && nTotal < (MAX_REGVALUE_SIZE - 5)) {
			register LPMYSTR	pDest	= szData ;
			Mylstrcpy (pDest, MYTEXT ("_")) ;			pDest ++ ;
			Mylstrcpy (pDest, pNode->m_strKana) ;		pDest += nKanaLen + 1 ;
			Mylstrcpy (pDest, MYTEXT ("_")) ;			pDest ++ ;
			Mylstrcpy (pDest, pNode->m_strKatakana) ;	pDest += nKataLen + 1 ;
			*pDest	= MYTEXT ('\0') ;
			if (MyRegSetValueEx (hSubKey, pNode->m_strPrefix, 0, REG_MULTI_SZ, (BYTE *)szData, sizeof (MYCHAR) * (nTotal + 5 /* ``_'' x 2 + NUL x 3*/)) != ERROR_SUCCESS) {
				DEBUGPRINTF ((MYTEXT ("RegSetValueEx: failed (\"%s\", \"%s\")\n"),
							  pNode->m_strPrefix, szData)) ;
				break ;
			}
		}
		pNode	= pNode->m_pNext ;
	}
	MyRegCloseKey (hSubKey) ;
	return	pNode == NULL ;
}

ROMAKANATABLENODE*
skkimeConfig_newRomaKanaTableNode (
	register TINYMEMMGR*	pMemMgr,
	register int			nPrefixLen,
	register int			nKanaLen,
	register int			nKataLen)
{
	register ROMAKANATABLENODE*	pNode ;

	/*	̒iKł͂܂ TCHAR ŊǗBMYCHAR ւ̕ϊ͍Ō̒iKŁB
	 *	ӂׂ nKataLen, nKanaLen ɂ '\0' ܂Ŋ܂񂾃TCY
	 *	B
	 */
	pNode		= TinyMalloc_Malloc (pMemMgr, sizeof (ROMAKANATABLENODE) + sizeof (MYCHAR) * (nPrefixLen + 1 + nKanaLen + nKataLen)) ;
	pNode->m_strPrefix		= (LPMYSTR)(pNode + 1) ;
	pNode->m_strKana		= pNode->m_strPrefix + nPrefixLen + 1 ;
	pNode->m_strKatakana	= pNode->m_strKana   + nKanaLen ;
	pNode->m_pPrev			= NULL ;
	pNode->m_pNext			= NULL ;
	return	pNode ;
}

BOOL
skkimeConfig_registerRomaKanaTableNode (
	register SkkimeConfigArg*		pArg,
	register ROMAKANATABLENODE*		pNode)
{
	register ROMAKANATABLENODE**	pListTop ;
	register ROMAKANATABLENODE*		pNextNode ;

	if (pNode == NULL || pArg == NULL)
		return	FALSE ;
	pListTop	= skkimeConfig_getRomaKanaTableListTop (pArg, pNode->m_chVowel) ;
	if (pListTop == NULL)	/* fatal */
		return	FALSE ;
	pNextNode		= *pListTop ;
	pNode->m_pNext	= pNextNode ;
	if (pNextNode != NULL)
		pNextNode->m_pPrev	= pNode ;
	*pListTop		= pNode ;
	return	TRUE ;
}

BOOL
skkimeConfig_unregisterRomaKanaTableNode (
	register SkkimeConfigArg*		pArg,
	register ROMAKANATABLENODE*		pNode)
{
	register ROMAKANATABLENODE**	pListTop ;
	register ROMAKANATABLENODE*		pPrevNode ;
	register ROMAKANATABLENODE*		pNextNode ;

	if (pNode == NULL || pArg == NULL)
		return	FALSE ;
	pListTop	= skkimeConfig_getRomaKanaTableListTop (pArg, pNode->m_chVowel) ;
	if (pListTop == NULL)	/* fatal */
		return	FALSE ;

	pPrevNode	= pNode->m_pPrev ;
	pNextNode	= pNode->m_pNext ;
	if (pPrevNode == NULL) {
		if (pNode != *pListTop) {
			DEBUGPRINTF ((MYTEXT ("fatal\n"))) ;
			return	FALSE ;
		}
		*pListTop	= pNextNode ;
	} else {
		pPrevNode->m_pNext	= pNextNode ;
	}
	if (pNextNode != NULL) 
		pNextNode->m_pPrev	= pPrevNode ;
	pNode->m_pPrev	= NULL ;
	pNode->m_pNext	= NULL ;
	return	TRUE ;
}

ROMAKANATABLENODE*
skkimeConfig_findRomaKanaTableNode (
	register SkkimeConfigArg*		pArg,
	register LPCMYSTR				strPrefix,
	register int					nChVowel)
{
	register ROMAKANATABLENODE**	plstTop ;
	register ROMAKANATABLENODE*		pNode ;

	if (strPrefix == NULL || pArg == NULL)	/* fatal */
		return	NULL ;
	plstTop	= skkimeConfig_getRomaKanaTableListTop (pArg, nChVowel) ;
	if (plstTop == NULL)	/* fatal */
		return	NULL ;
	pNode	= *plstTop ;
	while (pNode != NULL) {
		if (pNode->m_strPrefix != NULL && !Mylstrcmp (pNode->m_strPrefix, strPrefix))
			return	pNode ;	/* found */
		pNode	= pNode->m_pNext ;
	}
	return	NULL ;	/* not found */
}

ROMAKANATABLENODE**
skkimeConfig_getRomaKanaTableListTop (
	register SkkimeConfigArg*		pArg,
	register int					chVowel)
{
	switch (chVowel) {
	case	MYTEXT ('a'):
		return	&pArg->m_lstRomaKanaA ;
	case	MYTEXT ('i'):
		return	&pArg->m_lstRomaKanaI ;
	case	MYTEXT ('u'):
		return	&pArg->m_lstRomaKanaU ;
	case	MYTEXT ('e'):
		return	&pArg->m_lstRomaKanaE ;
	case	MYTEXT ('o'):
		return	&pArg->m_lstRomaKanaO ;
	default:
		break ;
	}
	return	NULL ;
}

